/*******************************************************************************
 * Copyright (c) 2009, Rockwell Automation, Inc.
 * All rights reserved.
 *
 ******************************************************************************/
#ifndef ENCAP_H_
#define ENCAP_H_

#include "typedefs.h"


/**  \defgroup ENCAP OpENer Ethernet encapsulation layer
 * The Ethernet encapsulation layer handles provides the abstraction between the Ethernet and the CIP layer.
 */

/*** defines ***/

#define ENCAPSULATION_HEADER_LENGTH 24
#define OPENER_ETHERNET_PORT 0xAF12

/* definition of status codes in encapsulation protocol */
#define OPENER_ENCAP_STATUS_SUCCESS                     0x0000
#define OPENER_ENCAP_STATUS_INVALID_COMMAND             0x0001
#define OPENER_ENCAP_STATUS_INSUFFICIENT_MEM            0x0002
#define OPENER_ENCAP_STATUS_INCORRECT_DATA              0x0003
#define OPENER_ENCAP_STATUS_INVALID_SESSION_HANDLE      0x0064
#define OPENER_ENCAP_STATUS_INVALID_LENGTH              0x0065
#define OPENER_ENCAP_STATUS_UNSUPPORTED_PROTOCOL        0x0069


/*** structs ***/
struct S_Encapsulation_Data
{
    EIP_UINT16 nCommand_code;
    EIP_UINT16 nData_length;
    EIP_UINT32 nSession_handle;
    EIP_UINT32 nStatus;
    /* The sender context is not needed any more with the new minimum data copy design */
    /* EIP_UINT8 anSender_context[SENDER_CONTEXT_SIZE];  */
    EIP_UINT32 nOptions;
    EIP_UINT8 *m_acCommBufferStart;       /*Pointer to the communication buffer used for this message */
    EIP_UINT8 *m_acCurrentCommBufferPos;  /*The current position in the communication buffer during the decoding process */
};

struct S_Encapsulation_Interface_Information
{
    EIP_UINT16 TypeCode;
    EIP_UINT16 Length;
    EIP_UINT16 EncapsulationProtocolVersion;
    EIP_UINT16 CapabilityFlags;
    EIP_INT8 NameofService[16];
};

// This struct describes the List Identity reply.
struct S_Encapsulation_ListIdentityResponse
{
    EIP_UINT16	command;                //
    EIP_UINT16  length;                 //
    EIP_UINT32  session_handle;         //
    EIP_UINT32  status;                 //
    EIP_UINT16  sender_context;         //
    EIP_UINT8   data_array[6];          //
    EIP_UINT32  options;                //
    EIP_UINT16  item_count;             //
    EIP_UINT16  item_id;                //
    EIP_UINT16  item_length;            //
    EIP_UINT16  encap_version;          //
    EIP_UINT8   sock_family[2];         //
    EIP_UINT8   sock_port[2];           //
    EIP_UINT8   sock_addr[4];           //
    EIP_UINT8   sock_zero[8];           //
    EIP_UINT16  vendor_id;              //
    EIP_UINT16  device_type;            //
    EIP_UINT16  product_code;           //
    EIP_UINT8   revision[2];            //
    EIP_UINT16  current_status;         //
    EIP_UINT16  serial_number_low;      // Note: due to 32-bit alignment issue with the Serial Number...
    EIP_UINT16  serial_number_high;     //       segment the 32-bit serial number into 2 short integers and..
    EIP_UINT8   name_length;            //       add the length of the product name plus
    EIP_UINT8   product_name[32];       //       the maximum characters are contained in the product name
    EIP_UINT8   product_state;          //       and product state.
};
#define LIST_RESPONSE_NAME_STATE_LENTGH 33
#define LIST_RESPONSE_LENGTH_WITOUT_NAME (sizeof(struct S_Encapsulation_ListIdentityResponse) - LIST_RESPONSE_NAME_STATE_LENTGH)

// This struct describes the Register Session Response fields.
struct S_Encapsulation_RegisterSessionResponse
{
    EIP_UINT16	command;
    EIP_UINT16  length;
    EIP_UINT32  session_handle;
    EIP_UINT32  status;
    EIP_UINT16  sender_context;
    EIP_UINT8   data_array[6];
    EIP_UINT32  options;
    EIP_UINT16  protocol_version;
    EIP_UINT16  option_flags;
};

// This struct describes the Common Packet format fields for address & data items.
typedef struct _S_Common_Packet_HeaderT
{
    EIP_UINT16  type_id;
    EIP_UINT16  length;
}S_Common_Packet_Header;

// This struct describes the UCMM Explicit Message Service Code Request fields.
struct S_Service_Code_Request_Format
{
    EIP_UINT32	session_handle;				// Session Handle

    // UDP IO connection parameters
    EIP_UINT32	o_to_t_connection_id;		// O->T network connection id
    EIP_UINT32	t_to_o_connection_id;		// T->O network connection id
    EIP_UINT32	o_to_t_rpi;					// O->T requested packet interval (us)
    EIP_UINT32	t_to_o_rpi;					// T->O requested packet interval (us)
    EIP_UINT32	o_to_t_sequence_num;		    // O->T sequence number
    EIP_UINT32	t_to_o_sequence_num;		    // T->O sequence number
    EIP_UINT32	interface_handle;			// Connection Interface

    EIP_UINT16	connection_serial_number;	// Connection Serial Number
    EIP_UINT16	vendor_id;					// Vendor identification number.

    EIP_UINT32	originator_serial_number;	// Originator Serial Number

    EIP_UINT8	timeout;					    // Connection time-out period
    EIP_UINT8	spare1;						// 32-bit alignment spare
    EIP_UINT8	spare2;						// 32-bit alignment spare
    EIP_UINT8	spare3;						// 32-bit alignment spare

    // TCP service code request parameters
    EIP_UINT8  service_code;
    EIP_UINT8  logical_class;
    EIP_UINT8  logical_instance;
    EIP_UINT8  logical_attribute;

    EIP_UINT16 output_assembly_size;
    EIP_UINT16 input_assembly_size;
    EIP_UINT8  output_assembly_object;
    EIP_UINT8  input_assembly_object;

    EIP_UINT16 data_length;
    EIP_BYTE   data_buf[512];

};

// This struct describes the UCMM Explicit Message Common Industrial Protocol Request fields.
struct S_Service_Request_Format
{
    EIP_UINT8  service_request;
    EIP_UINT8  request_path_size;
    EIP_UINT8  logical_class_segment;
    EIP_UINT8  logical_class;
    EIP_UINT8  logical_instance_segment;
    EIP_UINT8  logical_instance;
    EIP_UINT8  logical_attribute_segment;
    EIP_UINT8  logical_attribute;
};

typedef struct _S_Connection_PathT
{
    EIP_UINT8  logical_class_segment;
    EIP_UINT8  logical_class;
    EIP_UINT8  logical_instance_segment;
    EIP_UINT8  logical_instance;
}S_Connection_Path;

struct S_Service_Reply_Format
{
    EIP_UINT8  response;
    EIP_UINT8  status;
};

// This struct describes the UCMM Explicit Message Response fields.
struct S_Encapsulation_UCMMResponse
{
    EIP_UINT16	command;
    EIP_UINT16  length;
    EIP_UINT32  session_handle;
    EIP_UINT32  status;
    EIP_UINT8   sender_context_data[8];
    EIP_UINT32  options;
    EIP_UINT32  interface_handle;
    EIP_UINT16  timeout;
    EIP_UINT16  item_count;
    S_Common_Packet_Header  address_item;
    S_Common_Packet_Header  data_item;
    EIP_UINT8   service_response;       // Service Response = Request | 0x80
    EIP_UINT8   reserved_spare;
    EIP_UINT8   service_status;			// Service Status
    EIP_UINT8   additional_status_size;	// Additional Status if needed.
};

// This struct describes the UCMM Explicit Message Response fields.
struct S_Encapsulation_SendRRDataResponse
{
    EIP_UINT16	command;
    EIP_UINT16  length;
    EIP_UINT32  session_handle;
    EIP_UINT32  status;
    EIP_UINT8   sender_context_data[8];
    EIP_UINT32  options;
    EIP_UINT32  interface_handle;
    EIP_UINT16  timeout;
    EIP_UINT16  item_count;
    EIP_UINT16  address_item_type_id;
    EIP_UINT16  address_item_length;
    EIP_UINT16  data_item_type_id;
    EIP_UINT16  data_item_length;
    EIP_UINT8   service_response;       // Service Response = Request | 0x80
    EIP_UINT8   reserved_spare;
    EIP_UINT8   service_status;			// Service Status
    EIP_UINT8   additional_status_size;	// Additional Status if needed.
};


// This struct describes the Forward Open Request fields.
struct S_Service_Request_ForwardOpen
{
    EIP_UINT32  session_handle;

    EIP_UINT32  interface_handle;
    EIP_UINT16  time_out;
    EIP_UINT16  item_count;              // Command Specific Data

    S_Common_Packet_Header address_item;
    S_Common_Packet_Header data_item;

    EIP_UINT8  service_request;          // Service Request
    EIP_UINT8  request_path_size;

    S_Connection_Path request_path;      // Connection Manager request path class, instance

    EIP_UINT8  priority_time_tick;
    EIP_UINT8  timeout_ticks;
    EIP_UINT32 o_to_t_net_conn_id;      	// O->T network connection id
    EIP_UINT32 t_to_o_net_conn_id;      	// T->O network connection id
    EIP_UINT16 conn_serial_number;
    EIP_UINT16 originator_vendor_id;
    EIP_UINT32 originator_serial_number;
    EIP_UINT8  conn_timeout_multiplier;
    EIP_UINT8  reserve_1;
    EIP_UINT8  reserve_2;
    EIP_UINT8  reserve_3;
    EIP_UINT32 o_to_t_rpi;               // O->T requested packet interval
    EIP_UINT16 o_to_t_net_conn_param;    // Forward Open O->T network connection parameters
    EIP_UINT32 t_to_o_rpi;               // O->T requested packet interval
    EIP_UINT16 t_to_o_net_conn_param;    // Forward Open T->O network connection parameters
    EIP_UINT8  transport_trigger;        // Transport Class Trigger

    EIP_UINT8  connection_path_size;     // Number of 16-bit words in the connection path field.
    S_Connection_Path connection_path;   // Message Router connection path class, instance

};

// This struct describes the Forward Open Reply fields.
struct S_Service_Reply_ForwardOpen
{
    EIP_UINT16	command;
    EIP_UINT16  length;
    EIP_UINT32  session_handle;
    EIP_UINT32  status;
    EIP_UINT8	sender_context_data[8];
    EIP_UINT32  options;
    EIP_UINT32  interface_handle;
    EIP_UINT16  timeout;
    EIP_UINT16  item_count;             // Command Specific Data

    S_Common_Packet_Header address_item;
    S_Common_Packet_Header data_item;

    EIP_UINT8  service_response;        // Service Response = Request | 0x80
    EIP_UINT8  reserved_spare;
    EIP_UINT8  service_status;			// Service Status
    EIP_UINT8  additional_status_size;	// Additional Status if needed.

    EIP_UINT32 o_to_t_net_conn_id;  	    // O->T network connection id
    EIP_UINT32 t_to_o_net_conn_id;		// T->O network connection id

    EIP_UINT16 connection_serial_number;
    EIP_UINT16 vendor_id;

    EIP_UINT32 originator_serial_number;

    EIP_UINT32 o_to_t_api;               // O->T application packet interval
    EIP_UINT32 t_to_o_api;               // T->O application packet interval

    EIP_UINT8  app_reply_size;
    EIP_UINT8  alignment_spare1;
    EIP_UINT8  alignment_spare2;
    EIP_UINT8  alignment_spare3;

};

// This struct describes the IO Connection received data fields.
struct S_Connected_UDP_IO_Packet
{
    EIP_UINT16  item_count;              // item count
    S_Common_Packet_Header address_item; // address type id and length
    EIP_UINT32 connection_id;			 // connection id number
    EIP_UINT32 sequence_id;				 // sequence id number
    S_Common_Packet_Header data_item;    // data type id and length
    EIP_UINT16 sequence_number;          // 16-bit version of the sequence number
    EIP_UINT8  data_buf[2048];			 // buffer containing assembly data
};

/*** global variables (public) ***/

/*** public functions ***/
/*! \ingroup ENCAP
 * \brief Initialize the encapsulation layer.
 */
void encapInit(void);

/*! \ingroup ENCAP
 * \brief Shutdown the encapsulation layer.
 *
 * This means that all open sessions including their sockets are closed.
 */
void encapShutDown(void);


/*! \ingroup ENCAP
 * \brief Handle delayed encapsulation message responses
 *
 * Certain encapsulation message requests require a delayed sending of the response
 * message. This functions checks if messages need to be sent and performs the
 * sending.
 */
void manageEncapsulationMessages();

/*! \ingroup ENCAP
 * \brief Send encapsulation message list Identity request with maximum reponse delay.
 *
 */
EIP_INT32 encapsulateListIdentyRequestMessage(EIP_BYTE *pa_pacCommBuf, EIP_UINT16 MaxResponseDelay);

/*! \outgroup ENCAP
 * \brief Send encapsulation message list interfaces request. EIP standard dose not specify data
 * so this may not be useful at this time.
 *
 */
EIP_INT32 encapsulateListInterfacesRequestMessage(EIP_BYTE *pa_pacCommBuf);

/*! \outgroup ENCAP
 * \brief Send encapsulation message list services request.
 *
 */
EIP_INT32 encapsulateListServicesRequestMessage(EIP_BYTE *pa_pacCommBuf);

/*! \outgroup ENCAP
 * \brief Send encapsulation message NOP request. If needed, used to keep connection active.
 *
 */
EIP_INT32 encapsulateNopRequestMessage(EIP_BYTE *pa_pacCommBuf);

/*! \outgroup ENCAP
 * \brief Send encapsulation message register session request.
 *
 */
EIP_INT32 encapsulateRegisterSessionRequestMessage(EIP_BYTE *pa_pacCommBuf);

/*! \outgroup ENCAP
 * \brief Send encapsulation message un-register session request.
 *
 */
EIP_INT32 encapsulateUnRegisterSessionRequestMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request);

/*! \outgroup ENCAP
 * \brief Send encapsulation message UCMM SendRRData request.
 *
 */
EIP_INT32 encapsulateUCMMSendRRDataRequestMessage(EIP_BYTE *pa_pacCommBuf,
        struct S_Service_Code_Request_Format *pa_cip_request);


/*! \outgroup ENCAP
 * \brief Send encapsulation message explicit messaging ForwardOpen SendRRData request.
 *
 */
EIP_INT32 encapsulateSendForwardOpenRequestMessage(EIP_BYTE *pa_pacCommBuf,
        struct S_Service_Code_Request_Format *pa_cip_request);

/*! \outgroup ENCAP
 * \brief Send encapsulation message explicit messaging LargeForwardOpen SendRRData request.
 *
 */
EIP_INT32 encapsulateSendLargeForwardOpenRequestMessage(EIP_BYTE *pa_pacCommBuf,
        struct S_Service_Code_Request_Format *pa_cip_request);

/*! \outgroup ENCAP
 * \brief Send encapsulation message I/O messaging ForwardOpen SendRRData request.
 *
 */
EIP_INT32 encapsulateSendIOLargeForwardOpenRequestMessage(EIP_BYTE *pa_pacCommBuf,
        struct S_Service_Code_Request_Format *pa_cip_request);

/*! \outgroup ENCAP
 * \brief Send encapsulation message I/O messaging ForwardOpen SendRRData request.
 *
 */
EIP_INT32 encapsulateSendIOForwardOpenRequestMessage(EIP_BYTE *pa_pacCommBuf,
        struct S_Service_Code_Request_Format *pa_cip_request);



/*! \outgroup ENCAP
 * \brief Send encapsulation message ForwardClose SendRRData request.
 *
 */
EIP_INT32 encapsulateSendForwardCloseRequestMessage(EIP_BYTE *pa_pacCommBuf, EIP_UINT32 session_handle);

/*! \outgroup ENCAP
 * \brief Send encapsulation message explicit messaging SendUnitData request.
 *
 */
EIP_INT32 encapsulateSendUnitDataRequestMessage(EIP_BYTE *pa_pacCommBuf,
        struct S_Service_Code_Request_Format *pa_cip_request);

/*! \outgroup ENCAP
 * \brief Send i/o message implicit messaging request.
 *
 */
EIP_INT32 encapsulateSendIOKeepAliveMessage(EIP_BYTE *pa_pacCommBuf,
        struct S_Service_Code_Request_Format *pa_cip_request);
#endif /*ENCAP_H_*/
