/********************************************************************************************************
* OMRON Scientific Technologies Inc									  									*
* Project: 0949 - Next generation Laser Scanner, OS32C													*
* Build: OS32C client application    																    *
* 																										*
* $Id: os32c_client.cpp 2273 2013-01-02 18:47:57Z tom.charlton $
* Description: Defines the entry point for the EtherNet/IP console application.							*
********************************************************************************************************/

/*
*********************************************************************************************************
*                                              EXAMPLE CODE
*
*               (c) Copyright Omron Scientific Technologies, Inc. Fremont, CA
*
*               This OS32C EtherNet/IP example code is intended to assist developers of communication
*               software. This software is provided "AS IS" without any expressed or implied warranties. 
*********************************************************************************************************
*/

#include "stdafx.h"
#include <winsock2.h>
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <time.h>
#include <iostream>
#include <fstream>
using namespace std;

#include "typedefs.h"
#include "encap.h"
#include "os32c_eip.h"
#include "os32c_eip_interfacedll.h"

#define TXBUFSIZE			512			/* Size of Tx buffer */
#define RXBUFSIZE			(4*1024)	/* Size of Rx buffer */
#define MAX_UDP_MTU			1472		/* Maximum UDP transmission unit size */

// Build Options
//#define __USE_INPUT_ARGUMENTS__         /* Use input arguments during console execution to define local and target IP addresses */
//#define __USE_EIP_DLL__                 /* Use supplied DLL communication functions which can be used with C#(C-sharp) applications.  */
#define __LOG_TIMESTAMP_RESULTS__	    /* Log TCP explicit request/response timestamp test results to a file   */
#define __ENABLE_UDP_IO_MESSAGES__      /* Enable EtherNet/IP UDP IO connection messages which uses timer based communications  */

// Define starting and ending measurement angles plus the target angle(beam) used in this example. Arguments - float  Angle (-0.4 to 270.4 degrees)
#define START_ANGLE (float)(45.2)
#define END_ANGLE (float)(225.2)
#define TARGET_ANGLE (float)(135.2)

// Test samples required for this example.
#define NUM_OF_TCP_SAMPLES     30
#define NUM_OF_UDP_SAMPLES     20

#define LOCAL_IP_ADDRESS       "192.168.1.7";    // Set default PC local IP address
#define OS32C_IP_ADDRESS	   "192.168.1.1";    // Set default OS32C target IP address

//*********************************************************************************************************************
// Local Prototypes
//*********************************************************************************************************************
// main functions
int Initialize_UserSockets(void);
int Get_ScannerIdentity(void);
int RegisterScannerSession(struct S_Service_Code_Request_Format *pa_cip_request);
int Configure_ScannerMeasurementReport(struct S_Service_Code_Request_Format *pa_cip_request);
int Initiate_ScannerTCPMeasurements(struct S_Service_Code_Request_Format *pa_cip_request, EIP_UINT32 num_cycles);
int Initiate_ScannerUDPIOMeasurements(struct S_Service_Code_Request_Format *pa_cip_request, EIP_UINT32 num_cycles);
int Read_ScannerAttributes(struct S_Service_Code_Request_Format *pa_cip_request);
int KeepAliveTCPconnection(struct S_Service_Code_Request_Format *pa_cip_request, EIP_UINT32 num_cycles);
int UnRegisterScannerSession(struct S_Service_Code_Request_Format *pa_cip_request);
void Close_UserSockets(void);

// helper functions
bool SetupBeamSelectionMask(float startAngle, float endAngle, unsigned int *beamSelectionData);
EIP_UINT16 GetBeamIndex(float startAngle, float endAngle, float targetAngle);
EIP_UINT16 Get16bit_VSAttribute(EIP_UINT8 vsObject, EIP_UINT8 vsAttribute, EIP_UINT32 session_handle);
EIP_UINT32 Get32bit_VSAttribute(EIP_UINT8 vsObject, EIP_UINT8 vsAttribute, EIP_UINT32 session_handle);
void RecordEipMessageBytes(string fileName, EIP_BYTE *pa_CommBuf, unsigned int msgSize);
void Wait_ms(unsigned int mseconds);

// EtherNet/IP request packet routines
EIP_INT32 Setup_ListIdentityRequestMessage(EIP_BYTE *pa_pacCommBuf, EIP_UINT16 MaxResponseDelay);
EIP_INT32 Setup_RegisterSessionRequestMessage(EIP_BYTE *pa_pacCommBuf);
EIP_INT32 Setup_UnRegisterSessionRequestMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request);
EIP_INT32 Setup_SendRRDataRequestMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request);
EIP_INT32 Setup_NopRequestMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request);
EIP_INT32 Setup_IOLargeForwardOpenRequestMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request);
EIP_INT32 Setup_IOKeepAliveMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request);
EIP_INT32 Setup_IOLargeForwardCloseRequestMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request);

// EtherNet/IP response packet routines
EIP_UINT8 Get_IOLargeForwardOpenResponseMessage(struct S_Service_Code_Request_Format *pa_cip_request, struct S_Service_Reply_ForwardOpen *reply, char *buffer);
void ProcessReceived_IOData(EIP_UINT8 * ioData, EIP_UINT16 dataLen, EIP_UINT8 inputAssembly);
//*********************************************************************************************************************

//*********************************************************************************************************************
// Global Variables
//*********************************************************************************************************************
char			TxBuf[TXBUFSIZE];
char			RxBuf[RXBUFSIZE];
char			RxIoBuf[RXBUFSIZE];
char			*os32c_IpAddr;
char			*local_IpAddr;

// Socket variables
WSADATA			wsaData;
SOCKET          tcpLocalSocket;
SOCKADDR_IN     tcpLocalAddr;
SOCKADDR_IN     tcpServerAddr;
unsigned int    os32c_TcpPort = 44818;
SOCKET          udpLocalSocket;
SOCKADDR_IN     udpLocalAddr;
SOCKADDR_IN     udpServerAddr;
unsigned int    os32c_UdpPort = 44818;
int             udpAddrLen = sizeof(udpServerAddr);
SOCKET          udpIoLocalSocket;
SOCKADDR_IN     udpIoLocalAddr;
SOCKADDR_IN     udpIoServerAddr;
unsigned int    os32c_UdpIoPort = 2222;
int             udpIoAddrLen = sizeof(udpIoServerAddr);

// Session connection variables
S_Service_Code_Request_Format   connectionInfo;

// Scanner configuration variables
S_EIP_RTM_Configuration		    scannerConfig;

// Scanner assembly data variables
S_EIP_SystemStatusInfo		    systemReport;
S_EIP_RTM_Header                measurementReport;
EIP_UINT16					    measurementData[(NUMBER_OF_BEAMS_IN_FRONTSCAN + 1) * 2];
EIP_UINT8					    protectionZoneData[NUM_OF_DETECTION_ZONE_BYTES];
EIP_UINT8					    warningZone1Data[NUM_OF_DETECTION_ZONE_BYTES];
EIP_UINT8					    warningZone2Data[NUM_OF_DETECTION_ZONE_BYTES];

// Scanner Run-time Data
EIP_UINT32		                scanPeriod;
EIP_UINT16		                rangeMeasurementData;
EIP_UINT16		                reflectivityMeasurementData;
EIP_UINT16		                rangeDataIndex;
EIP_UINT16		                reflectivityDataIndex;


//*********************************************************************************************************************

//*********************************************************************************************************************
// main()
//*********************************************************************************************************************
int main(int argc, char **argv)
{
    int retCode;

#ifdef __USE_INPUT_ARGUMENTS__ 
    if (argc < 2)     /* Test for correct number of arguments */
    {
        fprintf(stderr, "Usage: %s [local_IpAddr]  %s [os32c_IpAddr]  \r\n", argv[0], argv[1]);
        return 1;
    }
	local_IpAddr		= argv[0];
	os32c_IpAddr		= argv[1];
#else
	local_IpAddr		= LOCAL_IP_ADDRESS;   // Set default PC local IP address
	os32c_IpAddr		= OS32C_IP_ADDRESS;   // Set default OS32C target IP address
#endif

	// Initialize TCP/IP and UDP/IO sockets used to transmit and receive EtherNet/IP messages.
	retCode = Initialize_UserSockets();
	if(retCode == SOCKET_ERROR)
	{
		printf("\r\nTCP/IP or UDP/IP sockets initialization failed, check connection and IP address assignments.\r\n");
	    	printf("\r\nExisting Application!...please wait\r\n\r\n");
		system("pause");
		return(1);
	}

    // Get the Serial Number and Product Name from the Scanner.
	retCode = Get_ScannerIdentity();
	if(retCode == SOCKET_ERROR)
	{
		printf("\r\nSocket Errors, check connection and IP address assignments.\r\n");
        Close_UserSockets();
		return(1);
	}

    // Register a Session with the Scanner and record the session handle for future use.
    retCode = RegisterScannerSession(&connectionInfo);
    if(retCode == SOCKET_ERROR)
	{
		printf("\r\nSocket Errors, check connection and IP address assignments.\r\n");
        Close_UserSockets();
		return(1);
	}else{
        printf("Session Handle : 0x%08X\r\n", connectionInfo.session_handle);
    }

    // Configure the range, reflectivity and beam selection mask.
    retCode = Configure_ScannerMeasurementReport(&connectionInfo);
    if(retCode == SOCKET_ERROR)
	{
		printf("\r\nSocket Errors, check connection and IP address assignments.\r\n");
        Close_UserSockets();
		return(1);
    }

    // Initiate TCP vendor specific object measurement data request/reply messaging.
    retCode = Initiate_ScannerTCPMeasurements(&connectionInfo, NUM_OF_TCP_SAMPLES);
    if(retCode == SOCKET_ERROR)
	{
		printf("\r\nSocket Errors, check connection and IP address assignments.\r\n");
        Close_UserSockets();
		return(1);
    }

    // Initiate UDP I/O assembly object measurement data streaming
    retCode = Initiate_ScannerUDPIOMeasurements(&connectionInfo, NUM_OF_UDP_SAMPLES);
    if(retCode == SOCKET_ERROR)
	{
		printf("\r\nSocket Errors, check connection and IP address assignments.\r\n");
        Close_UserSockets();
		return(1);
    }

    // Read Scanner System Status and individual attributes.
    retCode = Read_ScannerAttributes(&connectionInfo);
    if(retCode == SOCKET_ERROR)
	{
		printf("\r\nSocket Errors, check connection and IP address assignments.\r\n");
        Close_UserSockets();
		return(1);
    }

    // Keep Alive the TCP connection if there is no current activity. Send only 3 requests.
    retCode = KeepAliveTCPconnection(&connectionInfo, 3);
    if(retCode == SOCKET_ERROR)
	{
		printf("\r\nSocket Errors, check connection and IP address assignments.\r\n");
        Close_UserSockets();
		return(1);
	}

    // UnRegister a Session with the Scanner using the session handle.
    retCode = UnRegisterScannerSession(&connectionInfo);
    if(retCode == SOCKET_ERROR)
	{
		printf("\r\nSocket Errors, check connection and IP address assignments.\r\n");
        Close_UserSockets();
		return(1);
	}else{
        printf("Un-registering Session Handle : 0x%08X\r\n", connectionInfo.session_handle);
    }

    // Close All Sockets
    Close_UserSockets();

    return 0;
}
//*********************************************************************************************************************

//*********************************************************************************************************************
// Display the product name and serial number of the scanner.  
//*********************************************************************************************************************
int Get_ScannerIdentity(void)
{
    int         i, msgSize, retCode;
    EIP_UINT32	deviceSerialNumber;
    EIP_UINT16	maxListIdResponseDelay = 500; // ms
    S_Encapsulation_ListIdentityResponse    listIdentityReply;

    // *********************************************************************************************************************************************
    // TCP example:		Send the List Identity command to retreive the identity structure from the scanner.
    //					Display the product name and serial number members of the identity structure.  
    // *********************************************************************************************************************************************

    // Sends Tcp data to server/receiver...
	printf("\r\nSending TCP List Identity Request... \r\n");

#ifdef __USE_EIP_DLL__
	msgSize = EIP_EncapsulateListIdentityRequestMessage((EIP_BYTE *)&TxBuf, maxListIdResponseDelay);
#else	
	msgSize = Setup_ListIdentityRequestMessage((EIP_BYTE *)&TxBuf, maxListIdResponseDelay);
#endif
    RecordEipMessageBytes("EIP_Setup_ListIdentityRequestMessage.txt", (EIP_BYTE *)&TxBuf, msgSize);

	// Send the request message.
	retCode = send(tcpLocalSocket, TxBuf, msgSize, 0);
	if (retCode == SOCKET_ERROR) 
	{
		printf("\r\nTcp send() error %ld.\r\n", WSAGetLastError());
		return (SOCKET_ERROR);
	}
	
	// Get the reply message from the server/receiver...
	msgSize = recv(tcpLocalSocket, RxBuf, RXBUFSIZE, 0);
	if (msgSize == SOCKET_ERROR) 
	{
		printf("\r\nTcp recv() error %ld.\r\n", WSAGetLastError());
		return (SOCKET_ERROR);
	}
	else
	{
		// extract the returned data from the rx packet.
		(void *)memcpy(&listIdentityReply, &RxBuf[0], msgSize);
		for(i=0; i<listIdentityReply.name_length; i++)
		{
			printf("%c", listIdentityReply.product_name[i]);
		}
		printf("\r\n");

		deviceSerialNumber = (listIdentityReply.serial_number_high << 16) | listIdentityReply.serial_number_low;	
		if ((listIdentityReply.vendor_id == OSTI_VENDOR_ID) && (listIdentityReply.product_code == OS32C_PROD_CODE))
		{
			printf("OS32C Serial Number : AS%05d \r\n", deviceSerialNumber);
		}
		else
		{
			printf("Serial Number : %d \r\n", deviceSerialNumber);
		}
        RecordEipMessageBytes("EIP_Setup_ListIdentityResponseMessage.txt", (EIP_BYTE *)&RxBuf, msgSize);
	}

    // *********************************************************************************************************************************************
    // UDP example:		Send the List Identity command to retreive the identity structure from the scanner.
    //					Display the product name and serial number members of the identity structure.  
    //
    //					Note: this command can be used with a subnet mask to retrieive all of the identities of the devices on the EIP network.  			
    // *********************************************************************************************************************************************
	
    // Sends Udp data to server/receiver...
	printf("\r\nSending UDP List Identity Request... \r\n");
#ifdef __USE_EIP_DLL__
	msgSize = EIP_EncapsulateListIdentityRequestMessage((EIP_BYTE *)&TxBuf, maxListIdResponseDelay);
#else	
	msgSize = Setup_ListIdentityRequestMessage((EIP_BYTE *)&TxBuf, maxListIdResponseDelay);
#endif
	// Send the request message.
	retCode = sendto(udpLocalSocket, TxBuf, msgSize, 0, (SOCKADDR *) &udpServerAddr, udpAddrLen);
	if (retCode == SOCKET_ERROR) 
	{
		printf("\r\nUdp send() error %ld.\r\n", WSAGetLastError());
		return (SOCKET_ERROR);
	}
	
	// Get the reply message from the server/receiver...
	msgSize = recvfrom(udpLocalSocket, RxBuf, RXBUFSIZE, 0, (SOCKADDR *) &udpServerAddr, &udpAddrLen);
	if (msgSize == SOCKET_ERROR) 
	{
		printf("\r\nUdp recv() error %ld.\r\n", WSAGetLastError());
		return (SOCKET_ERROR);
	}
	else
	{
		// extract the returned data from the rx packet.
		(void *)memcpy(&listIdentityReply, &RxBuf[0], msgSize);
		for(i=0; i<listIdentityReply.name_length; i++)
		{
			printf("%c", listIdentityReply.product_name[i]);
		}
		printf("\r\n");

		deviceSerialNumber = (listIdentityReply.serial_number_high << 16) | listIdentityReply.serial_number_low;	
		if ((listIdentityReply.vendor_id == OSTI_VENDOR_ID) && (listIdentityReply.product_code == OS32C_PROD_CODE))
		{
			printf("OS32C Serial Number : AS%05d \r\n", deviceSerialNumber);
		}
		else
		{
			printf("Serial Number : %d \r\n", deviceSerialNumber);
		}
	}
    	return (0);
}
// *********************************************************************************************************************************************

//*********************************************************************************************************************
// Register a session with the scanner and save the session number for future use.
// *********************************************************************************************************************************************
int RegisterScannerSession(struct S_Service_Code_Request_Format *pa_cip_request)
{
    int msgSize, retCode;
    S_Encapsulation_RegisterSessionResponse registerSessionReply;

    // *********************************************************************************************************************************************
    // TCP example : Register a communication session with the scanner.
    //				 Save the session handle from the session reply for use during the remaining portion of the session.
    // *********************************************************************************************************************************************
	// Sends Tcp data to server/receiver...
	printf("\r\nSending TCP Register Session Request... \r\n");
#ifdef __USE_EIP_DLL__
	msgSize = EIP_EncapsulateTcpRegisterSessionRequestMessage((EIP_BYTE *)&TxBuf);
#else
	msgSize = Setup_RegisterSessionRequestMessage((EIP_BYTE *)&TxBuf);
#endif
    RecordEipMessageBytes("EIP_Setup_RegisterSessionRequestMessage.txt", (EIP_BYTE *)&TxBuf, msgSize);

	// Send the request message.
	retCode = send(tcpLocalSocket, TxBuf, msgSize, 0);
	if (retCode == SOCKET_ERROR) 
	{
		printf("\r\nTcp send() error %ld.\r\n", WSAGetLastError());
		return (SOCKET_ERROR);
	}
	
	// Get the reply message from the server/receiver...
	msgSize = recv(tcpLocalSocket, RxBuf, RXBUFSIZE, 0);
	if (msgSize == SOCKET_ERROR) 
	{
		printf("\r\nTcp recv() error %ld.\r\n", WSAGetLastError());
		return (SOCKET_ERROR);
	}
	else
	{
		// extract the returned data from the rx packet.
		(void *)memcpy(&registerSessionReply, &RxBuf[0], msgSize);
		
		// get the session handle from the scanner to be used in the remaining requests.
		pa_cip_request->session_handle = registerSessionReply.session_handle;

        RecordEipMessageBytes("EIP_Setup_RegisterSessionResponseMessage.txt", (EIP_BYTE *)&RxBuf, msgSize);

	}
    return(0);
}
// *********************************************************************************************************************************************

//*********************************************************************************************************************
// Un-Register the session with the scanner after required request/response messages.
// *********************************************************************************************************************************************
int UnRegisterScannerSession(struct S_Service_Code_Request_Format *pa_cip_request)
{
    int msgSize, retCode;

    // *********************************************************************************************************************************************
    // TCP example :	 Un-Registering a session with the scanner.
    //				 Terminate the communication session with the scanner and close the TCP socket.
    // *********************************************************************************************************************************************
	
    // Sends Tcp data to server/receiver...
	printf("\r\nSending TCP Un-Register Session Request... \r\n");
#ifdef __USE_EIP_DLL__
	msgSize = EIP_EncapsulateTcpUnRegisterSessionRequestMessage((EIP_BYTE *)&TxBuf, pa_cip_request);
#else
	msgSize = Setup_UnRegisterSessionRequestMessage((EIP_BYTE *)&TxBuf, pa_cip_request);
#endif
    RecordEipMessageBytes("EIP_Setup_UnRegisterSessionRequestMessage.txt", (EIP_BYTE *)&TxBuf, msgSize);
	
    // Send the request message.
	retCode = send(tcpLocalSocket, TxBuf, msgSize, 0);
	if (retCode == SOCKET_ERROR) 
	{
		printf("Tcp send() error %ld.\r\n", WSAGetLastError());
	}
	
	// Note: 1) There is no reply to the un-register session message request.
	//		 2) Server/OS32C will close the tcp socket connection after receiving this request.

    return(0);
}
// *********************************************************************************************************************************************

//*********************************************************************************************************************
// Configure the run-time measurement attributes used by the scanner.
// *********************************************************************************************************************************************
int Configure_ScannerMeasurementReport(struct S_Service_Code_Request_Format *pa_cip_request)
{
    int msgSize;
    S_Encapsulation_SendRRDataResponse      sendRRDataReply;
    
    // Seutp session control structure and get session handle for this scanner connection.
    S_Service_Code_Request_Format	local_connectionInfo;
    local_connectionInfo.session_handle = pa_cip_request->session_handle;

    // *********************************************************************************************************************************************
    // TCP example : 	Write single attributes to the scanner.
    //					Write new values for the range, reflectivity and the beam selection mask. 
    //					1) Set the range format for 50 meter measurements. 
    //					2) Set the reflectivity format for picosecond measurements.
    //					3) Set the beam selection mask for measurments between 45 and 225 degrees.
    // *********************************************************************************************************************************************
	// Sends Tcp data to server/receiver...
	printf("\r\nSending Configuration TCP SendRRData Explicit Message Service Requests... \r\n");
    local_connectionInfo.service_code = CIP_SC_SET_ATTR_SINGLE;			// Set Single Attribute service code.   
	local_connectionInfo.logical_class = OS32C_VS_RTM_CONFIG_DATA;		// Vendor Specific object
	local_connectionInfo.logical_instance = OS32C_VS_INSTANCE;			// Vendor Specific instance
	local_connectionInfo.logical_attribute = OS32C_VS_RANGE_ATTRIBUTE;  // Range Format attribute

	printf("Writing new Range Format... \r\n");
	local_connectionInfo.data_length = sizeof(scannerConfig.RangeFormat);
	scannerConfig.RangeFormat = RANGE_MEASURE_50M;
	(void *)memcpy(&local_connectionInfo.data_buf[0], &scannerConfig.RangeFormat, sizeof(scannerConfig.RangeFormat));  

#ifdef __USE_EIP_DLL__
	msgSize = EIP_EncapsulateTcpSendRRDataExplicitRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#else
	msgSize = Setup_SendRRDataRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#endif
    RecordEipMessageBytes("EIP_Setup_SendRRData_WriteRequestMessage.txt", (EIP_BYTE *)&TxBuf, msgSize);

	// Send the request message.
	send(tcpLocalSocket, TxBuf, msgSize, 0);
		
	// Get the reply message from the server/receiver...
	msgSize = recv(tcpLocalSocket, RxBuf, RXBUFSIZE, 0);
	if (msgSize != SOCKET_ERROR) 
	{
		(void *)memcpy(&sendRRDataReply, &RxBuf[0], sizeof(sendRRDataReply));
		if (sendRRDataReply.status == EIP_STATUS_SUCCESS)
		{
			printf("New Range Format = %d\r\n", scannerConfig.RangeFormat);
		}
		else
		{
			printf("Status Error : 0x%04X\r\n", sendRRDataReply.status);
		}
        RecordEipMessageBytes("EIP_Setup_SendRRData_WriteResponseMessage.txt", (EIP_BYTE *)&RxBuf, msgSize);

    }else{ 
        return(SOCKET_ERROR);
    }

	local_connectionInfo.logical_attribute = OS32C_VS_REFLECTIVITY_ATTRIBUTE;    	// Reflectivity Format attribute
	printf("Writing new Reflectivity Format... \r\n");
	local_connectionInfo.data_length = sizeof(scannerConfig.ReflectivityFormat);
	scannerConfig.ReflectivityFormat = REFLECT_MEASURE_TOT_4PS;
	(void *)memcpy(&local_connectionInfo.data_buf[0], &scannerConfig.ReflectivityFormat, sizeof(scannerConfig.ReflectivityFormat));  

#ifdef __USE_EIP_DLL__
	msgSize = EIP_EncapsulateTcpSendRRDataExplicitRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#else
	msgSize = Setup_SendRRDataRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#endif
	// Send the request message.
	send(tcpLocalSocket, TxBuf, msgSize, 0);
		
	// Get the reply message from the server/receiver...
	msgSize = recv(tcpLocalSocket, RxBuf, RXBUFSIZE, 0);
	if (msgSize != SOCKET_ERROR) 
	{
		(void *)memcpy(&sendRRDataReply, &RxBuf[0], sizeof(sendRRDataReply));
		if (sendRRDataReply.status == EIP_STATUS_SUCCESS)
		{
			printf("New Reflectivity Format = %d\r\n", scannerConfig.ReflectivityFormat);
		}
		else
		{
			printf("Status Error : 0x%04X\r\n", sendRRDataReply.status);
		}
    }else{ 
        return(SOCKET_ERROR);
    }
	
	// select the measurement beams between 45 degrees to 225 degrees and provide 180 degrees of coverage.
	memset((unsigned char *)&scannerConfig.BeamReportMask, 0, sizeof(scannerConfig.BeamReportMask));
	SetupBeamSelectionMask(START_ANGLE, END_ANGLE, scannerConfig.BeamReportMask);
	
	local_connectionInfo.logical_attribute = OS32C_VS_BEAM_MASK_ATTRIBUTE;		// Beam Selection Mask attribute
	printf("Writing new Beam Selection Mask... \r\n");
	local_connectionInfo.data_length = sizeof(scannerConfig.BeamReportMask);
	(void *)memcpy(&local_connectionInfo.data_buf[0], &scannerConfig.BeamReportMask, sizeof(scannerConfig.BeamReportMask));  

#ifdef __USE_EIP_DLL__
	msgSize = EIP_EncapsulateTcpSendRRDataExplicitRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#else
	msgSize = Setup_SendRRDataRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#endif
	// Send the request message.
	send(tcpLocalSocket, TxBuf, msgSize, 0);
		
	// Get the reply message from the server/receiver...
	msgSize = recv(tcpLocalSocket, RxBuf, RXBUFSIZE, 0);
	if (msgSize != SOCKET_ERROR) 
	{
		(void *)memcpy(&sendRRDataReply, &RxBuf[0], sizeof(sendRRDataReply));
		if (sendRRDataReply.status == EIP_STATUS_SUCCESS)
		{
			printf("New Beam Selection Mask\r\n");
		}
		else
		{
			printf("Status Error : 0x%04X\r\n", sendRRDataReply.status);
		}
    }else{ 
        return(SOCKET_ERROR);
    }

    return(0);
}
// *********************************************************************************************************************************************

//*********************************************************************************************************************
// Initiate a TCP request/response session with the scanner and gather Range and Reflectivity measurement data.
//*********************************************************************************************************************
int Initiate_ScannerTCPMeasurements(struct S_Service_Code_Request_Format *pa_cip_request, EIP_UINT32 num_cycles)
{
    int             msgSize;
    EIP_UINT32      reflectivityMeasurementData_inPicoseconds;  
	S_Encapsulation_SendRRDataResponse      sendRRDataReply;

    unsigned int	i, totalCount, totalDataCount;
    LARGE_INTEGER	ctr1, ctr2, ctr3, freq;
	double			elapseTime, eStartTime;
	double			eTime[NUM_OF_TCP_SAMPLES];
	
#ifdef __LOG_TIMESTAMP_RESULTS__	
	unsigned int    j,k;
	ofstream		LogFile; 
	double			ePeriod, eDelta, eDeltaNeg, eDeltaPos;
	double			eTimeMin[NUM_OF_TCP_SAMPLES];
	double			eTimeMax[NUM_OF_TCP_SAMPLES];
#endif
    
    // Seutp session control structure and get session handle for this scanner connection.
    S_Service_Code_Request_Format	local_connectionInfo;
    local_connectionInfo.session_handle = pa_cip_request->session_handle;

    // *********************************************************************************************************************************************
    // TCP example :	Read the vendor specific run-time measurement data, single attribute data, from the scanner.
    //					Send x (NUM_OF_TCP_SAMPLES) measurment data requests and extract the current measurements provided in the reply message.
    //					Display the measurment results for each sample and conclude with the final throughput measurements.
    //					Note: For accurate measurements limit other computer activities during this test cycle.
    // *********************************************************************************************************************************************
	// Sends Tcp data to server/receiver...
	printf("\r\nSending TCP SendRRData UCMM Explicit Message Service Requests... \r\n");
	local_connectionInfo.service_code = CIP_SC_GET_ATTR_SINGLE;			// Get Single Attribute service code 0x0E.   

#if 1
	local_connectionInfo.logical_class = OS32C_VS_RTM_TOF_AND_TOT_DATA;	// Vendor Specific object    0x75 
	local_connectionInfo.logical_instance = OS32C_VS_INSTANCE;			// Vendor Specific instance  0x01
	local_connectionInfo.logical_attribute = OS32C_VS_OBJ_ATTRIBUTE;	// Vendor Specific assembly data attribute 0x03
#else
	local_connectionInfo.logical_class = OS32C_ASM_OBJECT_CLASS;		// Assembly Data object
	plocal_connectionInfo.logical_instance = OS32C_ASM_RTM_TOF_DATA;	// Assembly Data instance
	local_connectionInfo.logical_attribute = OS32C_ASM_OBJ_DATA;		// Assembly Data attribute
#endif

	// Initialize cycle counters
	totalCount = 0;
	totalDataCount = 0;

	// Start process timing measurements.
	QueryPerformanceFrequency((LARGE_INTEGER *)&freq);
	QueryPerformanceCounter((LARGE_INTEGER *)&ctr1);

	// Report the measurements from the front center of the scanner at target angle degrees.
	rangeDataIndex = GetBeamIndex(START_ANGLE, END_ANGLE, TARGET_ANGLE);

	eStartTime = (ctr1.QuadPart * 1.0) / freq.QuadPart;

	// Repeat the run-time measurements
	for (i=0; i< num_cycles; i++)
	{
#ifdef __USE_EIP_DLL__
		msgSize = EIP_EncapsulateTcpSendRRDataExplicitRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#else
		msgSize = Setup_SendRRDataRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#endif
        RecordEipMessageBytes("EIP_Setup_SendRRData_ReadObjectDataRequestMessage.txt", (EIP_BYTE *)&TxBuf, msgSize);		
        
        // Send the request message.
		send(tcpLocalSocket, TxBuf, msgSize, 0);
	
		// Get the reply message from the server/receiver...
		msgSize = recv(tcpLocalSocket, RxBuf, RXBUFSIZE, 0);
		if (msgSize != SOCKET_ERROR) 
		{
			QueryPerformanceCounter((LARGE_INTEGER *)&ctr3);
			eTime[i] = (ctr3.QuadPart * 1.0) / freq.QuadPart;

			// Increment process timing measurement counts.
			totalCount++;
			totalDataCount += msgSize;

		    (void *)memcpy(&sendRRDataReply, &RxBuf[0], sizeof(sendRRDataReply));
		    if (sendRRDataReply.status == EIP_STATUS_SUCCESS)
		    {
                // Collected the received data packets.
			    (void *)memcpy(&measurementReport, &RxBuf[sizeof(sendRRDataReply)], sizeof(measurementReport));
			    (void *)memcpy(&measurementData,   &RxBuf[sizeof(sendRRDataReply) + sizeof(measurementReport)], msgSize - sizeof(sendRRDataReply) - sizeof(measurementReport));
			
			    // Report the range measurements from the selected index.
			    (void *)memcpy(&rangeMeasurementData,  &measurementData[rangeDataIndex], sizeof(EIP_UINT16));

			    // Report the reflectivity measurements from the selected index plus the number of beams sent in this packet.
			    reflectivityDataIndex = rangeDataIndex + measurementReport.DataLength;
			    (void *)memcpy(&reflectivityMeasurementData,  &measurementData[reflectivityDataIndex], sizeof(EIP_UINT16));
				reflectivityMeasurementData_inPicoseconds = reflectivityMeasurementData * 4;

			    if (rangeMeasurementData >= NO_REFLECTION_INDICATOR){
				    printf("\r\nNo Range Measurement !\r\n");
			    } else if (rangeMeasurementData == BLINDED_BEAM_INDICATOR) {
				    printf("\r\nRange Measurement Invalid !\r\n");
			    } else {
				    printf("\r\nRange Measurement for beam index #%d = %d (mm) \r\n", rangeDataIndex, rangeMeasurementData);
				    printf("Reflectivity Measurement for beam index #%d = %d (ps) \r\n", reflectivityDataIndex, reflectivityMeasurementData_inPicoseconds);
			    }
		    }
		    else
		    {
			    printf("Status Error : 0x%04X\r\n", sendRRDataReply.status);
		    }
			
		}else{
            return(SOCKET_ERROR);
        }

		Wait_ms(2); // Note: the scanner needs to wait a few milliseconds between tcp/ip request/reply transactions.
	}
    RecordEipMessageBytes("EIP_Setup_SendRRData_ReadObjectDataResponseMessage.txt", (EIP_BYTE *)&RxBuf, msgSize);

    // Finish process timing measurements.
	QueryPerformanceCounter((LARGE_INTEGER *)&ctr2);
	
	// Report processing results.
	printf("\r\nNumber of beams in measurement report : %d\r\n", measurementReport.DataLength);
	printf("Received bytes of data per packet : %d\r\n", msgSize);
	printf("Total number of packets received  : %d\r\n", totalCount);
	printf("Sending time (second): %.6f\r\n", ((ctr2.QuadPart-ctr1.QuadPart)*1.0)/freq.QuadPart );
	elapseTime = ((ctr2.QuadPart - ctr1.QuadPart) * 1.0) / freq.QuadPart;
	printf("Elapse Time: %.6f seconds.\r\n", elapseTime);
	printf("Packet Timing: %.6f seconds.\r\n", elapseTime/num_cycles);
	printf("Total Data Sent(MB): %.3f\r\n", (totalDataCount*1.0)/(1024*1024));
	printf("TCP Throughput: %.3f (KBytes/sec).\r\n", totalDataCount/(1000.0*elapseTime));
	
    printf("Scan Period\t: %d (us)\r\n", Get32bit_VSAttribute(OS32C_VS_RTM_TOF_DATA, OS32C_VS_SCAN_PERIOD_ATTRIBUTE, local_connectionInfo.session_handle));
    printf("SendRRDataReply Size = %d\r\n", sizeof(sendRRDataReply));
	printf("Measurement Report Header Size = %d\r\n", sizeof(measurementReport));

#ifdef __LOG_TIMESTAMP_RESULTS__
	LogFile.open ("OS32C_DataLog.txt");
	
	eDeltaPos = 0.000001;
	eDeltaNeg = 0.100000;
	j = k = 0;
	for(i = 4; i < num_cycles; i++)
	{
		eDelta = eTime[i] - eTime[i-1];
		if (eDelta < eDeltaNeg) 
		{
			eDeltaNeg = eDelta;
			eTimeMin[j++] = eDelta;
		}
		if (eDelta > eDeltaPos)
		{
			eDeltaPos = eDelta;
			eTimeMax[k++] = eDelta;
		}
	}
	scanPeriod = Get32bit_VSAttribute(OS32C_VS_RTM_TOF_DATA, OS32C_VS_SCAN_PERIOD_ATTRIBUTE, local_connectionInfo.session_handle);
    ePeriod = (double)scanPeriod/1000000; 
	if (eDeltaNeg > eDeltaPos)
	{
		printf("maximum jitter = +/- %.6f seconds.\r\n", ePeriod- eDeltaNeg);
	}
	else
	{
		printf("maximum jitter = +/- %.6f seconds.\r\n", eDeltaPos - ePeriod);
	}
	// Show the top jitter readings, discard the first since it may be the very first reading.
	printf("\r\n");
	for (i=1; i<j; i++)
	{
		if (eTimeMin[i] < ePeriod)
		printf("maximum negative jitter = +/- %.6f seconds.\r\n", ePeriod - eTimeMin[i]);
	}
	printf("\r\n");
	for (i=1; i<k; i++)
	{
		if (eTimeMax[i] > ePeriod)
		printf("maximum positive jitter = +/- %.6f seconds.\r\n", eTimeMax[i] - ePeriod);
	}

	// write data to a text file to be imported to an excel spreadsheet.
	LogFile << "\r\nSample\t TimeStamp (s)\t TS delta (ms)\t Deviation from Nominal Scan Period (ms)\t\r\n";
	for(i = 1; i < num_cycles; i++)
	{
		eDelta = eTime[i] -  eStartTime ; 
		LogFile << "timeStamp[" << i << "]\t" << eDelta << "\t" << (eTime[i] - eTime[i-1]) * 1000  << "\t" <<  (eTime[i] - eTime[i-1] - ePeriod) * 1000  << "\t\r\n";
	}
	LogFile.close();
#endif // #ifdef __LOG_TIMESTAMP_RESULTS__

    return(0);
}
// *********************************************************************************************************************************************

//*********************************************************************************************************************
// Initiate a UDP I/O streaming connection with the scanner and gather the range measurement assembly data. 
// *********************************************************************************************************************************************
int Initiate_ScannerUDPIOMeasurements(struct S_Service_Code_Request_Format *pa_cip_request, EIP_UINT32 num_cycles)
{

#ifdef __ENABLE_UDP_IO_MESSAGES__
    bool			connectionOpen = false;
    int             msgSize, retCode;
    unsigned int    msgCount, totalCount;
    S_Encapsulation_SendRRDataResponse      sendRRDataReply;
    S_Service_Reply_ForwardOpen				udpIOConnectionReply;

    // Seutp session control structure and get session handle for this scanner connection.
    S_Service_Code_Request_Format	local_connectionInfo;
    local_connectionInfo.session_handle = pa_cip_request->session_handle;
    
    // Initialize the interface handle to 0, it will be reassigned by the scanner later.
    local_connectionInfo.interface_handle = 0;

    // *********************************************************************************************************************************************
    // UDP example:		Create a UDP IO connection to stream IO Assembly 102 data, range data, at a specified timer rate.  
    //
    //					Note: this function type is typically used in PLC applications but can be used in PC applications that do not require 
    //					synchronous scan period communications.
    // *********************************************************************************************************************************************

    // Sends Udp data to server/receiver...
	printf("\r\nSending UDP IO Assembly Connection Request... \r\n");
		
	// Setup streaming for Range measurment data
	local_connectionInfo.input_assembly_object = OS32C_ASM_RTM_TOF_DATA; 
	local_connectionInfo.input_assembly_size = MAX_OS32C_ASM_RTM_TOF_DATA_SIZE;

#ifdef __USE_BASIC_OUTPUT_TRIGGER__		
	local_connectionInfo.output_assembly_object = OS32C_ASM_RTM_TRIGGER_DATA;
	local_connectionInfo.output_assembly_size = OS32C_ASM_RTM_TRIGGER_DATA_SIZE;
	local_connectionInfo.data_buf[0] = 0xFE;
	local_connectionInfo.data_buf[1] = 0xED;
#else  // Send configuration output assembly
	local_connectionInfo.output_assembly_object = OS32C_ASM_RTM_CONFIG_DATA;    
	local_connectionInfo.output_assembly_size = OS32C_ASM_RTM_CONFIG_DATA_SIZE; 
	(void *)memcpy(&local_connectionInfo.data_buf[0],  &scannerConfig,  sizeof(scannerConfig));
#endif

	srand((unsigned int)3145678);
	local_connectionInfo.o_to_t_connection_id = ((0xFFFFFFFF * rand()) / RAND_MAX);
	local_connectionInfo.t_to_o_connection_id = ((0xFFFFFFFF * rand()) / RAND_MAX + 1);

	// Setup the required packet intervals using the scanners scan period. 
	// Setup packet rate for 2 times the scan period and the keep alive rate 20 times the scan period.
	scanPeriod = Get32bit_VSAttribute(OS32C_VS_RTM_TOF_DATA, OS32C_VS_SCAN_PERIOD_ATTRIBUTE, local_connectionInfo.session_handle);

    local_connectionInfo.t_to_o_rpi = scanPeriod * 4;					      // repeat packet interval in microseconds
	local_connectionInfo.o_to_t_rpi = local_connectionInfo.t_to_o_rpi * 10;   // repeat packet interval in microseconds
	local_connectionInfo.timeout = 40;                       
	local_connectionInfo.vendor_id = OSTI_VENDOR_ID;					// customer defined vendor id number.
	local_connectionInfo.connection_serial_number = 0x8765;			    // customer defined connection number.
	local_connectionInfo.originator_serial_number = 0x4321;			    // customer defined serial number.

	// Send the Large Forward Open request to start the UDP IO Connection. This command is sent via the TCP port.
#ifdef __USE_EIP_DLL__
	msgSize = EIP_EncapsulateTcpStartUdpIOConnectionRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#else
	msgSize = Setup_IOLargeForwardOpenRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#endif
    RecordEipMessageBytes("EIP_Setup_IOLargeForwardOpenRequestMessage.txt", (EIP_BYTE *)&TxBuf, msgSize);

    // Send the request message.
	send(tcpLocalSocket, TxBuf, msgSize, 0);

	// Get the reply message from the server/receiver...
	retCode = recv(tcpLocalSocket, RxBuf, RXBUFSIZE, 0);
	if (retCode == SOCKET_ERROR) 
	{
		printf("Tcp recv() error %ld.\r\n", WSAGetLastError());
 		Wait_ms(2000);
	}
	else
	{
		if (Get_IOLargeForwardOpenResponseMessage(&local_connectionInfo, &udpIOConnectionReply, RxBuf) == 0)
		{
			connectionOpen = true;
            RecordEipMessageBytes("EIP_Get_IOLargeForwardOpenResponseMessage.txt", (EIP_BYTE *)&RxBuf, retCode);

	        // Keep alive UDP IO connection by sending the output assembly to the scanner. This command is sent via the UDP IO port.
#ifdef __USE_EIP_DLL__
	        msgSize = EIP_EncapsulateUdpIOConnectionKeepAliveRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#else
	        msgSize = Setup_IOKeepAliveMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#endif
            RecordEipMessageBytes("EIP_Setup_IOKeepAliveMessage.txt", (EIP_BYTE *)&TxBuf, msgSize);

            // Send the request message.
	        sendto(udpIoLocalSocket, TxBuf, msgSize, 0, (SOCKADDR *) &udpIoServerAddr, udpIoAddrLen);

	        // Increment the O->T connection sequence number.
            local_connectionInfo.o_to_t_sequence_num++;

            // Save interface handle for future use.
            pa_cip_request->interface_handle = local_connectionInfo.interface_handle;
		}
	} // endif retCode
		
	msgCount = 0;
	totalCount = 0;
	while (connectionOpen == true)
	{
		// Get the reply message from the server/receiver...
		printf("\r\nWaiting for UDP IO Assembly Data... \r\n");
		retCode = recvfrom(udpIoLocalSocket, RxIoBuf, MAX_UDP_MTU, 0, (SOCKADDR *) &udpIoServerAddr, &udpIoAddrLen);
		if (retCode == SOCKET_ERROR) 
		{
			printf("Udp IO recv() error %ld.\r\n", WSAGetLastError());
 			Wait_ms(2000);
			connectionOpen = false;
		}
		else
		{
			// Extract the returned data from the rx packet.
			ProcessReceived_IOData((EIP_UINT8 *)&RxIoBuf, retCode, local_connectionInfo.input_assembly_object);
			printf("Received bytes of data per packet : %d\r\n", retCode);
			
			// Report the measurement from the front center of the scanner at target angle degrees.
			rangeDataIndex = GetBeamIndex(START_ANGLE, END_ANGLE, TARGET_ANGLE);
			(void *)memcpy(&rangeMeasurementData,  &measurementData[rangeDataIndex], sizeof(EIP_UINT16));

			if (rangeMeasurementData >= NO_REFLECTION_INDICATOR){
				printf("\r\nNo Range Measurement !\r\n");
			} else if (rangeMeasurementData == BLINDED_BEAM_INDICATOR) {
				printf("\r\nRange Measurement Invalid !\r\n");
			} else {
				printf("\r\nRange Measurement for beam index #%d = %d (mm) \r\n", rangeDataIndex, rangeMeasurementData);
			}

			msgCount++;
			if (totalCount >= num_cycles)
			{
				connectionOpen = false;

                // Record last received message.
                RecordEipMessageBytes("EIP_ProcessReceived_IOData.txt", (EIP_BYTE *)&RxIoBuf, retCode);

				// Terminate the UDP IO connection. This command is sent via the TCP port.
				printf("\r\nTerminating UDP IO Assembly Connection... \r\n");

#ifdef __USE_EIP_DLL__
				msgSize = EIP_EncapsulateTcpStopUdpIOConnectionRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#else
				msgSize = Setup_IOLargeForwardCloseRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#endif
                RecordEipMessageBytes("EIP_Setup_IOLargeForwardCloseRequestMessage.txt", (EIP_BYTE *)&TxBuf, msgSize);
                
                // Send the request message.
                send(tcpLocalSocket, TxBuf, msgSize, 0);

                // Get the reply message from the server/receiver...
	            retCode = recv(tcpLocalSocket, RxBuf, RXBUFSIZE, 0);
	            if (retCode != SOCKET_ERROR) 
	            {
		            (void *)memcpy(&sendRRDataReply, &RxBuf[0], sizeof(sendRRDataReply));

                    if(sendRRDataReply.service_response != CIP_SC_LRG_CLOSE_RSP)
                    {
                        printf("\r\nClose Forward Open Error : Response 0x%02X : Status 0x%02X\r\n", sendRRDataReply.service_response, sendRRDataReply.service_status);
                    }
                    RecordEipMessageBytes("EIP_Setup_IOLargeForwardCloseResponseMessage.txt", (EIP_BYTE *)&RxBuf, retCode);
	            }                

			}
			else if (msgCount == 10)
			{
				// Keep the UDP IO connection alive by sending the output assembly . This command is sent via the UDP IO port.
#ifdef __USE_EIP_DLL__
                msgSize = EIP_EncapsulateUdpIOConnectionKeepAliveRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#else
                msgSize = Setup_IOKeepAliveMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#endif
                // Send the request message.
				sendto(udpIoLocalSocket, TxBuf, msgSize, 0, (SOCKADDR *) &udpIoServerAddr, udpIoAddrLen);
					
                // Increment the O->T connection sequence number.
                local_connectionInfo.o_to_t_sequence_num++; 

				msgCount = 0;
				totalCount += 10;
			}

		} // end if socket error
	} // end of while loop

#endif // #ifdef __ENABLE_UDP_IO_MESSAGES__

    return(0);
}
// *********************************************************************************************************************************************

//*********************************************************************************************************************
// Read the vendor specific configuration attributes and the system status assembly data
// *********************************************************************************************************************************************
int Read_ScannerAttributes(struct S_Service_Code_Request_Format *pa_cip_request)
{
    int         msgSize;
   	EIP_UINT16	attribute16bitVal;
    S_Encapsulation_SendRRDataResponse      sendRRDataReply;

    // Seutp session control structure and get session handle for this scanner connection.
    S_Service_Code_Request_Format	local_connectionInfo;
    local_connectionInfo.session_handle = pa_cip_request->session_handle;
    local_connectionInfo.interface_handle = pa_cip_request->interface_handle;

    // *********************************************************************************************************************************************
    // TCP example :	Read single attributes from the scanner. 
    //					Read the range and reflectivity measurement formats currently used by the scanner.
    // *********************************************************************************************************************************************
	// Sends Tcp data to server/receiver...
	printf("\r\nSending TCP SendRRData Explicit Message Service Requests... \r\n");
	local_connectionInfo.service_code = CIP_SC_GET_ATTR_SINGLE;		    // Get Single Attribute service code   
	local_connectionInfo.logical_class = OS32C_VS_RTM_CONFIG_DATA;	    // Vendor Specific object
	local_connectionInfo.logical_instance = OS32C_VS_INSTANCE;		    // Vendor Specific instance

	// Read a single attribute, measurement range format.
	local_connectionInfo.logical_attribute = OS32C_VS_RANGE_ATTRIBUTE;

#ifdef __USE_EIP_DLL__
	msgSize = EIP_EncapsulateTcpSendRRDataExplicitRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#else
	msgSize = Setup_SendRRDataRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#endif
    RecordEipMessageBytes("EIP_Setup_SendRRData_ReadRequestMessage.txt", (EIP_BYTE *)&TxBuf, msgSize);	
    
    // Send the request message.
	send(tcpLocalSocket, TxBuf, msgSize, 0);

	// Get the reply message from the server/receiver...
	msgSize = recv(tcpLocalSocket, RxBuf, RXBUFSIZE, 0);
	if (msgSize != SOCKET_ERROR) 
	{
		(void *)memcpy(&sendRRDataReply, &RxBuf[0], sizeof(sendRRDataReply));

        attribute16bitVal = 0;
        if(sendRRDataReply.service_response == (CIP_SC_GET_ATTR_SINGLE | 0x80))
        {
		    (void *)memcpy(&attribute16bitVal, &RxBuf[sizeof(sendRRDataReply)], sizeof(attribute16bitVal));
            printf("\r\nRange Format = %d\r\n", attribute16bitVal);
        }else{
            printf("\r\nGet Attribute Error : Response 0x%02X : Status 0x%02X\r\n", sendRRDataReply.service_response, sendRRDataReply.service_status);
            return(0);
        }
        RecordEipMessageBytes("EIP_Setup_SendRRData_ReadResponseMessage.txt", (EIP_BYTE *)&RxBuf, msgSize);	
	}

	// Read a single attribute, measurment reflectivity format.
	local_connectionInfo.logical_attribute = OS32C_VS_REFLECTIVITY_ATTRIBUTE;

#ifdef __USE_EIP_DLL__
	msgSize = EIP_EncapsulateTcpSendRRDataExplicitRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#else
	msgSize = Setup_SendRRDataRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#endif
	// Send the request message.
	send(tcpLocalSocket, TxBuf, msgSize, 0);

	// Get the reply message from the server/receiver...
	msgSize = recv(tcpLocalSocket, RxBuf, RXBUFSIZE, 0);
	if (msgSize != SOCKET_ERROR) 
	{
		(void *)memcpy(&sendRRDataReply, &RxBuf[0], sizeof(sendRRDataReply));
		(void *)memcpy(&attribute16bitVal, &RxBuf[sizeof(sendRRDataReply)], sizeof(attribute16bitVal));
		printf("Reflectivity Format = %d\r\n", attribute16bitVal);
	}
	Wait_ms(1000);

    // *********************************************************************************************************************************************
    // TCP example :	Read the system status assembly object data from the scanner using the explicit get single attribute command.
    //					Display the system status data using the structure members provided in the systemReport.
    // *********************************************************************************************************************************************
	// Sends Tcp data to server/receiver...
	printf("\r\nSending TCP UCMM Explicit Message Service Requests... \r\n");
	local_connectionInfo.service_code = CIP_SC_GET_ATTR_SINGLE;			// Get Single Attribute service code.   
	local_connectionInfo.logical_class = OS32C_ASM_OBJECT_CLASS;			// Assembly object class.
	local_connectionInfo.logical_instance = OS32C_ASM_SYS_STATUS_DATA;  	// Assembly object #100
	local_connectionInfo.logical_attribute = OS32C_ASM_OBJ_DATA;			// Assembly data attribute

#ifdef __USE_EIP_DLL__
	msgSize = EIP_EncapsulateTcpSendRRDataExplicitRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#else
	msgSize = Setup_SendRRDataRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#endif
	// Send the request message.
	send(tcpLocalSocket, TxBuf, msgSize, 0);

	// Get the reply message from the server/receiver...
	msgSize = recv(tcpLocalSocket, RxBuf, RXBUFSIZE, 0);
	if (msgSize != SOCKET_ERROR) 
	{
		// Collected the received data packets.
		(void *)memcpy(&sendRRDataReply,   &RxBuf[0], sizeof(sendRRDataReply));
		(void *)memcpy(&systemReport, &RxBuf[sizeof(sendRRDataReply)], sizeof(systemReport));

		switch (systemReport.MachineState)
		{
			case POST:
					printf("System State\t: POWER-ON-SELF-TEST\r\n");
					break;
			case STOP:
					printf("System State\t: STOP\r\n");
					break;
			case INTERLOCK:
					printf("System State\t: INTERLOCK\r\n");
					break;
			case RUN:
					printf("System State\t: RUN\r\n");
					break;
			case STANDBY:
					printf("System State\t: STANDBY\r\n");
					break;
			case DEEPSTANDBY:
					printf("System State\t: DEEPSTANDBY\r\n");
					break;
			case CONFIGURE:
					printf("System State\t: CONFIGURE\r\n");
					break;
			case FAULT:
					printf("System State\t: FAULT\r\n");
					break;
			default:
					printf("System State\t: %d\r\n", systemReport.MachineState);
					break;
		};
		printf("Active Zone\t: %d\r\n\r\n", systemReport.ActiveZoneSet);
	
		// Report the current detetion zone status
		if (systemReport.DetectionZoneStatus & EIP_PROTECTION_ZONE_VIOLATION)
		{
			printf("Protection Zone Violated ! \r\n");
		}
		if (systemReport.DetectionZoneStatus & EIP_WARNING_ZONE_1_VIOLATION)
		{
			printf("Warning Zone 1 Violated ! \r\n");
		}
		if (systemReport.DetectionZoneStatus & EIP_WARNING_ZONE_2_VIOLATION)
		{
			printf("Warning Zone 2 Violated ! \r\n");
		}
		if (systemReport.DetectionZoneStatus & EIP_DUST_DETETION_WARNING)
		{
			printf("Dust Detection Warning... clean the scanner window ! \r\n");
		}
		if (systemReport.DetectionZoneStatus == DEF_BIT_NONE)
		{
			printf("Detections Zones Clear !\r\n");
		}
	}  //endif (msgSize != SOCKET_ERROR) 
	
	Wait_ms(1000);

    // *********************************************************************************************************************************************
    // TCP example : Read individual scanner attributes using the Get32bit_VSAttribute() & Get16bit_VSAttribute() functions.
    // *********************************************************************************************************************************************
	printf("\r\n");
	// Get data from vendor specific assembly 114
	printf("Scan Period\t: %d (us)\r\n", Get32bit_VSAttribute(OS32C_VS_RTM_TOF_DATA, OS32C_VS_SCAN_PERIOD_ATTRIBUTE, local_connectionInfo.session_handle));
	printf("Beam Period\t: %d (ns)\r\n", Get32bit_VSAttribute(OS32C_VS_RTM_TOF_DATA, OS32C_VS_BEAM_PERIOD_ATTRIBUTE, local_connectionInfo.session_handle));
	printf("Time Stamp\t: %u (us)\r\n", Get32bit_VSAttribute(OS32C_VS_RTM_TOF_DATA, OS32C_VS_TIMESTAMP_ATTRIBUTE, local_connectionInfo.session_handle));
	printf("Beam Data Count\t: %d \r\n", Get16bit_VSAttribute(OS32C_VS_RTM_TOF_DATA, OS32C_VS_BEAM_DATA_LENGTH_ATTRIBUTE, local_connectionInfo.session_handle));

	// Get data from vendor specific assembly 112
	printf("Machine State\t: %d \r\n", Get16bit_VSAttribute(OS32C_VS_SYSTEM_STATUS_DATA, OS32C_VS_MACHINE_STATE_ATTRIBUTE, local_connectionInfo.session_handle));
	printf("Active Zone Set\t: %d \r\n", Get16bit_VSAttribute(OS32C_VS_SYSTEM_STATUS_DATA, OS32C_VS_ACTIVE_ZONE_SET_ATTRIBUTE, local_connectionInfo.session_handle));
	printf("Zone Inputs\t: 0x%02X \r\n", Get16bit_VSAttribute(OS32C_VS_SYSTEM_STATUS_DATA, OS32C_VS_ZONE_INPUT_COMBO_ATTRIBUTE, local_connectionInfo.session_handle));
	printf("Non-Safety CRC\t: 0x%02X \r\n", Get16bit_VSAttribute(OS32C_VS_SYSTEM_STATUS_DATA, OS32C_VS_NONSAFETY_CRC_ATRIBUTE, local_connectionInfo.session_handle));
	printf("Safety CRC\t: 0x%02X \r\n", Get16bit_VSAttribute(OS32C_VS_SYSTEM_STATUS_DATA, OS32C_VS_SAFETY_CRC_ATRIBUTE, local_connectionInfo.session_handle));

    return(0);
}
//*********************************************************************************************************************

//*********************************************************************************************************************
// Keep the TCP connection alive during quite periods of 10 seconds or more.
// *********************************************************************************************************************************************
int KeepAliveTCPconnection(struct S_Service_Code_Request_Format *pa_cip_request, EIP_UINT32 num_cycles)
{    
    int msgSize, retCode;
    EIP_UINT32 i;

    // Seutp session control structure and get session handle for this scanner connection.
    S_Service_Code_Request_Format	local_connectionInfo;
    local_connectionInfo.session_handle = pa_cip_request->session_handle;

    // *********************************************************************************************************************************************
    // TCP example : TCP connection keep alive (EtherNet/IP NOP command).
    //					Send NOP commands to the scanner to keep the connection open for more than 10 minutes when there is no current activity.
    // *********************************************************************************************************************************************

	printf("\r\n");
	for(i=0; i < num_cycles; i++)
	{
		printf("Sending TCP KeepAlive Request, EtherNet/IP NOP command... \r\n");
#ifdef __USE_EIP_DLL__
		msgSize = EIP_EncapsulateTcpConnectionKeepAliveRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#else
		msgSize = Setup_NopRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#endif
        // Send the request message.
		retCode = send(tcpLocalSocket, TxBuf, msgSize, 0);
        if(retCode == SOCKET_ERROR)
        {
            return(retCode);
        }
		Wait_ms(1000);
	}
    RecordEipMessageBytes("EIP_Setup_NopRequestMessage.txt", (EIP_BYTE *)&TxBuf, msgSize);	

    // Note: There is no reply to the EtherNet/IP NOP command.
    return(0);
}
// *********************************************************************************************************************************************

//*********************************************************************************************************************
// Close all of the user sockets before terminating the application.
// *********************************************************************************************************************************************
void Close_UserSockets(void)
{
    // Wait to close sockets.
    Wait_ms(1000);

    // Shutdown the Tcp socket.
    if (shutdown(tcpLocalSocket, SD_SEND) != 0)
	{
        printf("Error in Tcp shutdown(). Error code: %ld\r\n", WSAGetLastError());
	}
	// Close the Tcp socket.
    if (closesocket(tcpLocalSocket) != 0)
	{     
		printf("Error in Tcp closesocket(). Error code: %ld\r\n", WSAGetLastError());
	}

    // Shutdown the Udp socket
    if (shutdown(udpLocalSocket, SD_SEND) != 0)
	{
        printf("Error in Udp shutdown(). Error code: %ld\r\n", WSAGetLastError());
	}

    // Close the Udp socket.
    if (closesocket(udpLocalSocket) != 0)
	{     
		printf("\r\nError in Udp closesocket(). Error code: %ld\r\n", WSAGetLastError());
	}

    // Shutdown the Udp I/O socket
    if (shutdown(udpIoLocalSocket, SD_SEND) != 0)
	{
        printf("Error in Udp shutdown(). Error code: %ld\r\n", WSAGetLastError());
	}

    // Close the Udp I/O socket.
    if (closesocket(udpIoLocalSocket) != 0)
	{     
		printf("\r\nError in Udp closesocket(). Error code: %ld\r\n", WSAGetLastError());
	}

    	// When your application is finished handling the connection, call WSACleanup.
    if (WSACleanup() != 0)
	{
		printf("WSACleanup() failed!...\r\n");
	}

	printf("\r\nExisting Application!...please wait\r\n\r\n");
	system("pause");

}
//*********************************************************************************************************************

//*********************************************************************************************************************
// Initialize all of the user sockets for this application.
//*********************************************************************************************************************
int Initialize_UserSockets(void)
{
	int retCode;

	// Initialize Winsock version 2.2
    WSAStartup(MAKEWORD(2,2), &wsaData);
    printf("\r\nWinsock DLL status is %s.\r\n", wsaData.szSystemStatus);

	// Configure Tcp socket type
    tcpLocalSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (tcpLocalSocket == INVALID_SOCKET)
    {
         printf("\r\nTcp socket() failed! Error code: %ld\r\n", WSAGetLastError());
         WSACleanup();
         return (SOCKET_ERROR);
    }

	// Bind the socket to the local address and specified port.
	tcpLocalAddr.sin_family = AF_INET;
    tcpLocalAddr.sin_port = htons(os32c_TcpPort);
    tcpLocalAddr.sin_addr.s_addr = inet_addr(local_IpAddr);
	if (bind(tcpLocalSocket, (SOCKADDR *)&tcpLocalAddr, sizeof(tcpLocalAddr)) == SOCKET_ERROR)
	{
		if (WSAGetLastError() != WSAEADDRINUSE)
		{
			printf("\r\nTCP socket bind() failed! Error code: %ld\r\n", WSAGetLastError());
			closesocket(tcpLocalSocket);
			WSACleanup();
			return (SOCKET_ERROR);
		} else {
			printf("\r\nUsing previously connected TCP socket...\r\n");
		}
	}

	// Setup the remote or target address and specified port.
	tcpServerAddr.sin_family = AF_INET;
    tcpServerAddr.sin_port = htons(os32c_UdpPort);
    tcpServerAddr.sin_addr.s_addr = inet_addr(os32c_IpAddr);

    // Make a Tcp connection to the os32c(server) with socket tcpLocalSocket.
    retCode = connect(tcpLocalSocket, (SOCKADDR *) &tcpServerAddr, sizeof(tcpServerAddr));
    if(retCode != 0)
    {
         printf("\r\nTcp connect() failed! Error code: %ld\r\n", WSAGetLastError());
		 printf("\r\nAttempted to connected to Tcp Server using (IP Address : %s, Port # : %d)\r\n", inet_ntoa(tcpServerAddr.sin_addr), htons(tcpServerAddr.sin_port));
		 closesocket(tcpLocalSocket);
         WSACleanup();
		 return (SOCKET_ERROR);
    }

    // Get some information from the server side...
    getsockname(tcpLocalSocket, (SOCKADDR *)&tcpServerAddr, (int *)sizeof(tcpServerAddr));
	printf("\r\nConnected to Tcp Server (IP Address : %s, Port # : %d)\r\n", inet_ntoa(tcpServerAddr.sin_addr), htons(tcpServerAddr.sin_port));

	// Configure Udp socket type
    udpLocalSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (udpLocalSocket == INVALID_SOCKET)
    {
         printf("\r\nUDP socket() failed! Error code: %ld\r\n", WSAGetLastError());
   		 closesocket(tcpLocalSocket);
		 WSACleanup();
		 return (SOCKET_ERROR);
    }
	
	// Bind the socket to the local address and specified port.
	udpLocalAddr.sin_family = AF_INET;
    udpLocalAddr.sin_port = htons(os32c_UdpPort);
    udpLocalAddr.sin_addr.s_addr = inet_addr(local_IpAddr);

	// Setup the remote or target address and specified port.
	udpServerAddr.sin_family = AF_INET;
    udpServerAddr.sin_port = htons(os32c_UdpPort);
    udpServerAddr.sin_addr.s_addr = inet_addr(os32c_IpAddr);

	if (bind(udpLocalSocket, (SOCKADDR *)&udpLocalAddr, sizeof(udpLocalAddr)) == SOCKET_ERROR)
	{
		printf("\r\nUDP socket bind() failed! Error code: %ld\r\n", WSAGetLastError());
		closesocket(udpLocalSocket);
		shutdown(tcpLocalSocket, SD_SEND);
		closesocket(tcpLocalSocket);
		WSACleanup();
		return (SOCKET_ERROR);
	}
    else
    {
        printf("UDP socket() opened,  port : %d \r\n", os32c_UdpPort);
    }

	// Configure Udp IO socket type
    udpIoLocalSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (udpIoLocalSocket == INVALID_SOCKET)
    {
        printf("\r\nUDP IO socket() failed! Error code: %ld\r\n", WSAGetLastError());
	}
	else
	{
		printf("UDP IO socket() opened,  port : %d \r\n", os32c_UdpIoPort);

		// Bind the socket to the local address and specified port.
		udpIoLocalAddr.sin_family = AF_INET;
		udpIoLocalAddr.sin_port = htons(os32c_UdpIoPort);
		udpIoLocalAddr.sin_addr.s_addr = inet_addr(local_IpAddr);

		// Setup the remote or target address and specified port.
		udpIoServerAddr.sin_family = AF_INET;
		udpIoServerAddr.sin_port = htons(os32c_UdpPort);
		udpIoServerAddr.sin_addr.s_addr = inet_addr(os32c_IpAddr);

		if (bind(udpIoLocalSocket, (SOCKADDR *)&udpIoLocalAddr, sizeof(udpIoLocalAddr)) == SOCKET_ERROR)
		{
			printf("\r\nUDP IO socket bind() failed! Error code: %ld\r\n", WSAGetLastError());
            Close_UserSockets();
			return (SOCKET_ERROR);
		}
    }
	return (0);
}

// *********************************************************************************************************************************************
// Helper functions
// *********************************************************************************************************************************************

//******************************************************************************
// Wait_ms()
// ******************************************************************************
void Wait_ms(unsigned int mseconds)
{
    clock_t goal = mseconds + clock();
    while (goal > clock());
}

/******************************************************************************
**  Function/Task: SetupBeamSelectionMask
**
**  Purpose:    Select the beams used within the measurement data reports.
**
**  Arguments:  float  startAngle (-0.4 to 270.4 degrees)
**				float  startEnd  (-0.4 to 270.4 degrees)
**              uint16 *BeamSelectionData (pointer to Beam Selection Array)
**
**  Return:     bool TRUE/FALSE, PASS/FAIL 
**
**  Notes:     
**
*******************************************************************************/
bool SetupBeamSelectionMask(float startAngle, float endAngle, unsigned int *beamSelectionData)
{
	signed short i;
    signed short startBeam = (signed short)((startAngle * 10) / 4);       // startAngle/0.4 degrees per beam.
    signed short endBeam   = (signed short)(((endAngle * 10) + 2) / 4);   // endAngle/0.4 degrees per beam.
	unsigned int beamSelectionMask = 0x00000001;
	unsigned int beamSelctionCtr   = 0;

	if ((startBeam < -1) || (startBeam > 676) ||
	    (  endBeam < -1) || (endBeam   > 676) ||
		(startBeam > endBeam))
	{
	  	return(false);
	} 
    
    // setup initial mask variables.
	for (i = -1; i < startBeam; i++)
	{  
		if (beamSelectionMask == 0x80000000)
		{
			beamSelctionCtr++;
			beamSelectionMask = 0x00000001;
		}
		else
		{
			beamSelectionMask = beamSelectionMask << 1;
		}
	}

	// setup required measurement beams.
	for (i = startBeam; i < endBeam; i++)
	{  
		beamSelectionData[beamSelctionCtr] |= beamSelectionMask;
		if (beamSelectionMask == 0x80000000)
		{
			beamSelctionCtr++;
			beamSelectionMask = 0x00000001;
		}
		else
		{
			beamSelectionMask = beamSelectionMask << 1;
		}
	}
    return(true);
}	

/******************************************************************************
**  Function/Task: GetBeamIndex
**
**  Purpose:    Select the beams used within the measurement data reports.
**
**  Arguments:  float  startAngle  (-0.4 to 270.4 degrees)
**				float  startEnd    (-0.4 to 270.4 degrees)
**				float  targetAngle (-0.4 to 270.4 degrees)
**
**  Return:     unsigned short Array Index Value 
**
**  Notes:     
**
*******************************************************************************/
unsigned short GetBeamIndex(float startAngle, float endAngle, float targetAngle)
{
	signed short i,j;
    signed short startBeam	= (signed short)(( startAngle * 10) / 4);        // startAngle/0.4 degrees per beam.
    signed short endBeam    = (signed short)(((endAngle * 10) + 2) / 4);     // endAngle/0.4 degrees per beam.
    signed short targetBeam	= (signed short)(((targetAngle * 10)) / 4);      // targetAngle/0.4 degrees per beam.


	// get beam index
	j=0;
	for (i = startBeam; i < endBeam; i++)
	{  
		if (i == targetBeam)
		{
			break;
		}
		j++;
	}
	return (j);
}

// *********************************************************************************************************************************************
// TCP example : for reading the vendor specific 16bit attributes from the scanner.
// *********************************************************************************************************************************************
EIP_UINT16 Get16bit_VSAttribute(EIP_UINT8 vsObject, EIP_UINT8 vsAttribute, EIP_UINT32 session_handle)
{
	EIP_UINT32	msgSize;
	EIP_UINT16	retVal = 0xFFFF;
    
    // Seutp session control structure and get session handle for this scanner connection.
    S_Service_Code_Request_Format	local_connectionInfo;
    local_connectionInfo.session_handle = session_handle;

	local_connectionInfo.service_code = CIP_SC_GET_ATTR_SINGLE;		// Get Single Attribute service code.   
	local_connectionInfo.logical_class = vsObject;					// Vendor Specific object
	local_connectionInfo.logical_instance = OS32C_VS_INSTANCE;		// Vendor Specific instance
	local_connectionInfo.logical_attribute = vsAttribute;			// Vendor Specific attribute

#ifdef __USE_EIP_DLL__
	msgSize = EIP_EncapsulateTcpSendRRDataExplicitRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#else
	msgSize = Setup_SendRRDataRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#endif
	// Send the request message.
	send(tcpLocalSocket, TxBuf, msgSize, 0);

	// Get the reply message from the server/receiver...
	if (recv(tcpLocalSocket, RxBuf, RXBUFSIZE, 0) != SOCKET_ERROR) 
	{
		// Collected the received data packets.
		(void *)memcpy(&retVal, &RxBuf[sizeof(S_Encapsulation_UCMMResponse)], sizeof(retVal));
	}
	return (retVal);
}

// *********************************************************************************************************************************************
// TCP example : for reading the vendor specific 32bit attributes from the scanner.
// *********************************************************************************************************************************************
EIP_UINT32 Get32bit_VSAttribute(EIP_UINT8 vsObject, EIP_UINT8 vsAttribute, EIP_UINT32 session_handle)
{
	EIP_UINT32	msgSize;
	EIP_UINT32	retVal = 0xFFFFFFFF;

    // Seutp session control structure and get session handle for this scanner connection.
    S_Service_Code_Request_Format	local_connectionInfo;
    local_connectionInfo.session_handle = session_handle;

	local_connectionInfo.service_code = CIP_SC_GET_ATTR_SINGLE;		// Get Single Attribute service code.   
	local_connectionInfo.logical_class = vsObject;					// Vendor Specific object
	local_connectionInfo.logical_instance = OS32C_VS_INSTANCE;		// Vendor Specific instance
	local_connectionInfo.logical_attribute = vsAttribute;			// Vendor Specific attribute

#ifdef __USE_EIP_DLL__
	msgSize = EIP_EncapsulateTcpSendRRDataExplicitRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#else
	msgSize = Setup_SendRRDataRequestMessage((EIP_BYTE *)&TxBuf, &local_connectionInfo);
#endif
	// Send the request message.
	send(tcpLocalSocket, TxBuf, msgSize, 0);

	// Get the reply message from the server/receiver...
	if (recv(tcpLocalSocket, RxBuf, RXBUFSIZE, 0) != SOCKET_ERROR) 
	{
		// Collected the received data packets.
		(void *)memcpy(&retVal, &RxBuf[sizeof(S_Encapsulation_UCMMResponse)], sizeof(retVal));
	}
	return (retVal);
}

// *********************************************************************************************************************************************
// UDP example :	Read IO connection response from the scanner.
//					Save the interface handle, connection IDs and connection serial number to be used in the IO keep alive message.
// *********************************************************************************************************************************************
EIP_UINT8 Get_IOLargeForwardOpenResponseMessage(struct S_Service_Code_Request_Format *pa_cip_request, struct S_Service_Reply_ForwardOpen *reply, char *buffer)
{
			// Get the response to the connection request.
			(void *)memcpy(reply, buffer, sizeof(struct S_Service_Reply_ForwardOpen));

			// -----------------------------------------------------------------------------------
			// Save the following connection information to be used to keep the connection alive.
			pa_cip_request->interface_handle = reply->interface_handle;
			pa_cip_request->o_to_t_connection_id = reply->o_to_t_net_conn_id;
			pa_cip_request->t_to_o_connection_id = reply->t_to_o_net_conn_id;
			pa_cip_request->connection_serial_number = reply->connection_serial_number;
			// -----------------------------------------------------------------------------------

#ifdef __ENABLE_TRACE_DEBUG__
			printf("\r\n*** SendRRData Attributes ***\r\n");
			printf("command\t\t\t: 0x%02X\r\n", reply->command);
			printf("length\t\t\t: %d \r\n", reply->length);
			printf("session_handle\t\t: %d \r\n",reply->session_handle);
			printf("status\t\t\t: %d \r\n", reply->status);
			for(EIP_INT8 i=0; i<8; i++)
			{
				printf("sender_context data[%d]\t: %d \r\n", i, reply->sender_context_data[i]);
			}
			printf("options\t\t\t: %d \r\n", reply->options);
			printf("interface_handle\t: %d \r\n", reply->interface_handle);
			printf("timeout\t\t\t: %d \r\n", reply->timeout);
			printf("item_count\t\t: %d \r\n", reply->item_count);
			printf("address_item.type_id\t: 0x%02X \r\n", reply->address_item.type_id);
			printf("address_item.length\t: %d \r\n", reply->address_item.length);
			printf("data_item.type_id\t: 0x%02X \r\n", reply->data_item.type_id);
			printf("data_item.length\t: %d \r\n", reply->data_item.length);
			printf("service_response\t: 0x%02X \r\n", reply->service_response);
			printf("service_status\t\t: 0x%02X \r\n", reply->service_status);
			printf("additional_status_size\t: 0x%02X\r\n", reply->additional_status_size);
			if (reply->service_status != 0)
			{
				printf("additional_status\t: 0x%04X\r\n", reply->o_to_t_net_conn_id); // contains error code information 
			}
			else
			{
				printf("O->T Connection ID\t: 0x%08X\r\n", reply->o_to_t_net_conn_id);
				printf("T->O Connection ID\t: 0x%08X\r\n", reply->t_to_o_net_conn_id);
				printf("Connection Serial Number: %d\r\n", reply->connection_serial_number);
				printf("Originator Vendor ID\t: %d\r\n", reply->vendor_id);
				printf("Originator Serial Number: %d\r\n", reply->originator_serial_number);
				printf("O->T API\t\t: %d\r\n", reply->o_to_t_api);
				printf("T->O API\t\t: %d\r\n", reply->t_to_o_api);
				printf("Application Reply Size\t: %d \r\n", reply->app_reply_size);
			} // endif reply->service_status
#endif // #ifdef __ENABLE_TRACE_DEBUG__

	return(reply->service_status);
}

/*=======================================================================
Function:   GetLittleEndian16
Parameters: pointer where the Little-Endian data is stored
Returns:    16 bit number

This function returns the value (in Little-Endian) of the passed pointer.
=========================================================================*/
EIP_UINT16 GetLittleEndian16 (EIP_UINT8 *pos)
{
	EIP_UINT16 return_val;

#ifdef SWAP_ENDIAN
	return_val = pos[0];
	return_val = (EIP_UINT16)((return_val << 8) | pos[1]);
#else
	return_val = pos[1];
	return_val = (EIP_UINT16)((return_val << 8) | pos[0]);
#endif

	return(return_val);
}

/*=======================================================================
Function:   GetLittleEndian32
Parameters: pointer where the Little-Endian data is stored
Returns:    32 bit number

This function returns the value (in Little-Endian) of the passed pointer.
=========================================================================*/
EIP_UINT32 GetLittleEndian32 (EIP_UINT8 *pos)
{
	EIP_UINT32 return_val;

#ifdef SWAP_ENDIAN
	return_val = pos[0];
	return_val = (return_val << 8) | pos[1];
	return_val = (return_val << 8) | pos[2];
	return_val = (return_val << 8) | pos[3];
#else
	return_val = pos[3];
	return_val = (return_val << 8) | pos[2];
	return_val = (return_val << 8) | pos[1];
	return_val = (return_val << 8) | pos[0];
#endif

	return(return_val);
}
//=======================================================================

// *********************************************************************************************************************************************
// OS32C IO Data example :	Extract the IO connection data from the scanners response packet.
// *********************************************************************************************************************************************
void ProcessReceived_IOData(EIP_UINT8 * ioData, EIP_UINT16 dataLen, EIP_UINT8 inputAssembly)
{
   /*
	  ProcessReceived_IOData received message
   */
	S_Connected_UDP_IO_Packet   ioMessage;
    EIP_UINT16 buf_ix = 0;
	EIP_UINT16 i, idx;

	ioMessage.item_count = GetLittleEndian16(&ioData[buf_ix]);
	buf_ix += sizeof(UINT16);

	ioMessage.address_item.type_id = GetLittleEndian16(&ioData[buf_ix]);
	buf_ix += sizeof(UINT16);

	ioMessage.address_item.length = GetLittleEndian16(&ioData[buf_ix]);
	buf_ix += sizeof(UINT16);

	ioMessage.connection_id = GetLittleEndian32(&ioData[buf_ix]);
	buf_ix += sizeof(UINT32);

	ioMessage.sequence_id = GetLittleEndian32(&ioData[buf_ix]);
	buf_ix += sizeof(UINT32);

	ioMessage.data_item.type_id = GetLittleEndian16(&ioData[buf_ix]);
	buf_ix += sizeof(UINT16);

	ioMessage.data_item.length = GetLittleEndian16(&ioData[buf_ix]);
	buf_ix += sizeof(UINT16);

	ioMessage.sequence_number = GetLittleEndian16(&ioData[buf_ix]);
	buf_ix += sizeof(UINT16);

	for (i = 0; i < ioMessage.data_item.length - sizeof(UINT16); i++)
	{
	  ioMessage.data_buf[i] = ioData[buf_ix];
	  buf_ix++;
	}

	switch (inputAssembly)
	{
		case OS32C_ASM_SYS_STATUS_DATA:
			// Copy the current data to the system status structure.
			(void *)memcpy(&systemReport, &ioMessage.data_buf[0], sizeof(systemReport));
			break;

		case OS32C_ASM_SYS_DETECT_STATUS_DATA:
			// Copy the current data to the system status structure.
			(void *)memcpy(&systemReport, &ioMessage.data_buf[0], sizeof(systemReport));
			idx = sizeof(systemReport);

			// Copy additional data to the protection zone data.
			(void *)memcpy(&protectionZoneData, &ioMessage.data_buf[idx], sizeof(protectionZoneData));
			idx += sizeof(protectionZoneData);
		
			// Copy additional data to the warning zone data.
			(void *)memcpy(&warningZone1Data, &ioMessage.data_buf[idx], sizeof(warningZone1Data));
			idx += sizeof(warningZone1Data);

			// Copy remaining data to the warning zone data.
			(void *)memcpy(&warningZone2Data, &ioMessage.data_buf[idx], sizeof(warningZone2Data));

			break;
		
		case OS32C_ASM_RTM_TOF_DATA:
		case OS32C_ASM_RTM_TOT_DATA:
			// Copy the current data to the system status structure.
			(void *)memcpy(&measurementReport, &ioMessage.data_buf[0], sizeof(measurementReport));
			idx = sizeof(measurementReport);

			// Copy additional data to the protection zone data.
			(void *)memcpy(&measurementData, &ioMessage.data_buf[idx], sizeof(measurementData));

			break;

		default:
			break;
	}

}
// *********************************************************************************************************************************************

// *********************************************************************************************************************************************
void RecordEipMessageBytes (string fileName, EIP_BYTE *pa_CommBuf, unsigned int msgSize)
{
  unsigned int i, n;
  ofstream LogFile;
  char buffer [16];

  LogFile.open (fileName);
  LogFile << "Message Size: " << msgSize << "\n";
  for (i=0; i < (msgSize); i++)
  {  
	  n = sprintf_s (buffer, "%02X ", pa_CommBuf[i]);
	  if (n > 3) 
	  {
		  // Remove leading FF characters.
		  buffer[0] = buffer[n-3];
		  buffer[1] = buffer[n-2];
		  buffer[2] = buffer[n-1];
		  buffer[3] = buffer[n];
	  }	
	  LogFile << buffer << " ";
  }
  LogFile.close();
}
// *********************************************************************************************************************************************

/* little-endian-to-host unsigned 16 bit*/

EIP_UINT16
ltohs(EIP_UINT8 ** pa_buf)
{
  EIP_BYTE *p = (EIP_BYTE *) *pa_buf;
  EIP_UINT16 data = p[0] | p[1] << 8;
  *pa_buf += 2;
  return data;
}

/*   UINT32 ltohl(INT8 **pa_buf)
 *   reads UINT32 from *pa_buf and converts little endian to host.
 *      *pa_buf pointer where data should be reed. 
 *  return value
 */EIP_UINT32
ltohl(EIP_UINT8 ** pa_buf)
{
  EIP_BYTE *p = (EIP_BYTE *) *pa_buf;
  EIP_UINT32 data = p[0] | p[1] << 8 | p[2] << 16 | p[3] << 24;
  *pa_buf += 4;
  return data;
}

/*   void htols(UINT16 data, UINT8 **pa_buf)
 *   converts UINT16 data from host to little endian an writes it to pa_buf.
 *      data value to be written
 *      *pa_buf pointer where data should be written.
 */
void
htols(EIP_UINT16 data, EIP_UINT8 ** pa_buf)
{
  EIP_BYTE *p = (EIP_BYTE *) *pa_buf;

  p[0] = (EIP_BYTE) data;
  p[1] = (EIP_BYTE) (data >> 8);
  *pa_buf += 2;
}

/*   void htoll(UINT32 data, INT8 **pa_buf)
 *   converts UINT32 data from host to little endian an writes it to pa_buf.
 *      data value to be written
 *      *pa_buf pointer where data should be written.
 */
void
htoll(EIP_UINT32 data, EIP_UINT8 ** pa_buf)
{
  EIP_BYTE *p = (EIP_BYTE *) *pa_buf;

  p[0] = (EIP_BYTE) data;
  p[1] = (EIP_BYTE) (data >> 8);
  p[2] = (EIP_BYTE) (data >> 16);
  p[3] = (EIP_BYTE) (data >> 24);
  *pa_buf += 4;
}

// *********************************************************************************************************************************************
EIP_INT32 Setup_ListIdentityRequestMessage(EIP_BYTE *pa_pacCommBuf, EIP_UINT16 MaxResponseDelay)
{
  /*
	  ListIdentity command request
  */

  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;

  htols(COMMAND_LISTIDENTITY, &pacCommBufRunner); /* uint16 command COMMAND_LISTIDENTITY = 0x63 */
  htols(0, &pacCommBufRunner);                    /* uint16 length 			*/
  htoll(0, &pacCommBufRunner);                    /* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                    /* uint32 status 			*/
  htols(MaxResponseDelay, &pacCommBufRunner);     /* uint32 sender context 	*/
  htols(0, &pacCommBufRunner);                    /* uint8 data[6] set to 0	*/
  htols(0, &pacCommBufRunner);
  htols(0, &pacCommBufRunner);
  htoll(0, &pacCommBufRunner);                    /* uint32 options */

  return (pacCommBufRunner - pa_pacCommBuf);
}
// *********************************************************************************************************************************************

// *********************************************************************************************************************************************
EIP_INT32 Setup_RegisterSessionRequestMessage(EIP_BYTE *pa_pacCommBuf)
{
  /*
	  RegisterSession command request
  */
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;

  htols(COMMAND_REGISTERSESSION,&pacCommBufRunner);						/* uint16 command COMMAND_REGISTERSESSION = 0x65 */
  htols(ENCAPSULATION_HEADER_SESSION_HANDLE_POS, &pacCommBufRunner); 	/* uint16 length ENCAPSULATION_HEADER_SESSION_HANDLE_POS = 4 */
  htoll(0, &pacCommBufRunner);                     						/* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                     						/* uint32 status 			*/
  htols(0, &pacCommBufRunner);     				   						/* uint32 sender context 	*/
  htols(0, &pacCommBufRunner);                     						/* uint8 data[6] set to 0	*/
  htols(0, &pacCommBufRunner);
  htols(0, &pacCommBufRunner);
  htoll(0, &pacCommBufRunner);                     						/* uint32 options 			*/
  htols(SUPPORTED_PROTOCOL_VERSION, &pacCommBufRunner);                 // uint16 protocol version SUPPORTED_PROTOCOL_VERSION = 1 */
  htols(0, &pacCommBufRunner);                                          // uint16 option_flags      */

  return pacCommBufRunner - pa_pacCommBuf;
}
// *********************************************************************************************************************************************

// *********************************************************************************************************************************************
EIP_INT32 Setup_UnRegisterSessionRequestMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request)
{
  /*
	  Un-RegisterSession command request
  */
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;

  htols(COMMAND_UNREGISTERSESSION, &pacCommBufRunner); 			/* uint16 command COMMAND_UNREGISTERSESSION = 0x66 */
  htols(0, &pacCommBufRunner);                    				/* uint16 length 			*/
  htoll(pa_cip_request->session_handle, &pacCommBufRunner);    	/* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                    				/* uint32 status 			*/
  htols(0, &pacCommBufRunner);                    				/* uint32 sender context 	*/
  htols(0, &pacCommBufRunner);                    				/* uint8 data[6] set to 0	*/
  htols(0, &pacCommBufRunner);
  htols(0, &pacCommBufRunner);
  htoll(0, &pacCommBufRunner);                    				/* uint32 options */

  return pacCommBufRunner - pa_pacCommBuf;
}
// *********************************************************************************************************************************************


// *********************************************************************************************************************************************
EIP_INT32 Setup_SendRRDataRequestMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request)
{
  /*
	  SendRRData UCMM Explicit message command
  */
  EIP_UINT16 i;
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;
  EIP_BYTE *acIdLenBuf, *acIdCmdSpecData;

  htols(COMMAND_SENDRRDATA,&pacCommBufRunner);					   		/* uint16 command COMMAND_SENDRRDATA = 0x6F	*/

  acIdLenBuf = pacCommBufRunner; 		   /*at this location the real length will be inserted later */
  pacCommBufRunner += 2;

  htoll(pa_cip_request->session_handle, &pacCommBufRunner);      		/* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                     						/* uint32 status 			*/

  htols(0, &pacCommBufRunner);     				   						/* uint8 sender context data[8]	*/
  htols(0, &pacCommBufRunner);
  htols(0, &pacCommBufRunner);
  htols(0, &pacCommBufRunner);

  htoll(0, &pacCommBufRunner);                     						/* uint32 options 		      	*/

  acIdCmdSpecData = pacCommBufRunner;         /* record the starting point of the command specific data */

  htoll(0, &pacCommBufRunner);                							// uint32 interface hangle      */
  htols(0, &pacCommBufRunner);                                          // uint16 timeout               */

  htols(2, &pacCommBufRunner);                                          // uint16 item count            */
  htols(0, &pacCommBufRunner);                                          // uint16 address item type id  */
  htols(0, &pacCommBufRunner);                                          // uint16 address item length   */
  htols(0xb2, &pacCommBufRunner);                                       // uint16 data item type id     */
  
  // set the data length
  if (pa_cip_request->service_code == CIP_SC_SET_ATTR_SINGLE)			// CIP_SC_SET_ATTR_SINGLE = 0x10
  {
	htols(pa_cip_request->data_length + 8, &pacCommBufRunner);          // uint16 data item length      */	
  }
  else
  {
	htols(8, &pacCommBufRunner);                                        // uint16 data item length      */	
  }

  *(pacCommBufRunner)++ = pa_cip_request->service_code; 			  	// example 0x0e;  	uint8 service request
  *(pacCommBufRunner)++ = 3;            								// 					uint8 path size in words
  *(pacCommBufRunner)++ = 0x20;											//					uint8
  *(pacCommBufRunner)++ = pa_cip_request->logical_class; 	   		 	// example 0x04 	uint8 class request
  *(pacCommBufRunner)++ = 0x24;											//					uint8
  *(pacCommBufRunner)++ = pa_cip_request->logical_instance;				// example 0x64  	uint8 instance request
  *(pacCommBufRunner)++ = 0x30;											//					uint8
  *(pacCommBufRunner)++ = pa_cip_request->logical_attribute;		 	// example 0x03  	uint8 attribute request

  // setup the data if needed.
  if (pa_cip_request->service_code == CIP_SC_SET_ATTR_SINGLE)			// CIP_SC_SET_ATTR_SINGLE = 0x10
  {
	  for(i=0; i < pa_cip_request->data_length; i++)
	  {
		 *(pacCommBufRunner)++ = pa_cip_request->data_buf[i];
      }
  }
  htols(pacCommBufRunner - acIdCmdSpecData, &acIdLenBuf);              /* update the command length field     */

  return pacCommBufRunner - pa_pacCommBuf;
}
// *********************************************************************************************************************************************


// *********************************************************************************************************************************************
EIP_INT32 Setup_NopRequestMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request)
{
  /*
	  NOP command request
  */
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;

  htols(COMMAND_NOP,&pacCommBufRunner);  						/* uint16 command COMMAND_NOP = 0x00	*/
  htols(0, &pacCommBufRunner);                    				/* uint16 length 			*/
  htoll(pa_cip_request->session_handle, &pacCommBufRunner);    	/* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                    				/* uint32 status 			*/
  htols(0, &pacCommBufRunner);                    				/* uint32 sender context 	*/
  htols(0, &pacCommBufRunner);                    				/* uint8 data[6] set to 0	*/
  htols(0, &pacCommBufRunner);
  htols(0, &pacCommBufRunner);
  htoll(0, &pacCommBufRunner);                    				/* uint32 options */

  return pacCommBufRunner - pa_pacCommBuf;
}
// *********************************************************************************************************************************************

// *********************************************************************************************************************************************
// Large Forward Open Request (max 4095 bytes)
// *********************************************************************************************************************************************
EIP_INT32 Setup_IOLargeForwardOpenRequestMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request)
{
   /*
	  SendRRData command with a Large Forward Open command request
   */
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;
  EIP_BYTE *acIdLenBuf, *acIdCmdSpecData, *acIdDataItemLenBuf;
  EIP_UINT32 temp32;

  // Seutp command request command
  htols(COMMAND_SENDRRDATA,&pacCommBufRunner);					   		/* uint16 command COMMAND_SENDRRDATA = 0x6F */
																		/* uint16 command length    */
  acIdLenBuf = pacCommBufRunner; 		   /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  htoll(pa_cip_request->session_handle, &pacCommBufRunner);   			/* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                     						/* uint32 status 			*/

  htols(0x00, &pacCommBufRunner);     				   		 			/* uint8 sender context data[8]	*/
  htols(0x00, &pacCommBufRunner);
  htols(0x00, &pacCommBufRunner);
  htols(0x00, &pacCommBufRunner);

  htoll(0, &pacCommBufRunner);                							/* uint32 options               */

  acIdCmdSpecData = pacCommBufRunner;         /* record the starting point of the command specific data */

  htoll(pa_cip_request->interface_handle, &pacCommBufRunner);    		/* uint32 interface handle      */
  htols(pa_cip_request->timeout, &pacCommBufRunner);                    /* uint16 timeout               */

  htols(2, &pacCommBufRunner);                                          /* uint16 item count            */
  htols(0, &pacCommBufRunner);                                          /* uint16 address item type id  */
  htols(0, &pacCommBufRunner);                                          /* uint16 address item length   */
  htols(0xb2, &pacCommBufRunner);                                       /* uint16 data item type id     */

																		/* uint16 data item length      */
  acIdDataItemLenBuf =  pacCommBufRunner;    /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  *(pacCommBufRunner)++ = 0x5B; 			  							// uint8 service request , Large Forward Open request

  *(pacCommBufRunner)++ = 2;            								// uint8 path size in words
  
  *(pacCommBufRunner)++ = 0x20;											// uint8 logical_class_segment
  *(pacCommBufRunner)++ = 0x06; 	   		 							// uint8 class request
  *(pacCommBufRunner)++ = 0x24;											// uint8 logical_instance_segment
  *(pacCommBufRunner)++ = 0x01;											// uint8 instance request

  *(pacCommBufRunner)++ = 0x06;                              			// uint8 priority_time_tick

  *(pacCommBufRunner)++ = pa_cip_request->timeout;				     	// uint8 timeout_ticks

  // setup to initial connection id's
  htoll(pa_cip_request->o_to_t_connection_id, &pacCommBufRunner);       // uint32  o_to_t_net_conn_id
  htoll(pa_cip_request->t_to_o_connection_id, &pacCommBufRunner);       // uint32  t_to_o_net_conn_id

  // setup the initial sequence numbers
  pa_cip_request->o_to_t_sequence_num = 1;
  pa_cip_request->t_to_o_sequence_num = 1;

  htols(pa_cip_request->connection_serial_number, &pacCommBufRunner);   /* uint16 connection serial number   */
  htols(pa_cip_request->vendor_id, &pacCommBufRunner);					/* uint16 vendor id                  */

  htoll(pa_cip_request->originator_serial_number, &pacCommBufRunner);  	// uint32  originator serial number */

  *(pacCommBufRunner)++ = 0;                          					// uint8 conn_timeout_multiplier
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved

  htoll(pa_cip_request->o_to_t_rpi, &pacCommBufRunner);          		// uint32  o_to_t_rpi (us)

																		// Owner Exclusive, Conn Type Point to Point,
																		// Scheduled Priority connection size x bytes (x+6?)
  temp32 = 0x48000000 + pa_cip_request->output_assembly_size + 6;		// Fixed sizes used for output assemblies.
  htoll(temp32, &pacCommBufRunner);                           			// uint32  o_to_t_net_conn_parameters.

  /*
   Network Connection Parameters, Large Forward Open
   Bit 31 		- Redundant Owner (0=Exclusive, Input-Only or Listen-Only, 1 = Redundant Owner in O->T direction)
   Bit 30 - 29  - Connection Type (00=Null, 01=Multicast, 10=Point-to-Point, 11=Reserved)
   Bit 28 		- Reserved
   Bit 27 - 26 	- Priority   (00= Low, 01= High, 10 = *Scheduled, 11 = Urgent
   Bit 25 		- Fixed or Variable (0=Fixed, 1 Variable)
   Bit 24 - 16	- Reserved
   Bit 15 - 0 	- Connection Size   (Size of data up to 65007)
  */

  htoll(pa_cip_request->t_to_o_rpi, &pacCommBufRunner);            		// uint32  t_to_o_rpi  (us)

  // Variable sizes used for input assemblies.

																		// Owner Exclusive, Conn Type Point to Point,
  temp32 = 0x4A000000 + pa_cip_request->input_assembly_size + 2;		// Scheduled Priority connection size x bytes (x+2?)
  htoll(temp32, &pacCommBufRunner);										// uint32  t_to_o_net_conn_parameters.

  *(pacCommBufRunner)++ = 0x01;                        					// uint8 Transport Class Trigger, Client(0), Cyclic(0). Class 1

  *(pacCommBufRunner)++ = 4;						   					// uint8 Number of 16-bit words in the connection path field.
  *(pacCommBufRunner)++ = 0x20;                        					// uint8 connection path class, instance
  *(pacCommBufRunner)++ = 0x04;                        					// uint8 Assembly Object connection_path.logical_class
  *(pacCommBufRunner)++ = 0x24;                         				// uint8 connection_path.logical_instance_segment
  *(pacCommBufRunner)++ = 0x01;                         				// uint8 connection_path.logical_instance

  *(pacCommBufRunner)++ = 0x2c;                         				// uint8 connection_point o->t
  *(pacCommBufRunner)++ = pa_cip_request->output_assembly_object;    	// uint8 connection_point o->t
  *(pacCommBufRunner)++ = 0x2c;                         				// uint8 connection_point t->o
  *(pacCommBufRunner)++ = pa_cip_request->input_assembly_object;		// uint8 connection_point t->o

  EIP_INT32 dataLen =  pacCommBufRunner - acIdCmdSpecData;
  htols(dataLen, &acIdLenBuf);       									/* update the command length field     */

  EIP_INT32 cmdLen = pacCommBufRunner - acIdDataItemLenBuf - 2;
  htols(cmdLen , &acIdDataItemLenBuf);									/* update the data length field     */

  return pacCommBufRunner - pa_pacCommBuf;
}
// *********************************************************************************************************************************************

// *********************************************************************************************************************************************
EIP_INT32 Setup_IOKeepAliveMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request)
{
   /*
	  Send UDP I/O KeepAliveMessage command request (UDP I/O socket only)
   */
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;
  EIP_UINT16 i;

  // Seutp keep alive command
  htols(2, &pacCommBufRunner);                          			    /* uint16 item count            */
  htols(0x8002, &pacCommBufRunner);                     			    /* uint16 address item type id  */
  htols(8, &pacCommBufRunner);                          			    /* uint16 address item length   */
  htoll(pa_cip_request->o_to_t_connection_id, &pacCommBufRunner);  	    /* uint32 connection id			*/
  htoll(pa_cip_request->o_to_t_sequence_num, &pacCommBufRunner);   	    /* uint32 sequence number		*/

  htols(0xb1, &pacCommBufRunner);                       			    /* uint16 data item type id     */
  
  i =  pa_cip_request->output_assembly_size + 6;
  htols(i, &pacCommBufRunner);                         			        /* uint16 data item length      */

  htols(pa_cip_request->o_to_t_sequence_num, &pacCommBufRunner);  	    /* uint16 sequence number		*/
  htoll(0x0003, &pacCommBufRunner);   								    /* uint32 header				*/

  // Send the output data
  for(i=0; i < pa_cip_request->output_assembly_size; i++)
  {	
	*(pacCommBufRunner)++ = pa_cip_request->data_buf[i];     		    /* uint8 data					*/
  }
  
  return pacCommBufRunner - pa_pacCommBuf;
}
// *********************************************************************************************************************************************

// *********************************************************************************************************************************************
EIP_INT32 Setup_IOLargeForwardCloseRequestMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request)
{
   /*
	  SendRRData command with a Forward Close command request
   */
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;
  EIP_BYTE *acIdLenBuf, *acIdCmdSpecData, *acIdDataItemLenBuf;

  // Seutp command request command
  htols(COMMAND_SENDRRDATA,&pacCommBufRunner);					   		/* uint16 command COMMAND_SENDRRDATA = 0x6F */
																		/* uint16 command length    */
  acIdLenBuf = pacCommBufRunner; 		   /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  htoll(pa_cip_request->session_handle, &pacCommBufRunner);    			/* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                     						/* uint32 status 			*/

  htols(0x00, &pacCommBufRunner);     				   		 			/* uint8 sender context data[8]	*/
  htols(0x00, &pacCommBufRunner);
  htols(0x00, &pacCommBufRunner);
  htols(0x00, &pacCommBufRunner);
  
  htoll(0, &pacCommBufRunner);                							/* uint32 options               */

  acIdCmdSpecData = pacCommBufRunner;         /* record the starting point of the command specific data */

  htoll(pa_cip_request->interface_handle, &pacCommBufRunner);    		/* uint32 interface hangle      */
  htols(pa_cip_request->timeout, &pacCommBufRunner);                    /* uint16 timeout               */

  htols(2, &pacCommBufRunner);                                          /* uint16 item count            */
  htols(0, &pacCommBufRunner);                                          /* uint16 address item type id  */
  htols(0, &pacCommBufRunner);                                          /* uint16 address item length   */
  htols(0xb2, &pacCommBufRunner);                                       /* uint16 data item type id     */

																		/* uint16 data item length      */
  acIdDataItemLenBuf =  pacCommBufRunner;    /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  *(pacCommBufRunner)++ = 0x4e; 			  							// uint8 service request Forward Close request
  
  *(pacCommBufRunner)++ = 2;            								// uint8 path size in words
  *(pacCommBufRunner)++ = 0x20;										    // uint8 logical_class_segment
  *(pacCommBufRunner)++ = 0x06; 	   		 	                        // uint8 class request - Message Router.
  *(pacCommBufRunner)++ = 0x24;										    // uint8 logical_instance_segment
  *(pacCommBufRunner)++ = 0x01;										    // uint8 instance request

  *(pacCommBufRunner)++ = 0x06;                               		    // uint8 priority_time_tick

  *(pacCommBufRunner)++ = pa_cip_request->timeout;         			    // uint8 timeout_ticks

  htols(pa_cip_request->connection_serial_number, &pacCommBufRunner);   /* uint16 connection serial number   */
  htols(pa_cip_request->vendor_id, &pacCommBufRunner);                  /* uint16 vendor id                  */

  htoll(pa_cip_request->originator_serial_number, &pacCommBufRunner);  	/* uint32  originator serial number */

  *(pacCommBufRunner)++ = 2;						   					// uint8 Number of 16-bit words in the connection path field.
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved

  *(pacCommBufRunner)++ = 0x20;										    // uint8 logical_class_segment
  *(pacCommBufRunner)++ = 0x02; 	   		 	                        // uint8 class request - Connection Manager.
  *(pacCommBufRunner)++ = 0x24;										    // uint8 logical_instance_segment
  *(pacCommBufRunner)++ = 0x01;										    // uint8 instance request

  EIP_INT32 dataLen =  pacCommBufRunner - acIdCmdSpecData;
  htols(dataLen, &acIdLenBuf);       		/* update the command length field     */

  EIP_INT32 cmdLen = pacCommBufRunner - acIdDataItemLenBuf - 2;
  htols(cmdLen , &acIdDataItemLenBuf);     /* update the data length field     */

  return pacCommBufRunner - pa_pacCommBuf;
}
// *********************************************************************************************************************************************
