/********************************************************************************************************
* OMRON Scientific Technologies Inc									  									*
* Project: 0949 - Next generation Laser Scanner, OS32													*
* Build: OS32C client application    																    *
* 																										*
* $Id: os32c_eip.h 2246 2012-11-19 23:55:21Z tom.charlton $
* Description: OS32C EIP header file																	*
********************************************************************************************************/

#ifndef  __OS32C_EIP_H__
#define  __OS32C_EIP_H__

/*
*********************************************************************************************************
*                                         GLOBAL DEFINITIONS
*********************************************************************************************************
*/
#define OSTI_VENDOR_ID							405			// Omron Scientific Technologies, Inc. Vendor Identification 
#define OS32C_PROD_CODE							12			// OS32C Product Code 

/* definition of status codes in encapsulation protocol */
#define EIP_STATUS_SUCCESS						0x0000
#define EIP_STATUS_INVALID_COMMAND				0x0001
#define EIP_STATUS_INSUFFICIENT_MEM				0x0002
#define EIP_STATUS_INCORRECT_DATA				0x0003
#define EIP_STATUS_INVALID_SESSION_HANDLE		0x0064
#define EIP_STATUS_INVALID_LENGTH				0x0065
#define EIP_STATUS_UNSUPPORTED_PROTOCOL			0x0069

/* definition of known encapsulation commands */
#define COMMAND_NOP                     0x0000
#define COMMAND_LISTSERVICES            0x0004
#define COMMAND_LISTIDENTITY            0x0063
#define COMMAND_LISTINTERFACES          0x0064
#define COMMAND_REGISTERSESSION         0x0065
#define COMMAND_UNREGISTERSESSION       0x0066
#define COMMAND_SENDRRDATA              0x006F
#define COMMAND_SENDUNITDATA            0x0070

#define SUPPORTED_PROTOCOL_VERSION				1
#define ENCAPSULATION_HEADER_SESSION_HANDLE_POS 4   /*the position of the session handle within the encapsulation header*/


#define OS32C_ASM_OBJECT_CLASS					0x04  // Assembly Object Class Type
#define OS32C_ASM_OBJ_DATA						0x03  // Assembly Object Data Type  
#define OS32C_ASM_SYS_STATUS_DATA				0x64  // Assembly Object Input Data (T->O), System Status data.
#define OS32C_ASM_SYS_DETECT_STATUS_DATA        0x65  // Assembly Object Input Data (T->O), System & Detection Status data.
#define OS32C_ASM_RTM_TOF_DATA					0x66  // Assembly Object Input Data (T->O), Range(TOF) measurement data.
#define OS32C_ASM_RTM_TOT_DATA					0x67  // Assembly Object Input Data (T->O), Reflectivity measurement data.
#define OS32C_ASM_RTM_TRIGGER_DATA				0x70  // Assembly Object Output Data (O->T), Output data to trigger UDP connection reports.
#define OS32C_ASM_RTM_CONFIG_DATA				0x71  // Assembly Object Output Data (O->T), Output data to trigger & configure UDP connection reports.

#define MAX_OS32C_ASM_RTM_TOF_DATA_SIZE			1410  // Maximum UDP IO data size
#define MAX_OS32C_ASM_RTM_TOT_DATA_SIZE			1410  // Maximum UDP IO data size
#define OS32C_ASM_RTM_TRIGGER_DATA_SIZE			2	  // UDP IO connection trigger data size
#define OS32C_ASM_RTM_CONFIG_DATA_SIZE			104	  // UDP IO connection configuration data size

#define OS32C_VS_SYSTEM_STATUS_DATA				0x70  // Vendor Specific Input Assembly (T->O), System Status data.
#define OS32C_VS_RTM_TOF_DATA					0x72  // Vendor Specific Input Assembly (T->O), Range(TOF) measurement data.
#define OS32C_VS_RTM_CONFIG_DATA                0x73  // Vendor Specific Output Assembly (O->T), Output data to configure measurement formats.
#define OS32C_VS_RTM_TOT_DATA					0x74  // Vendor Specific Input Assembly (T->O), Reflectivity measurement data.
#define OS32C_VS_RTM_TOF_AND_TOT_DATA			0x75  // Vendor Specific Input Assembly (T->O), Range & Reflectivity measurement data.

#define OS32C_VS_INSTANCE						1		// Vendor Specific Instance.
#define OS32C_VS_OBJ_SIZE_ATTRIBUTE	  			1		// Vendor Specific Object Size Attribute 
#define OS32C_VS_OBJ_ATTRIBUTE					3		// Vendor Specific Object Attribute 

// Measurement Data Attributes
#define OS32C_VS_RANGE_ATTRIBUTE   		        4		// Vendor Specific Attribute, Range Format
#define OS32C_VS_REFLECTIVITY_ATTRIBUTE			5		// Vendor Specific Attribute, Reflectivity Format 
#define OS32C_VS_BEAM_MASK_ATTRIBUTE            12		// Vendor Specific Attribute, Beam Selection Mask 
#define OS32C_VS_SCAN_COUNT_ATTRIBUTE			13		// Vendor Specific Attribute, Scan Count 
#define OS32C_VS_SCAN_PERIOD_ATTRIBUTE			14		// Vendor Specific Attribute, Scan Period
#define OS32C_VS_TIMESTAMP_ATTRIBUTE            15		// Vendor Specific Attribute, Time Stamp
#define OS32C_VS_BEAM_PERIOD_ATTRIBUTE			16		// Vendor Specific Attribute, Beam Period
#define OS32C_VS_BEAM_DATA_LENGTH_ATTRIBUTE		17		// Vendor Specific Attribute, Beam Data Length, number of beams in measurement reports.

// System Status Attributes
#define OS32C_VS_MACHINE_STATE_ATTRIBUTE        4       // Vendor Specific Attribute, Machine State
#define OS32C_VS_MACHINE_STOP_REASON_ATTRIBUTE  5       // Vendor Specific Attribute, Machine Stop Reason 
#define OS32C_VS_ACTIVE_ZONE_SET_ATTRIBUTE      6       // Vendor Specific Attribute, Active Zone Set 
#define OS32C_VS_ZONE_INPUT_COMBO_ATTRIBUTE     7       // Vendor Specific Attribute, Zone Input Combination 
#define OS32C_VS_DETECT_ZONE_STATUS_ATRIBUTE 	8       // Vendor Specific Attribute, Detection Zone Status 
#define OS32C_VS_OUTPUT_STATUS_ATRIBUTE		    9       // Vendor Specific Attribute, Output Status 
#define OS32C_VS_INPUT_STATUS_ATRIBUTE 			10      // Vendor Specific Attribute, Input Status 
#define OS32C_VS_7SEG_DISPLAY_ATRIBUTE 			11      // Vendor Specific Attribute, Seven Segment Display 
#define OS32C_VS_NONSAFETY_CRC_ATRIBUTE   	    12      // Vendor Specific Attribute, Non-Safety Configuration CRC Value 
#define OS32C_VS_SAFETY_CRC_ATRIBUTE 	  		13      // Vendor Specific Attribute, Safety Configuration CRC Value 

#define CIP_SC_GET_ATTR_SINGLE     				0x0E   	// Get Single Attribute command
#define CIP_SC_SET_ATTR_SINGLE     				0x10   	// Set Single Attribute command

#define CIP_SC_FWD_OPEN_REQ    			        0x54    // Open Standard connection (1 - 511 bytes)
#define CIP_SC_FWD_OPEN_RSP    			        0xD4    // Open Standard response

#define CIP_SC_LRG_FWD_OPEN_REQ    			    0x5B    // Open Large connection (512-65535 bytes)
#define CIP_SC_LRG_FWD_OPEN_RSP    			    0xDB    // Open Large connection response

#define CIP_SC_LRG_CLOSE_REQ     			    0x4E    // Close Large or small connection
#define CIP_SC_LRG_CLOSE_RSP     			    0xCE    // Close Large or small connection response
/*
*********************************************************************************************************
*                                         GLOBAL DECLARATIONS
*********************************************************************************************************
*/
enum RANGE_MEASUREMENT_FORMATS_OPTIONS
{
    NO_TOF_MEASUREMENTS = 0,	     // No measurement required for the specific format. To be used to disable Reflectivity measurements.

    RANGE_MEASURE_50M  = 1,          // Bit 0 to 15 (16-bit) distance measurement value (0 to 50,000 millimeters, 50 meters).

    RANGE_MEASURE_32M_PZ = 2,        // Bit 0 to 14 (15-bit) distance measurement value (0 to 32,766 millimeters, 32.766 meters).
                                     // Bit 15: measure value detected (beam blocked) within protection zone.

    RANGE_MEASURE_16M_WZ1PZ = 3,     // Bit 0 to 13 (14-bit) distance measurement value (0 to 16,382 millimeters, 16.382 meters).
                                     // Bit 14: measure value detected (beam blocked) within the warning zone 1.
                                     // Bit 15: measure value detected (beam blocked) within protection zone.

    RANGE_MEASURE_8M_WZ2WZ1PZ = 4,   // Bit 0 to 12 (13-bit) distance measurement value (0 to 8,190 millimeters, 8.190 meters).
                                     // Bit 13: measure value detected (beam blocked) within the warning zone 2.
                                     // Bit 14: measure value detected (beam blocked) within the warning zone 1.
                                     // Bit 15: measure value detected (beam blocked) within protection zone.

    RANGE_MEASURE_TOF_4PS  = 5       // Bit 0 to 15 (16-bit) TOF measurement value (0 to 65,534 x 4ps).
};


enum REFLECTIVITY_MEASUREMENT_FORMATS_OPTIONS
{
    NO_TOT_MEASUREMENTS = 0,	  	 // No measurement required for the specific format. To be used to disable Reflectivity measurements.

    REFLECT_MEASURE_TOT_ENCODED = 1, // Bit 0 to 9 (10-bit) TOT measurement value (0 to 1,000)
                                     // Bit 10: unused.
                                     // Bit 11: measure value detected (beam blocked) within the warning zone 2.
                                     // Bit 12: measure value detected (beam blocked) within the warning zone 1.
                                     // Bit 13: measure value detected (beam blocked) within protection zone.
                                     // Bit 14: noisy beam detected.
                                     // Bit 15: no reflection detected.

    REFLECT_MEASURE_TOT_4PS = 2	     // Bit 0 to 15 (16-bit) TOT measurement value (0 to 65,535 x 4ps).

};

#define	NUMBER_OF_BEAMS_IN_FRONTSCAN	677
#define BITS_PER_UINT32					32
#define DEFAULT_START_BEAM              0 															// Starting beam for default monitor data
#define DEFAULT_END_BEAM 			 	(NUMBER_OF_BEAMS_IN_FRONTSCAN - 1)    						// Ending beam for default monidata
#define NUM_OF_32BIT_BEAM_MASKS      	(((NUMBER_OF_BEAMS_IN_FRONTSCAN - 1) / BITS_PER_UINT32) + 1)	// Maintain 32-bit alignment.
#define NUM_OF_DETECTION_ZONE_BYTES		(NUM_OF_32BIT_BEAM_MASKS * sizeof(EIP_UINT32))

#define  DEF_BIT_NONE                   0x00u

#define  DEF_BIT_00						0x01u
#define  DEF_BIT_01                     0x02u
#define  DEF_BIT_02                     0x04u
#define  DEF_BIT_03                     0x08u
#define  DEF_BIT_04                     0x10u
#define  DEF_BIT_05                     0x20u
#define  DEF_BIT_06                     0x40u
#define  DEF_BIT_07                     0x80u

#define  DEF_BIT_08                     0x0100u
#define  DEF_BIT_09                     0x0200u
#define  DEF_BIT_10                     0x0400u
#define  DEF_BIT_11                     0x0800u
#define  DEF_BIT_12                     0x1000u
#define  DEF_BIT_13                     0x2000u
#define  DEF_BIT_14                     0x4000u
#define  DEF_BIT_15                     0x8000u

#define  DEF_BIT_16                     0x00010000u
#define  DEF_BIT_17                     0x00020000u
#define  DEF_BIT_18                     0x00040000u
#define  DEF_BIT_19                     0x00080000u
#define  DEF_BIT_20                     0x00100000u
#define  DEF_BIT_21                     0x00200000u
#define  DEF_BIT_22                     0x00400000u
#define  DEF_BIT_23                     0x00800000u

#define  DEF_BIT_24                     0x01000000u
#define  DEF_BIT_25                     0x02000000u
#define  DEF_BIT_26                     0x04000000u
#define  DEF_BIT_27                     0x08000000u
#define  DEF_BIT_28                     0x10000000u
#define  DEF_BIT_29                     0x20000000u
#define  DEF_BIT_30                     0x40000000u
#define  DEF_BIT_31                     0x80000000u

#define MEASUREMENT_8M_ZONEC_BLOCKED	    (DEF_BIT_13)
#define MEASUREMENT_8M_ZONEB_BLOCKED	    (DEF_BIT_14)
#define MEASUREMENT_8M_ZONEA_BLOCKED	    (DEF_BIT_15)

#define MEASUREMENT_16M_ZONEB_BLOCKED	(DEF_BIT_14)
#define MEASUREMENT_16M_ZONEA_BLOCKED	(DEF_BIT_15)

#define MEASUREMENT_32M_ZONEA_BLOCKED	(DEF_BIT_15)

#define BLINDED_BEAM_INDICATOR			1    			// Blinded beam indicator for range measurement data.
#define NO_REFLECTION_INDICATOR			0xFFFFFFFF		// No Reflection beam indicator for range measurement data.

#define MEASURED_TOT_ENCODE_NO_REFLECTION	(DEF_BIT_15)
#define MEASURED_TOT_ENCODE_NOISY_BEAM		(DEF_BIT_14)
#define MEASURED_TOT_ENCODE_ZONEA_BLOCKED	(DEF_BIT_13)
#define MEASURED_TOT_ENCODE_ZONEB_BLOCKED	(DEF_BIT_12)
#define MEASURED_TOT_ENCODE_ZONEC_BLOCKED	(DEF_BIT_11)

#define EIP_PROTECTION_ZONE_VIOLATION		(DEF_BIT_00)
#define EIP_WARNING_ZONE_1_VIOLATION	        (DEF_BIT_01)
#define EIP_WARNING_ZONE_2_VIOLATION        (DEF_BIT_02)
#define EIP_DUST_DETETION_WARNING			(DEF_BIT_03)

#define EIP_OSSD_OUTPUT_ACTIVE 	  		    (DEF_BIT_00)
#define EIP_AUXILIARY_OUTPUT_ACTIVE			(DEF_BIT_01)
#define EIP_WARNING_OUTPUT_ACTIVE  		    (DEF_BIT_02)

// This struct describes the System Status Info associated with each monitoring session.
#define EIP_STATUS_NUM_UNUSED 6
typedef struct _S_EIP_SystemStatusInfo
{
    EIP_UINT16 MachineState;			        // Machine State or OSSD state
    EIP_UINT16 MachineStopReason;          	// Reason OSSDs were turned-off
    EIP_UINT16 ActiveZoneSet;              	// Active Zone Set currently selected
    EIP_UINT16 ZoneInputCombo;				// All Active Zone Inputs
    EIP_UINT16 DetectionZoneStatus;        	// Detection Zone Violation Status,
                                            // Protection Zone = bit 0, Warning Zone #1 = bit 1, Warning Zone #2 = bit 2, Window Contamination = bit 3
    EIP_UINT16 OutputStatus;         		// Combined System Output Status,
                                            // OSSD Output= bit 0, Auxiliary Output = bit 1, Warning Output = bit 2
    EIP_UINT16 InputStatus;        		    // Combined System Input Status,
                                            // StandBy Input = bit 0, Start Input = bit 1 , EDM Input = bit 2
    EIP_UINT16 SevenSegDisplayCode;        	// Seven Segment display's code

    EIP_UINT16 NonSafetyConfigCRC; 		    // Non-Safety-Critical Configuration CRC value
    EIP_UINT16 SafetyConfigCRC;  			// Safety-Critical Configuration CRC value
    EIP_UINT16 Spare[EIP_STATUS_NUM_UNUSED];// Padded spares for future use

    // Total 32 bytes
} S_EIP_SystemStatusInfo;

// This enumeration defines the different possible reasons to go to machine stop
enum STOP_REASONS
{
    DIRTY_WINDOW		         = 0,   // Dirty window (must be the first in the enum as well)
    ZONE_SELECT_NUM_ACTIVE	 = 1,   // Invalid zone inputs, with wrong number of active inputs
    ZONE_SELECT_INVALID      = 2,   // Invalid zone inputs, with correct number of active inputs
    QUALIFIED_ZONE_VIOLATION = 3,   // Qualified zone violation
    STAND_BY_REQUEST         = 4,   // Stand-by or deep stand-by request
    POLARITY_ERROR           = 5,   // RX inversion polarity error
    NUM_STOP_REASONS         = 6,
    NOT_IN_MACHINE_STOP      = 7,   // System is not in machine stop mode and one of the corresponding machine states.
    SYTEM_FAULT              = 8    // Enum type but not used as array index
};

// This enumeration defines the different possible operating states of the system
enum SYSTEM_STATE
{
    POST                = 0,
    STOP			        = 1,
    INTERLOCK			= 2,
    RUN					= 3,
    STANDBY				= 4,
    DEEPSTANDBY			= 5,
    CONFIGURE			= 6,
    FAULT				= 7,
    SYSTEM_STATE_NUM
};

// This structure describes the EIP measurement report header provided during run-time monitoring.
// 16-bit Measurement data follows this header up to the define DataLength.
typedef struct _S_EIP_RTM_Header
{
    EIP_UINT32 ScanCount; 						// Current scan (angular rotation) count for time stamping messages.
    EIP_UINT32 ScanPeriod; 						// Scan period in microseconds to be used for time stamping messages.
    EIP_UINT32 TimeStamp;						// Scan count time Scan Period = Time stamp of the current scan cycle.
    EIP_UINT32 BeamPeriod;						// Beam to beam period.
    S_EIP_SystemStatusInfo SystemStatus;	        // 32 bytes of EIP run-time system status information.
    EIP_UINT16 RangeFormat;    			   		// Range (TOF) Measurement Format
    EIP_UINT16 ReflectivityFormat; 		   		// Reflectivitiy (TOT) Measurement Format
    EIP_UINT16 Spare;                           // Spare to maintain 32-bit alignment.
    EIP_UINT16 DataLength;						// Size of the data report to follow.
} S_EIP_RTM_Header;


// This structure describes the configuration parameters needed for the run-time montioring task.
typedef struct _S_EIP_RTM_Configuration
{
    EIP_UINT16 RangeFormat;    			   				// Range (TOF) Measurement Format
    EIP_UINT16 ReflectivityFormat; 		   				// Reflectivitiy (TOT) Measurement Format
    EIP_UINT16 Spare[6];               			        // Future use spares.
    EIP_UINT32 BeamReportMask[NUM_OF_32BIT_BEAM_MASKS]; // Selected Measurement Beam Mask,  beam bit x = 0 (ignore) , beam bit x = 1 (report).

} S_EIP_RTM_Configuration;



#endif /* __OS32C_EIP_H__ */
