/*******************************************************************************
 * Copyright (c) 2009, Rockwell Automation, Inc.
 * All rights reserved. 
 *
 ******************************************************************************/
#include <string.h>
#include <stdlib.h>
#include "cpf.h"
#include "encap.h"
#include "endianconv.h"
#include "cipcommon.h"
#include "cipmessagerouter.h"
#include "cipconnectionmanager.h"
#include "cipidentity.h"

/*Identity data from cipidentity.c*/
#ifdef __EIP_SERVER__
extern EIP_UINT16 VendorID;
extern EIP_UINT16 DeviceType;
extern EIP_UINT16 ProductCode;
extern S_CIP_Revision Revison;
extern EIP_UINT16 ID_Status;
extern EIP_UINT32 SerialNumber;
extern S_CIP_Short_String ProductName;

#else
EIP_UINT16 VendorID;
EIP_UINT16 DeviceType;
EIP_UINT16 ProductCode;
S_CIP_Revision Revison;
EIP_UINT16 ID_Status;
EIP_UINT32 SerialNumber;
S_CIP_Short_String ProductName;
#endif // #ifdef __EIP_SERVER__

/*ip address data taken from TCPIPInterfaceObject*/
#ifdef __EIP_SERVER__
extern S_CIP_TCPIPNetworkInterfaceConfiguration Interface_Configuration;
#endif

/*** defines ***/

#define ITEM_ID_LISTIDENTITY 0x000C

#define SUPPORTED_PROTOCOL_VERSION 1

#define SUPPORTED_OPTIONS_MASK  0x00  /*Mask of which options are supported as of the current CIP specs no other option value as 0 should be supported.*/

#define ENCAPSULATION_HEADER_SESSION_HANDLE_POS 4   /*the position of the session handle within the encapsulation header*/

#define INVALID_SESSION -1

#define SENDER_CONTEXT_SIZE 8                   /*size of sender context in encapsulation header*/

/* definition of known encapsulation commands */
#define COMMAND_NOP                     0x0000
#define COMMAND_LISTSERVICES            0x0004
#define COMMAND_LISTIDENTITY            0x0063
#define COMMAND_LISTINTERFACES          0x0064
#define COMMAND_REGISTERSESSION         0x0065
#define COMMAND_UNREGISTERSESSION       0x0066
#define COMMAND_SENDRRDATA              0x006F
#define COMMAND_SENDUNITDATA            0x0070

/* definition of capability flags */
#define SUPPORT_CIP_TCP                 0x0020
#define SUPPORT_CIP_UDP_CLASS_0_OR_1    0x0100

#define ENCAP_NUMBER_OF_SUPPORTED_DELAYED_ENCAP_MESSAGES 2 /*According to EIP spec at least 2 delayed message requests schould be supporte */

#define ENCAP_MAX_DELAYED_ENCAP_MESSAGE_SIZE  (ENCAPSULATION_HEADER_LENGTH + 39 + sizeof(OPENER_DEVICE_NAME)) /* currently we only have the size of an encapsulation message */

/* Encapsulation layer data  */

struct SDelayedEncapsulationMessage
{
  EIP_INT32 m_unTimeOut;
  EIP_INT32 m_nSocket;
  struct sockaddr_in m_stSendToAddr;
  EIP_BYTE m_anMsg[ENCAP_MAX_DELAYED_ENCAP_MESSAGE_SIZE];
  EIP_UINT32 m_unMessageSize;
};

struct S_Encapsulation_Interface_Information g_stInterfaceInformation;

int anRegisteredSessions[OPENER_NUMBER_OF_SUPPORTED_SESSIONS];

struct SDelayedEncapsulationMessage g_stDelayedEncapsulationMessages[ENCAP_NUMBER_OF_SUPPORTED_DELAYED_ENCAP_MESSAGES];

/*** private functions ***/
void
handleReceivedListServicesCmd(struct S_Encapsulation_Data *pa_stReceiveData);
void
handleReceivedListInterfacesCmd(struct S_Encapsulation_Data *pa_stReceiveData);

void
handleReceivedListIdentityCmdTCP(struct S_Encapsulation_Data *pa_stReceiveData);

void
handleReceivedListIdentityCmdUDP(int pa_nSocket,
    struct sockaddr_in *pa_pstFromAddr,
    struct S_Encapsulation_Data *pa_stReceiveData);

void
handleReceivedRegisterSessionCmd(int pa_nSockfd,
    struct S_Encapsulation_Data *pa_stReceiveData);
EIP_STATUS
handleReceivedUnregisterSessionCmd(
    struct S_Encapsulation_Data *pa_stReceiveData);
EIP_STATUS
handleReceivedSendUnitDataCmd(struct S_Encapsulation_Data *pa_stReceiveData);
EIP_STATUS
handleReceivedSendRRDataCmd(struct S_Encapsulation_Data *pa_stReceiveData);

EIP_INT32
getFreeSessionIndex(void);
EIP_INT16
createEncapsulationStructure(EIP_BYTE * buf, EIP_INT32 length,
    struct S_Encapsulation_Data *pa_S_ReceiveData);
EIP_STATUS
checkRegisteredSessions(struct S_Encapsulation_Data *pa_S_ReceiveData);
EIP_INT32
encapsulate_data(struct S_Encapsulation_Data *pa_S_SendData);

void
determineDelayTime(EIP_BYTE *pa_acBufferStart,
    struct SDelayedEncapsulationMessage *pa_pstDelayedMessageBuffer);

EIP_INT32
encapsulateListIdentyResponseMessage(EIP_BYTE *pa_pacCommBuf);

/*   void encapInit(void)
 *   initialize sessionlist and interfaceinformation.
 */

void
encapInit(void)
{
  EIP_UINT32 i;

  determineEndianess();

  /*initialize random numbers for random delayed response message generation
   * we use the ip address as seed as suggested in the spec */
  // srand(Interface_Configuration.IPAddress);

  /* initialize Sessions to invalid == free session */
  for (i = 0; i < OPENER_NUMBER_OF_SUPPORTED_SESSIONS; i++)
    {
      anRegisteredSessions[i] = EIP_INVALID_SOCKET;
    }

  for (i = 0; i < ENCAP_NUMBER_OF_SUPPORTED_DELAYED_ENCAP_MESSAGES; i++)
    {
      g_stDelayedEncapsulationMessages[i].m_nSocket = -1;
    }

  /*TODO make the interface information configurable*/
  /* initialize interface information */
  g_stInterfaceInformation.TypeCode = CIP_ITEM_ID_LISTSERVICE_RESPONSE;
  g_stInterfaceInformation.Length = sizeof(g_stInterfaceInformation);
  g_stInterfaceInformation.EncapsulationProtocolVersion = 1;
  g_stInterfaceInformation.CapabilityFlags = SUPPORT_CIP_TCP | SUPPORT_CIP_UDP_CLASS_0_OR_1;
  
#ifdef __EIP_SERVER__  
  strcpy((char *) g_stInterfaceInformation.NameofService, "communications");
#endif

}

EIP_INT32
handleReceivedExplictTCPData(EIP_INT32 pa_socket, EIP_BYTE * pa_buf,
	EIP_UINT32 pa_length, EIP_INT32 *pa_nRemainingBytes)
{
  EIP_INT32 nRetVal = 0;
  struct S_Encapsulation_Data sEncapData;
  /* eat the encapsulation header*/
  /* the structure contains a pointer to the encapsulated data*/
  /* returns how many bytes are left after the encapsulated data*/
  *pa_nRemainingBytes = createEncapsulationStructure(pa_buf, pa_length,
      &sEncapData);

  if (SUPPORTED_OPTIONS_MASK == sEncapData.nOptions) /*TODO generate appropriate error response*/
    {
      if (*pa_nRemainingBytes >= 0) /* check if the message is corrupt: header size + claimed payload size > than what we actually received*/
        {
          /* full package or more received */
          sEncapData.nStatus = OPENER_ENCAP_STATUS_SUCCESS;
          nRetVal = 1;
          /* most of these functions need a reply to be send */
          switch (sEncapData.nCommand_code)
            {
          case (COMMAND_NOP):
            /* Nop needs no reply and does nothing */
            nRetVal = 0;
            break;

          case (COMMAND_LISTSERVICES):
            handleReceivedListServicesCmd(&sEncapData);
            break;

          case (COMMAND_LISTIDENTITY):
			handleReceivedListIdentityCmdTCP(&sEncapData);
            break;

          case (COMMAND_LISTINTERFACES):
            handleReceivedListInterfacesCmd(&sEncapData);
            break;
#ifdef __EIP_SERVER__
          case (COMMAND_REGISTERSESSION):
            handleReceivedRegisterSessionCmd(pa_socket, &sEncapData);
            break;

          case (COMMAND_UNREGISTERSESSION):
            nRetVal = handleReceivedUnregisterSessionCmd(&sEncapData);
            break;
#endif // #ifdef __EIP_SERVER__
          case (COMMAND_SENDRRDATA):
            nRetVal = handleReceivedSendRRDataCmd(&sEncapData);
            break;

          case (COMMAND_SENDUNITDATA):
            nRetVal = handleReceivedSendUnitDataCmd(&sEncapData);
            break;

          default:
            sEncapData.nStatus = OPENER_ENCAP_STATUS_INVALID_COMMAND;
            sEncapData.nData_length = 0;
            break;
            }
          /* if nRetVal is greater then 0 data has to be sent */
          if (0 < nRetVal)
            {
              nRetVal = encapsulate_data(&sEncapData);
            }
        }
    }

  return nRetVal;
}

EIP_INT32
handleReceivedExplictUDPData(EIP_INT32 pa_socket, struct sockaddr_in *pa_pstFromAddr,
	EIP_BYTE* pa_buf, EIP_UINT32 pa_length, EIP_INT32 *pa_nRemainingBytes)
{
  EIP_INT32 nRetVal = 0;
  struct S_Encapsulation_Data sEncapData;
  /* eat the encapsulation header*/
  /* the structure contains a pointer to the encapsulated data*/
  /* returns how many bytes are left after the encapsulated data*/
  *pa_nRemainingBytes = createEncapsulationStructure(pa_buf, pa_length,
      &sEncapData);

  if (SUPPORTED_OPTIONS_MASK == sEncapData.nOptions) /*TODO generate appropriate error response*/
    {
      if (*pa_nRemainingBytes >= 0) /* check if the message is corrupt: header size + claimed payload size > than what we actually received*/
        {
          /* full package or more received */
          sEncapData.nStatus = OPENER_ENCAP_STATUS_SUCCESS;
          nRetVal = 1;
          /* most of these functions need a reply to be send */
          switch (sEncapData.nCommand_code)
            {
          case (COMMAND_LISTSERVICES):
            handleReceivedListServicesCmd(&sEncapData);
            break;
#ifdef __EIP_SERVER__
          case (COMMAND_LISTIDENTITY):
            handleReceivedListIdentityCmdUDP(pa_socket, pa_pstFromAddr,
                &sEncapData);
            nRetVal = EIP_OK; /* as the response has to be delayed do not send it now */
            break;
#endif // #ifdef __EIP_SERVER__
          case (COMMAND_LISTINTERFACES):
            handleReceivedListInterfacesCmd(&sEncapData);
            break;

            /* Fhe following commands are not to be sent via UDP */
          case (COMMAND_NOP):
          case (COMMAND_REGISTERSESSION):
          case (COMMAND_UNREGISTERSESSION):
          case (COMMAND_SENDRRDATA):
          case (COMMAND_SENDUNITDATA):
          default:
            sEncapData.nStatus = OPENER_ENCAP_STATUS_INVALID_COMMAND;
            sEncapData.nData_length = 0;
            break;
            }
          /* if nRetVal is greater then 0 data has to be sent */
          if (0 < nRetVal)
            {
              nRetVal = encapsulate_data(&sEncapData);
            }
        }
    }

  return nRetVal;
}

EIP_INT32
encapsulate_data(struct S_Encapsulation_Data *pa_stSendData)
{
  EIP_UINT8 *acCommBuf = &(pa_stSendData->m_acCommBufferStart[2]);
  /*htols(pa_stSendData->nCommand_code, &pa_stSendData->pEncapsulation_Data);*/
  htols(pa_stSendData->nData_length, &acCommBuf);
  /*the CommBuf should already contain the correct session handle*/
  /*htoll(pa_stSendData->nSession_handle, &pa_stSendData->pEncapsulation_Data); */
  acCommBuf += 4;
  htoll(pa_stSendData->nStatus, &acCommBuf);
  /*the CommBuf should already contain the correct sender context*/
  /*memcpy(pa_stSendData->pEncapsulation_Data, pa_stSendData->anSender_context, SENDER_CONTEXT_SIZE);*/
  /*pa_stSendData->pEncapsulation_Data += SENDER_CONTEXT_SIZE + 2;*//* the plus 2 is for the options value*/
  /*the CommBuf should already contain the correct  options value*/
  /*htols((EIP_UINT16)pa_stSendData->nOptions, &pa_stSendData->pEncapsulation_Data);*/

  return ENCAPSULATION_HEADER_LENGTH + pa_stSendData->nData_length;
}

/*   INT8 ListServices(struct S_Encapsulation_Data *pa_S_ReceiveData)
 *   generate reply with "Communications Services" + compatibility Flags.
 *      pa_S_ReceiveData pointer to structure with received data
 */
void
handleReceivedListServicesCmd(struct S_Encapsulation_Data *pa_stReceiveData)
{
  EIP_UINT8 *pacCommBuf = pa_stReceiveData->m_acCurrentCommBufferPos;

  pa_stReceiveData->nData_length = g_stInterfaceInformation.Length + 2;

  /* copy Interface data to nmsg for sending */
  htols(1, &pacCommBuf);
  htols(g_stInterfaceInformation.TypeCode, &pacCommBuf);
  htols((EIP_UINT16) (g_stInterfaceInformation.Length - 4), &pacCommBuf);
  htols(g_stInterfaceInformation.EncapsulationProtocolVersion, &pacCommBuf);
  htols(g_stInterfaceInformation.CapabilityFlags, &pacCommBuf);
  memcpy(pacCommBuf, g_stInterfaceInformation.NameofService,
      sizeof(g_stInterfaceInformation.NameofService));
}

void
handleReceivedListInterfacesCmd(struct S_Encapsulation_Data *pa_stReceiveData)
{
  EIP_UINT8 *pacCommBuf = pa_stReceiveData->m_acCurrentCommBufferPos;
  pa_stReceiveData->nData_length = 2;
  htols(0x0000, &pacCommBuf); /* copy Interface data to nmsg for sending */
}

void
handleReceivedListIdentityCmdTCP(struct S_Encapsulation_Data * pa_stReceiveData)
{
  pa_stReceiveData->nData_length = encapsulateListIdentyResponseMessage(
      pa_stReceiveData->m_acCurrentCommBufferPos);
}

void
handleReceivedListIdentityCmdUDP(EIP_INT32 pa_nSocket,
    struct sockaddr_in *pa_pstFromAddr,
    struct S_Encapsulation_Data *pa_stReceiveData)
{
  struct SDelayedEncapsulationMessage *pstDelayedMessageBuffer = NULL;
  EIP_UINT32 i;
  EIP_BYTE *pacCommBuf;

  for (i = 0; i < ENCAP_NUMBER_OF_SUPPORTED_DELAYED_ENCAP_MESSAGES; i++)
    {
      if (-1 == g_stDelayedEncapsulationMessages[i].m_nSocket)
        {
          pstDelayedMessageBuffer = &(g_stDelayedEncapsulationMessages[i]);
          break;
        }
    }

  if (NULL != pstDelayedMessageBuffer)
    {
      pstDelayedMessageBuffer->m_nSocket = pa_nSocket;
      memcpy((&pstDelayedMessageBuffer->m_stSendToAddr), pa_pstFromAddr,
          sizeof(struct sockaddr_in));

      determineDelayTime(pa_stReceiveData->m_acCommBufferStart,
          pstDelayedMessageBuffer);

      memcpy(&(pstDelayedMessageBuffer->m_anMsg[0]),
          pa_stReceiveData->m_acCommBufferStart, ENCAPSULATION_HEADER_LENGTH);

      pstDelayedMessageBuffer->m_unMessageSize =
		  encapsulateListIdentyResponseMessage(
              &(pstDelayedMessageBuffer->m_anMsg[ENCAPSULATION_HEADER_LENGTH]));

	  // TOM-TBD, MODIFIED FROM THE ORIGINAL CODE
	  // EIP_UINT8 *pacCommBuf = pstDelayedMessageBuffer->m_anMsg + 2;
	  pacCommBuf = pstDelayedMessageBuffer->m_anMsg + 2;
	  htols(pstDelayedMessageBuffer->m_unMessageSize, &pacCommBuf);
      pstDelayedMessageBuffer->m_unMessageSize += ENCAPSULATION_HEADER_LENGTH;
    }
}

EIP_INT32
encapsulateListIdentyResponseMessage(EIP_BYTE *pa_pacCommBuf)
{
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;
  EIP_BYTE *acIdLenBuf;

  htols(1, &pacCommBufRunner); /* one item */
  htols(ITEM_ID_LISTIDENTITY, &pacCommBufRunner);

  acIdLenBuf = pacCommBufRunner;
  pacCommBufRunner += 2; /*at this place the real length will be inserted below*/

  htols(SUPPORTED_PROTOCOL_VERSION, &pacCommBufRunner);

#ifdef __EIP_SERVER__
	encapsulateIPAdress(OPENER_ETHERNET_PORT, Interface_Configuration.IPAddress, pacCommBufRunner);
#endif

  pacCommBufRunner += 8;

  memset(pacCommBufRunner, 0, 8);
  pacCommBufRunner += 8;

  htols(VendorID, &pacCommBufRunner);
  htols(DeviceType, &pacCommBufRunner);
  htols(ProductCode, &pacCommBufRunner);
  *(pacCommBufRunner)++ = Revison.MajorRevision;
  *(pacCommBufRunner)++ = Revison.MinorRevision;
  htols(ID_Status, &pacCommBufRunner);
  htoll(SerialNumber, &pacCommBufRunner);
  *pacCommBufRunner++ = (EIP_BYTE) ProductName.Length;
  memcpy(pacCommBufRunner, ProductName.String, ProductName.Length);
  pacCommBufRunner += ProductName.Length;
  *pacCommBufRunner++ = 0xFF;

  htols(pacCommBufRunner - acIdLenBuf - 2, &acIdLenBuf); /* the -2 is for not counting the length field*/

  return pacCommBufRunner - pa_pacCommBuf;
}
// ---------------------------------------------------------------------------
EIP_INT32 encapsulateListIdentityRequestMessage(EIP_BYTE *pa_pacCommBuf, EIP_UINT16 MaxResponseDelay)
{
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;

  htols(COMMAND_LISTIDENTITY, &pacCommBufRunner); /* uint16 command 		*/
  htols(0, &pacCommBufRunner);                    /* uint16 length 			*/
  htoll(0, &pacCommBufRunner);                    /* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                    /* uint32 status 			*/
  htols(MaxResponseDelay, &pacCommBufRunner);     /* uint32 sender context 	*/
  htols(0, &pacCommBufRunner);                    /* uint8 data[6] set to 0	*/
  htols(0, &pacCommBufRunner);
  htols(0, &pacCommBufRunner);
  htoll(0, &pacCommBufRunner);                    /* uint32 options */

  return pacCommBufRunner - pa_pacCommBuf;
}
// ----------------------------------------------------------------------------

// ---------------------------------------------------------------------------
EIP_INT32 encapsulateListInterfacesRequestMessage(EIP_BYTE *pa_pacCommBuf)
{
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;

  htols(COMMAND_LISTINTERFACES,&pacCommBufRunner); /* uint16 command 			*/
  htols(0, &pacCommBufRunner);                     /* uint16 length 			*/
  htoll(0, &pacCommBufRunner);                     /* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                     /* uint32 status 			*/
  htols(0, &pacCommBufRunner);     				   /* uint32 sender context 	*/
  htols(0, &pacCommBufRunner);                     /* uint8 data[6] set to 0	*/
  htols(0, &pacCommBufRunner);
  htols(0, &pacCommBufRunner);
  htoll(0, &pacCommBufRunner);                     /* uint32 options */

  return pacCommBufRunner - pa_pacCommBuf;
}
// ----------------------------------------------------------------------------

// ---------------------------------------------------------------------------
EIP_INT32 encapsulateListServicesRequestMessage(EIP_BYTE *pa_pacCommBuf)
{
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;

  htols(COMMAND_LISTSERVICES,&pacCommBufRunner);   /* uint16 command 			*/
  htols(0, &pacCommBufRunner);                     /* uint16 length 			*/
  htoll(0, &pacCommBufRunner);                     /* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                     /* uint32 status 			*/
  htols(0, &pacCommBufRunner);     				   /* uint32 sender context 	*/
  htols(0, &pacCommBufRunner);                     /* uint8 data[6] set to 0	*/
  htols(0, &pacCommBufRunner);
  htols(0, &pacCommBufRunner);
  htoll(0, &pacCommBufRunner);                     /* uint32 options */

  return pacCommBufRunner - pa_pacCommBuf;
}
// ----------------------------------------------------------------------------

// ---------------------------------------------------------------------------
EIP_INT32 encapsulateNopRequestMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request)
{
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;

  htols(COMMAND_NOP,&pacCommBufRunner);  		                /* uint16 command 			*/
  htols(0, &pacCommBufRunner);                    				/* uint16 length 			*/
  htoll(pa_cip_request->session_handle, &pacCommBufRunner);    	/* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                    				/* uint32 status 			*/
  htols(0, &pacCommBufRunner);                    				/* uint32 sender context 	*/
  htols(0, &pacCommBufRunner);                    				/* uint8 data[6] set to 0	*/
  htols(0, &pacCommBufRunner);
  htols(0, &pacCommBufRunner);
  htoll(0, &pacCommBufRunner);                    				/* uint32 options */

  return pacCommBufRunner - pa_pacCommBuf;
}
// ----------------------------------------------------------------------------

// ---------------------------------------------------------------------------
EIP_INT32 encapsulateRegisterSessionRequestMessage(EIP_BYTE *pa_pacCommBuf)
{
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;

  htols(COMMAND_REGISTERSESSION,&pacCommBufRunner);						/* uint16 command */
  htols(ENCAPSULATION_HEADER_SESSION_HANDLE_POS, &pacCommBufRunner); 	/* uint16 length  */
  htoll(0, &pacCommBufRunner);                     						/* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                     						/* uint32 status 			*/
  htols(0, &pacCommBufRunner);     				   						/* uint32 sender context 	*/
  htols(0, &pacCommBufRunner);                     						/* uint8 data[6] set to 0	*/
  htols(0, &pacCommBufRunner);
  htols(0, &pacCommBufRunner);
  htoll(0, &pacCommBufRunner);                     						/* uint32 options 			*/
  htols(SUPPORTED_PROTOCOL_VERSION, &pacCommBufRunner);                 // uint16 protocol version  */
  htols(0, &pacCommBufRunner);                                          // uint16 option_flags      */

  return pacCommBufRunner - pa_pacCommBuf;
}
// ----------------------------------------------------------------------------

// ---------------------------------------------------------------------------
EIP_INT32 encapsulateUnRegisterSessionRequestMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request)
{
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;

  htols(COMMAND_UNREGISTERSESSION, &pacCommBufRunner); 			/* uint16 command 		*/
  htols(0, &pacCommBufRunner);                    				/* uint16 length 			*/
  htoll(pa_cip_request->session_handle, &pacCommBufRunner);    	/* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                    				/* uint32 status 			*/
  htols(0, &pacCommBufRunner);                    				/* uint32 sender context 	*/
  htols(0, &pacCommBufRunner);                    				/* uint8 data[6] set to 0	*/
  htols(0, &pacCommBufRunner);
  htols(0, &pacCommBufRunner);
  htoll(0, &pacCommBufRunner);                    				/* uint32 options */

  return pacCommBufRunner - pa_pacCommBuf;
}
// ----------------------------------------------------------------------------

// ---------------------------------------------------------------------------
EIP_INT32 encapsulateUCMMSendRRDataRequestMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request)
{
  /*
	  SendRRData UCMM Explicit message command
  */
  EIP_UINT16 i;
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;
  EIP_BYTE *acIdLenBuf, *acIdCmdSpecData;

  htols(COMMAND_SENDRRDATA,&pacCommBufRunner);					   		/* uint16 command 			*/

  acIdLenBuf = pacCommBufRunner; 		   /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  htoll(0, &pacCommBufRunner);                     	 					/* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                     						/* uint32 status 			*/

  htols(0, &pacCommBufRunner);     				   						/* uint8 sender context data[8]	*/
  htols(0, &pacCommBufRunner);
  htols(0, &pacCommBufRunner);
  htols(0, &pacCommBufRunner);

  htoll(0, &pacCommBufRunner);                     						/* uint32 options 		      	*/

  acIdCmdSpecData = pacCommBufRunner;         /* record the starting point of the command specific data */

  htoll(0, &pacCommBufRunner);                							// uint32 interface hangle      */
  htols(0, &pacCommBufRunner);                                          // uint16 timeout               */

  htols(2, &pacCommBufRunner);                                          // uint16 item count            */
  htols(0, &pacCommBufRunner);                                          // uint16 address item type id  */
  htols(0, &pacCommBufRunner);                                          // uint16 address item length   */
  htols(0xb2, &pacCommBufRunner);                                       // uint16 data item type id     */
  
  // set the data length
  if (pa_cip_request->service_code == CIP_SET_ATTRIBUTE_SINGLE)
  {
	htols(pa_cip_request->data_length + 8, &pacCommBufRunner);          // uint16 data item length      */	
  }
  else
  {
	htols(8, &pacCommBufRunner);                                        // uint16 data item length      */	
  }

  *(pacCommBufRunner)++ = pa_cip_request->service_code; 			  		// example 0x0e;  	uint8 service request
  *(pacCommBufRunner)++ = 3;            									// 					uint8 path size in words
  *(pacCommBufRunner)++ = 0x20;											//					uint8
  *(pacCommBufRunner)++ = pa_cip_request->logical_class; 	   		 	// example. 0x04 	uint8 class request
  *(pacCommBufRunner)++ = 0x24;											//					uint8
  *(pacCommBufRunner)++ = pa_cip_request->logical_instance;				// example 0x65  	uint8 instance request
  *(pacCommBufRunner)++ = 0x30;											//					uint8
  *(pacCommBufRunner)++ = pa_cip_request->logical_attribute;		 		// example 0x03  	uint8 attribute request

  // setup the data if needed.
  if (pa_cip_request->service_code == CIP_SET_ATTRIBUTE_SINGLE)
  {
	  for(i=0; i < pa_cip_request->data_length; i++)
	  {
		 *(pacCommBufRunner)++ = pa_cip_request->data_buf[i];
      }
  }
  htols(pacCommBufRunner - acIdCmdSpecData, &acIdLenBuf);              /* update the command length field     */

  return pacCommBufRunner - pa_pacCommBuf;
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Forward Open Request (max 511 bytes)
// ----------------------------------------------------------------------------
EIP_INT32 encapsulateSendForwardOpenRequestMessage(EIP_BYTE *pa_pacCommBuf,
											 struct S_Service_Code_Request_Format *pa_cip_request)
{
   /*
	  SendForwardOpen message command
   */
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;
  EIP_BYTE *acIdLenBuf, *acIdCmdSpecData, *acIdDataItemLenBuf;

  // Seutp command request command
  htols(COMMAND_SENDRRDATA,&pacCommBufRunner);					   		/* uint16 command 			*/
																		/* uint16 command length    */
  acIdLenBuf = pacCommBufRunner; 		   /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  htoll(pa_cip_request->session_handle, &pacCommBufRunner);   			/* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                     						/* uint32 status 			*/

  htols(0x01, &pacCommBufRunner);     				   		 			/* uint8 sender context data[8]	*/
  htols(0x02, &pacCommBufRunner);
  htols(0x04, &pacCommBufRunner);
  htols(0x08, &pacCommBufRunner);

  htoll(0, &pacCommBufRunner);                							/* uint32 options               */

  acIdCmdSpecData = pacCommBufRunner;         /* record the starting point of the command specific data */

  htoll(0, &pacCommBufRunner);                							/* uint32 interface hangle      */
  htols(3, &pacCommBufRunner);                                          /* uint16 timeout               */

  htols(2, &pacCommBufRunner);                                          /* uint16 item count            */
  htols(0, &pacCommBufRunner);                                          /* uint16 address item type id  */
  htols(0, &pacCommBufRunner);                                          /* uint16 address item length   */
  htols(0xb2, &pacCommBufRunner);                                       /* uint16 data item type id     */

																		/* uint16 data item length      */
  acIdDataItemLenBuf =  pacCommBufRunner;    /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  *(pacCommBufRunner)++ = 0x54; 			  								// uint8 service request

  *(pacCommBufRunner)++ = 2;            									// uint8 path size in words
  *(pacCommBufRunner)++ = 0x20;											// uint8 logical_class_segment
  *(pacCommBufRunner)++ = 0x06							; 	   		 	// uint8 class request
  *(pacCommBufRunner)++ = 0x24;											// uint8 logical_instance_segment
  *(pacCommBufRunner)++ = 0x01;											// uint8 instance request

  *(pacCommBufRunner)++ = 0x06;                              			// uint8 priority_time_tick

  *(pacCommBufRunner)++ = 40;                          					// uint8 timeout_ticks

  htoll(0, &pacCommBufRunner);                          					// uint32  o_to_t_net_conn_id
  htoll(1, &pacCommBufRunner);                           				// uint32  t_to_o_net_conn_id

  htols(0x0001, &pacCommBufRunner);                                     /* uint16 connection serial number   */
  htols(405, &pacCommBufRunner);                                        /* uint16 vendor id                  */

  htoll(0x32000020, &pacCommBufRunner);                              	// uint32  originator serial number */

  *(pacCommBufRunner)++ = 0;                          					// uint8 conn_timeout_multiplier
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved

  htoll(1000000, &pacCommBufRunner);                       				// uint32  o_to_t_rpi

																		// Owner Exclusive, Conn Type Point to Point,
																		// Low Priority connection size 511 bytes
  htols(0x43ff, &pacCommBufRunner);                           		    // uint16  o_to_t_net_conn_parameters.
  /*
   Network Connection Parameters, Forward Open
   Bit 15 		- Redundant Owner (0=*Exclusive,Input-Only or Listen-Only, 1 = Redundant Owner in O->T direction)
   Bit 14 - 13  - Connection Type (00=Null, 01=Multicast, 10=*Point-to-Point, 11=Reserved)
   Bit 12 		- Reserved
   Bit 11 - 10 	- Priority   (00= Low, 01= High, 10 = *Scheduled, 11 = Urgent
   Bit 9 		- Fixed or Variable (0=*Fixed, 1=Variable)
   Bit 8 - 0 	- Connection Size   (Size of data up to 511)
  */

  htoll(1000000, &pacCommBufRunner);                       				// uint32  t_to_o_rpi

																		// Owner Exclusive, Conn Type Point to Point,
																		// Low Priority connection size 511 bytes
  htols(0x43ff, &pacCommBufRunner);                           		    // uint16  t_to_o_net_conn_parameters.
  /*
   Network Connection Parameters, Forward Open
   Bit 15 		- Redundant Owner (0=*Exclusive,Input-Only or Listen-Only, 1 = Redundant Owner in O->T direction)
   Bit 14 - 13  - Connection Type (00=Null, 01=Multicast, 10=*Point-to-Point, 11=Reserved)
   Bit 12 		- Reserved
   Bit 11 - 10 	- Priority   (00= Low, 01= High, 10 = *Scheduled, 11 = Urgent
   Bit 9 		- Fixed or Variable (0=*Fixed, 1=Variable)
   Bit 8 - 0 	- Connection Size   (Size of data up to 511)
  */

  *(pacCommBufRunner)++ = 0xA3;                        					// uint8 Transport Class Trigger, Server, Application Obj. Class 3

  *(pacCommBufRunner)++ = 2;						   					// uint8 Number of 16-bit words in the connection path field.
  *(pacCommBufRunner)++ = 0x20;                        					// uint8 connection path class, instance
  *(pacCommBufRunner)++ = 0x02;                        					// uint8 Message Router connection_path.logical_class
  *(pacCommBufRunner)++ = 0x24;                         				// uint8 connection_path.logical_instance_segment
  *(pacCommBufRunner)++ = 0x01;                         				// uint8 connection_path.logical_instance


  EIP_INT32 dataLen =  pacCommBufRunner - acIdCmdSpecData;
  htols(dataLen, &acIdLenBuf);       		/* update the command length field     */

  EIP_INT32 cmdLen = pacCommBufRunner - acIdDataItemLenBuf - 2;
  htols(cmdLen , &acIdDataItemLenBuf);     /* update the data length field     */

  return pacCommBufRunner - pa_pacCommBuf;
}
//------------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Large Forward Open Request (max 4095 bytes)
// ----------------------------------------------------------------------------
EIP_INT32 encapsulateSendLargeForwardOpenRequestMessage(EIP_BYTE *pa_pacCommBuf,
												  struct S_Service_Code_Request_Format *pa_cip_request)
{
   /*
	  SendForwardOpen message command
   */
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;
  EIP_BYTE *acIdLenBuf, *acIdCmdSpecData, *acIdDataItemLenBuf;

  // Seutp command request command
  htols(COMMAND_SENDRRDATA,&pacCommBufRunner);					   		/* uint16 command 			*/
																		/* uint16 command length    */
  acIdLenBuf = pacCommBufRunner; 		   /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  htoll(pa_cip_request->session_handle, &pacCommBufRunner);  			/* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                     						/* uint32 status 			*/

  htols(0x01, &pacCommBufRunner);     				   		 			/* uint8 sender context data[8]	*/
  htols(0x02, &pacCommBufRunner);
  htols(0x04, &pacCommBufRunner);
  htols(0x08, &pacCommBufRunner);

  htoll(0, &pacCommBufRunner);                							/* uint32 options               */

  acIdCmdSpecData = pacCommBufRunner;         /* record the starting point of the command specific data */

  htoll(0, &pacCommBufRunner);                							/* uint32 interface hangle      */
  htols(3, &pacCommBufRunner);                                          /* uint16 timeout               */

  htols(2, &pacCommBufRunner);                                          /* uint16 item count            */
  htols(0, &pacCommBufRunner);                                          /* uint16 address item type id  */
  htols(0, &pacCommBufRunner);                                          /* uint16 address item length   */
  htols(0xb2, &pacCommBufRunner);                                       /* uint16 data item type id     */

																		/* uint16 data item length      */
  acIdDataItemLenBuf =  pacCommBufRunner;    /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  *(pacCommBufRunner)++ = 0x5b; 			  							// uint8 service request

  *(pacCommBufRunner)++ = 2;            								// uint8 path size in words
  *(pacCommBufRunner)++ = 0x20;											// uint8 logical_class_segment
  *(pacCommBufRunner)++ = 0x06							; 	   		 	// uint8 class request
  *(pacCommBufRunner)++ = 0x24;											// uint8 logical_instance_segment
  *(pacCommBufRunner)++ = 0x01;											// uint8 instance request

  *(pacCommBufRunner)++ = 0x06;                              			// uint8 priority_time_tick

  *(pacCommBufRunner)++ = 40;                          					// uint8 timeout_ticks

  htoll(0, &pacCommBufRunner);                          				// uint32  o_to_t_net_conn_id
  htoll(1, &pacCommBufRunner);                           				// uint32  t_to_o_net_conn_id

  htols(0x0001, &pacCommBufRunner);                                     /* uint16 connection serial number   */
  htols(405, &pacCommBufRunner);                                        /* uint16 vendor id                  */

  htoll(0x32000020, &pacCommBufRunner);                              	// uint32  originator serial number */

  *(pacCommBufRunner)++ = 0;                          					// uint8 conn_timeout_multiplier
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved

  htoll(1000000, &pacCommBufRunner);                       				// uint32  o_to_t_rpi

																		// Owner Exclusive, Conn Type Point to Point,
																		// Low Priority connection size 4095 bytes
  htoll(0x42000fff, &pacCommBufRunner);                           	    // uint32  o_to_t_net_conn_parameters.

  htoll(1000000, &pacCommBufRunner);                       				// uint32  t_to_o_rpi

																		// Owner Exclusive, Conn Type Point to Point,
																		// Low Priority connection size 4095 bytes
  htoll(0x42000fff, &pacCommBufRunner);                           	    // uint32  t_to_o_net_conn_parameters.

  *(pacCommBufRunner)++ = 0xA3;                        					// uint8 Transport Class Trigger, Server, Application Obj. Class 3

  *(pacCommBufRunner)++ = 2;						   					// uint8 Number of 16-bit words in the connection path field.
  *(pacCommBufRunner)++ = 0x20;                        					// uint8 connection path class, instance
  *(pacCommBufRunner)++ = 0x02;                        					// uint8 Message Router connection_path.logical_class
  *(pacCommBufRunner)++ = 0x24;                         				// uint8 connection_path.logical_instance_segment
  *(pacCommBufRunner)++ = 0x01;                         				// uint8 connection_path.logical_instance


  EIP_INT32 dataLen =  pacCommBufRunner - acIdCmdSpecData;
  htols(dataLen, &acIdLenBuf);       		/* update the command length field     */

  EIP_INT32 cmdLen = pacCommBufRunner - acIdDataItemLenBuf - 2;
  htols(cmdLen , &acIdDataItemLenBuf);     /* update the data length field     */

  return pacCommBufRunner - pa_pacCommBuf;
}
//------------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Forward Open Request (max 1410 bytes)
// ----------------------------------------------------------------------------
EIP_INT32 encapsulateSendIOForwardOpenRequestMessage(EIP_BYTE *pa_pacCommBuf,
											 struct S_Service_Code_Request_Format *pa_cip_request)
{
   /*
	  SendForwardOpen message command
   */
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;
  EIP_BYTE *acIdLenBuf, *acIdCmdSpecData, *acIdDataItemLenBuf;

  srand(1234565);

  // Seutp command request command
  htols(COMMAND_SENDRRDATA,&pacCommBufRunner);					   		/* uint16 command 			*/
																		/* uint16 command length    */
  acIdLenBuf = pacCommBufRunner; 		   /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  htoll(pa_cip_request->session_handle, &pacCommBufRunner);   			/* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                     						/* uint32 status 			*/

  htols(0x01, &pacCommBufRunner);     				   		 			/* uint8 sender context data[8]	*/
  htols(0x02, &pacCommBufRunner);
  htols(0x04, &pacCommBufRunner);
  htols(0x08, &pacCommBufRunner);

  htoll(0, &pacCommBufRunner);                							/* uint32 options               */

  acIdCmdSpecData = pacCommBufRunner;         /* record the starting point of the command specific data */

  htoll(0, &pacCommBufRunner);                							/* uint32 interface hangle      */
  htols(3, &pacCommBufRunner);                                          /* uint16 timeout               */

  htols(2, &pacCommBufRunner);                                          /* uint16 item count            */
  htols(0, &pacCommBufRunner);                                          /* uint16 address item type id  */
  htols(0, &pacCommBufRunner);                                          /* uint16 address item length   */
  htols(0xb2, &pacCommBufRunner);                                       /* uint16 data item type id     */

																		/* uint16 data item length      */
  acIdDataItemLenBuf =  pacCommBufRunner;    /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  *(pacCommBufRunner)++ = 0x54; 			  							// uint8 service request

  *(pacCommBufRunner)++ = 2;            								// uint8 path size in words
  *(pacCommBufRunner)++ = 0x20;											// uint8 logical_class_segment
  *(pacCommBufRunner)++ = 0x06							; 	   		 	// uint8 class request
  *(pacCommBufRunner)++ = 0x24;											// uint8 logical_instance_segment
  *(pacCommBufRunner)++ = 0x01;											// uint8 instance request

  *(pacCommBufRunner)++ = 0x06;                              			// uint8 priority_time_tick

  *(pacCommBufRunner)++ = 40;                          					// uint8 timeout_ticks

  // setup to initial connection id's
  pa_cip_request->o_to_t_connection_id = 0;
  pa_cip_request->t_to_o_connection_id = (0xFFFFFFFF * rand()) / RAND_MAX;
  htoll(pa_cip_request->o_to_t_connection_id, &pacCommBufRunner);       // uint32  o_to_t_net_conn_id
  htoll(pa_cip_request->t_to_o_connection_id, &pacCommBufRunner);       // uint32  t_to_o_net_conn_id

  // setup the initial sequence numbers
  pa_cip_request->o_to_t_sequence_num = 1;
  pa_cip_request->t_to_o_sequence_num = 1;

  htols(0x1000, &pacCommBufRunner);                                     /* uint16 connection serial number   */
  htols(405, &pacCommBufRunner);                                        /* uint16 vendor id                  */

  htoll(0x32000000, &pacCommBufRunner);                              	// uint32  originator serial number */

  *(pacCommBufRunner)++ = 0;                          					// uint8 conn_timeout_multiplier
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved

  htoll(900000, &pacCommBufRunner);                       				// uint32  o_to_t_rpi 300ms

																		// Owner Exclusive, Conn Type Point to Point,
																		// Scheduled Priority connection size 8 bytes (2+6?)
  htols(0x4808, &pacCommBufRunner);                          		    // uint16  o_to_t_net_conn_parameters.
  htoll(100000, &pacCommBufRunner);                       				// uint32  t_to_o_rpi  300ms

  if(pa_cip_request->input_assembly_object == 0x64)
  {
																		// Owner Exclusive, Conn Type Point to Point,
																		// Scheduled Priority connection size 34 bytes (32+2?)
		htols(0x4822, &pacCommBufRunner);                      		    // uint16  t_to_o_net_conn_parameters.
  }
  else
  {
																		// Owner Exclusive, Conn Type Point to Point,
																		// Scheduled Priority connection size 298 bytes (296+2?)
		htols(0x492a, &pacCommBufRunner);                      		    // uint16  t_to_o_net_conn_parameters.

  }
  
  *(pacCommBufRunner)++ = 0x01;                        					// uint8 Transport Class Trigger, Client(0), Cyclic(0). Class 1

  *(pacCommBufRunner)++ = 4;						   					// uint8 Number of 16-bit words in the connection path field.
  *(pacCommBufRunner)++ = 0x20;                        					// uint8 connection path class, instance
  *(pacCommBufRunner)++ = 0x04;                        					// uint8 Assembly Object connection_path.logical_class
  *(pacCommBufRunner)++ = 0x24;                         				// uint8 connection_path.logical_instance_segment
  *(pacCommBufRunner)++ = 0x01;                         				// uint8 connection_path.logical_instance

  *(pacCommBufRunner)++ = 0x2c;                         				// uint8 connection_point o->t
  *(pacCommBufRunner)++ = pa_cip_request->output_assembly_object;    	// uint8 connection_point o->t
  *(pacCommBufRunner)++ = 0x2c;                         				// uint8 connection_point t->o
  *(pacCommBufRunner)++ = pa_cip_request->input_assembly_object;		// uint8 connection_point t->o

  EIP_INT32 dataLen =  pacCommBufRunner - acIdCmdSpecData;
  htols(dataLen, &acIdLenBuf);       		/* update the command length field     */

  EIP_INT32 cmdLen = pacCommBufRunner - acIdDataItemLenBuf - 2;
  htols(cmdLen , &acIdDataItemLenBuf);     /* update the data length field     */

  return pacCommBufRunner - pa_pacCommBuf;
}
//------------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Large Forward Open Request (max 4095 bytes)
// ----------------------------------------------------------------------------
EIP_INT32 encapsulateSendIOLargeForwardOpenRequestMessage(EIP_BYTE *pa_pacCommBuf,
											 struct S_Service_Code_Request_Format *pa_cip_request)
{
   /*
	  SendLargeForwardOpen message command
   */
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;
  EIP_BYTE *acIdLenBuf, *acIdCmdSpecData, *acIdDataItemLenBuf;
  EIP_UINT32 temp32;

  // Seutp command request command
  htols(COMMAND_SENDRRDATA,&pacCommBufRunner);					   		/* uint16 command 			*/
																		/* uint16 command length    */
  acIdLenBuf = pacCommBufRunner; 		   /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  htoll(pa_cip_request->session_handle, &pacCommBufRunner);   			/* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                     						/* uint32 status 			*/

  htols(0x00, &pacCommBufRunner);     				   		 			/* uint8 sender context data[8]	*/
  htols(0x00, &pacCommBufRunner);
  htols(0x00, &pacCommBufRunner);
  htols(0x00, &pacCommBufRunner);

  htoll(0, &pacCommBufRunner);                							/* uint32 options               */

  acIdCmdSpecData = pacCommBufRunner;         /* record the starting point of the command specific data */

  htoll(pa_cip_request->interface_handle, &pacCommBufRunner);    		/* uint32 interface handle      */
  htols(pa_cip_request->timeout, &pacCommBufRunner);                    /* uint16 timeout               */

  htols(2, &pacCommBufRunner);                                          /* uint16 item count            */
  htols(0, &pacCommBufRunner);                                          /* uint16 address item type id  */
  htols(0, &pacCommBufRunner);                                          /* uint16 address item length   */
  htols(0xb2, &pacCommBufRunner);                                       /* uint16 data item type id     */

																		/* uint16 data item length      */
  acIdDataItemLenBuf =  pacCommBufRunner;    /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  *(pacCommBufRunner)++ = 0x5B; 			  							    // uint8 service request

  *(pacCommBufRunner)++ = 2;            								    // uint8 path size in words
  *(pacCommBufRunner)++ = 0x20;											// uint8 logical_class_segment
  *(pacCommBufRunner)++ = 0x06; 	   		 							    // uint8 class request
  *(pacCommBufRunner)++ = 0x24;											// uint8 logical_instance_segment
  *(pacCommBufRunner)++ = 0x01;											// uint8 instance request

  *(pacCommBufRunner)++ = 0x06;                              			// uint8 priority_time_tick

  *(pacCommBufRunner)++ = pa_cip_request->timeout;				     	// uint8 timeout_ticks

  // setup to initial connection id's
  htoll(pa_cip_request->o_to_t_connection_id, &pacCommBufRunner);       // uint32  o_to_t_net_conn_id
  htoll(pa_cip_request->t_to_o_connection_id, &pacCommBufRunner);       // uint32  t_to_o_net_conn_id

  // setup the initial sequence numbers
  pa_cip_request->o_to_t_sequence_num = 1;
  pa_cip_request->t_to_o_sequence_num = 1;

  htols(pa_cip_request->connection_serial_number, &pacCommBufRunner);   /* uint16 connection serial number   */
  htols(pa_cip_request->vendor_id, &pacCommBufRunner);					/* uint16 vendor id                  */

  htoll(pa_cip_request->originator_serial_number, &pacCommBufRunner);   	// uint32  originator serial number */

  *(pacCommBufRunner)++ = 0;                          					// uint8 conn_timeout_multiplier
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved

  htoll(pa_cip_request->o_to_t_rpi, &pacCommBufRunner);          		// uint32  o_to_t_rpi (us)

																		// Owner Exclusive, Conn Type Point to Point,
																		// Scheduled Priority connection size x bytes (x+6?)
  temp32 = 0x48000000 + pa_cip_request->output_assembly_size + 6;		// Fixed sizes used for output assemblies.
  htoll(temp32, &pacCommBufRunner);                           			// uint32  o_to_t_net_conn_parameters.

  /*
   Network Connection Parameters, Large Forward Open
   Bit 31 		- Redundant Owner (0=Exclusive, Input-Only or Listen-Only, 1 = Redundant Owner in O->T direction)
   Bit 30 - 29  - Connection Type (00=Null, 01=Multicast, 10=Point-to-Point, 11=Reserved)
   Bit 28 		- Reserved
   Bit 27 - 26 	- Priority   (00= Low, 01= High, 10 = *Scheduled, 11 = Urgent
   Bit 25 		- Fixed or Variable (0=Fixed, 1 Variable)
   Bit 24 - 16	- Reserved
   Bit 15 - 0 	- Connection Size   (Size of data up to 65007)
  */

  htoll(pa_cip_request->t_to_o_rpi, &pacCommBufRunner);            		// uint32  t_to_o_rpi  (us)

  // Variable sizes used for input assemblies.
																		// Owner Exclusive, Conn Type Point to Point,
  temp32 = 0x4A000000 + pa_cip_request->input_assembly_size + 2;		    // Scheduled Priority connection size x bytes (x+2?)
  htoll(temp32, &pacCommBufRunner);										// uint32  t_to_o_net_conn_parameters.

  *(pacCommBufRunner)++ = 0x01;                        					// uint8 Transport Class Trigger, Client(0), Cyclic(0). Class 1

  *(pacCommBufRunner)++ = 4;						   					    // uint8 Number of 16-bit words in the connection path field.
  *(pacCommBufRunner)++ = 0x20;                        					// uint8 connection path class, instance
  *(pacCommBufRunner)++ = 0x04;                        					// uint8 Assembly Object connection_path.logical_class
  *(pacCommBufRunner)++ = 0x24;                         				// uint8 connection_path.logical_instance_segment
  *(pacCommBufRunner)++ = 0x01;                         				// uint8 connection_path.logical_instance

  *(pacCommBufRunner)++ = 0x2c;                         				// uint8 connection_point o->t
  *(pacCommBufRunner)++ = pa_cip_request->output_assembly_object;    	// uint8 connection_point o->t
  *(pacCommBufRunner)++ = 0x2c;                         				// uint8 connection_point t->o
  *(pacCommBufRunner)++ = pa_cip_request->input_assembly_object;		// uint8 connection_point t->o

  EIP_INT32 dataLen =  pacCommBufRunner - acIdCmdSpecData;
  htols(dataLen, &acIdLenBuf);       									/* update the command length field     */

  EIP_INT32 cmdLen = pacCommBufRunner - acIdDataItemLenBuf - 2;
  htols(cmdLen , &acIdDataItemLenBuf);									/* update the data length field     */

  return pacCommBufRunner - pa_pacCommBuf;
}
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------

EIP_INT32 encapsulateSendIOKeepAliveMessage(EIP_BYTE *pa_pacCommBuf,
											 struct S_Service_Code_Request_Format *pa_cip_request)
{
   /*
	  SendIOKeepAliveMessage command
   */
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;
  EIP_UINT16 i;

  // Seutp keep alive command
  htols(2, &pacCommBufRunner);                          			    /* uint16 item count            */
  htols(0x8002, &pacCommBufRunner);                     			    /* uint16 address item type id  */
  htols(8, &pacCommBufRunner);                          			    /* uint16 address item length   */
  htoll(pa_cip_request->o_to_t_connection_id, &pacCommBufRunner);  	/* uint32 connection id   */
  htoll(pa_cip_request->o_to_t_sequence_num, &pacCommBufRunner);   	/* uint32 sequence number */

  htols(0xb1, &pacCommBufRunner);                       			    /* uint16 data item type id     */
  
  i =  pa_cip_request->output_assembly_size + 6;
  htols(i, &pacCommBufRunner);                         			    /* uint16 data item length      */

  htols(pa_cip_request->o_to_t_sequence_num, &pacCommBufRunner);  	/* uint16 sequence number      */
  htoll(0x0003, &pacCommBufRunner);   								/* uint32 header               */

  // Send the output data
  for(i=0; i < pa_cip_request->output_assembly_size; i++)
  {	
	*(pacCommBufRunner)++ = pa_cip_request->data_buf[i];     		/* uint8 data                 */
  }
  
  return pacCommBufRunner - pa_pacCommBuf;
}
//------------------------------------------------------------------------------



//------------------------------------------------------------------------------
EIP_INT32 encapsulateSendUnitDataRequestMessage(EIP_BYTE *pa_pacCommBuf,
								   struct S_Service_Code_Request_Format *pa_cip_request)
{
   /*
	  SendUnitData message command
   */
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;
  EIP_BYTE *acIdLenBuf, *acIdCmdSpecData, *acIdDataItemLenBuf;

  // Seutp command request command
  htols(COMMAND_SENDUNITDATA,&pacCommBufRunner);					     	/* uint16 command 			*/
																		/* uint16 command length    */
  acIdLenBuf = pacCommBufRunner; 		   /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  htoll(pa_cip_request->session_handle, &pacCommBufRunner);       		/* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                     						/* uint32 status 			*/

  htols(0x00, &pacCommBufRunner);     				   		 			/* uint8 sender context data[8]	*/
  htols(0x00, &pacCommBufRunner);
  htols(0x00, &pacCommBufRunner);
  htols(0x00, &pacCommBufRunner);

  htoll(0, &pacCommBufRunner);                							/* uint32 options               */

  acIdCmdSpecData = pacCommBufRunner;         /* record the starting point of the command specific data */

  htoll(pa_cip_request->interface_handle, &pacCommBufRunner);    		/* uint32 interface hangle      */
  htols(pa_cip_request->timeout, &pacCommBufRunner);                    /* uint16 timeout               */

  htols(2, &pacCommBufRunner);                                          /* uint16 item count            */
  htols(0xa1, &pacCommBufRunner);                                       /* uint16 address item type id  */
  htols(4, &pacCommBufRunner);                                          /* uint16 address item length   */

  htoll(pa_cip_request->o_to_t_connection_id, &pacCommBufRunner); 		/* uint32 o->t connection id from fwd open reply */

  htols(0xb1, &pacCommBufRunner);                                       /* uint16 data item type id     */

																		/* uint16 data item length      */
  acIdDataItemLenBuf =  pacCommBufRunner;    /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  htols(pa_cip_request->connection_serial_number, &pacCommBufRunner);  	/* uint16 connection serial number fwd open reply */

  *(pacCommBufRunner)++ = 0x0e; 			  							    // uint8 service request
  *(pacCommBufRunner)++ = 3;            								    // uint8 path size in words
  *(pacCommBufRunner)++ = 0x20;											// uint8 logical_class_segment
  *(pacCommBufRunner)++ = pa_cip_request->logical_class;   				// uint8 class request
  *(pacCommBufRunner)++ = 0x24;											// uint8 logical_instance_segment
  *(pacCommBufRunner)++ = pa_cip_request->logical_instance;	  			// uint8 instance request
  *(pacCommBufRunner)++ = 0x30;											// uint8 logical_attribute_segment
  *(pacCommBufRunner)++ = pa_cip_request->logical_attribute;		 		// uint8 attribute request

  EIP_INT32 dataLen =  pacCommBufRunner - acIdCmdSpecData;
  htols(dataLen, &acIdLenBuf);       		/* update the command length field     */

  EIP_INT32 cmdLen = pacCommBufRunner - acIdDataItemLenBuf - 2;
  htols(cmdLen , &acIdDataItemLenBuf);     /* update the data length field     */

  return pacCommBufRunner - pa_pacCommBuf;
}
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
EIP_INT32 encapsulateSendForwardCloseRequestMessage(EIP_BYTE *pa_pacCommBuf, struct S_Service_Code_Request_Format *pa_cip_request)
{
   /*
	  SendForwardClose message command
   */
  EIP_UINT8 *pacCommBufRunner = pa_pacCommBuf;
  EIP_BYTE *acIdLenBuf, *acIdCmdSpecData, *acIdDataItemLenBuf;

  // Seutp command request command
  htols(COMMAND_SENDRRDATA,&pacCommBufRunner);					   		/* uint16 command 			*/
																		/* uint16 command length    */
  acIdLenBuf = pacCommBufRunner; 		   /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  htoll(pa_cip_request->session_handle, &pacCommBufRunner);    			/* uint32 session handle 	*/
  htoll(0, &pacCommBufRunner);                     						/* uint32 status 			*/

  htols(0x00, &pacCommBufRunner);     				   		 			/* uint8 sender context data[8]	*/
  htols(0x00, &pacCommBufRunner);
  htols(0x00, &pacCommBufRunner);
  htols(0x00, &pacCommBufRunner);
  
  htoll(0, &pacCommBufRunner);                							/* uint32 options               */

  acIdCmdSpecData = pacCommBufRunner;         /* record the starting point of the command specific data */

  htoll(pa_cip_request->interface_handle, &pacCommBufRunner);    		/* uint32 interface hangle      */
  htols(pa_cip_request->timeout, &pacCommBufRunner);                    /* uint16 timeout               */

  htols(2, &pacCommBufRunner);                                          /* uint16 item count            */
  htols(0, &pacCommBufRunner);                                          /* uint16 address item type id  */
  htols(0, &pacCommBufRunner);                                          /* uint16 address item length   */
  htols(0xb2, &pacCommBufRunner);                                       /* uint16 data item type id     */

																		/* uint16 data item length      */
  acIdDataItemLenBuf =  pacCommBufRunner;    /*at this location the real length will be inserted later*/
  pacCommBufRunner += 2;

  *(pacCommBufRunner)++ = 0x4e; 			  							// uint8 service request
  
  *(pacCommBufRunner)++ = 2;            								// uint8 path size in words
  *(pacCommBufRunner)++ = 0x20;										    // uint8 logical_class_segment
  *(pacCommBufRunner)++ = 0x06; 	   		 	                        // uint8 class request - Message Router.
  *(pacCommBufRunner)++ = 0x24;										    // uint8 logical_instance_segment
  *(pacCommBufRunner)++ = 0x01;										    // uint8 instance request

  *(pacCommBufRunner)++ = 0x06;                               		    // uint8 priority_time_tick

  *(pacCommBufRunner)++ = pa_cip_request->timeout;         			    // uint8 timeout_ticks

  htols(pa_cip_request->connection_serial_number, &pacCommBufRunner);   /* uint16 connection serial number   */
  htols(pa_cip_request->vendor_id, &pacCommBufRunner);                  /* uint16 vendor id                  */

  htoll(pa_cip_request->originator_serial_number, &pacCommBufRunner);  	/* uint32  originator serial number */

  *(pacCommBufRunner)++ = 2;						   					// uint8 Number of 16-bit words in the connection path field.
  *(pacCommBufRunner)++ = 0;                          					// uint8 reserved

  *(pacCommBufRunner)++ = 0x20;										    // uint8 logical_class_segment
  *(pacCommBufRunner)++ = 0x02; 	   		 	                        // uint8 class request - Connection Manager.
  *(pacCommBufRunner)++ = 0x24;										    // uint8 logical_instance_segment
  *(pacCommBufRunner)++ = 0x01;										    // uint8 instance request

  EIP_INT32 dataLen =  pacCommBufRunner - acIdCmdSpecData;
  htols(dataLen, &acIdLenBuf);       		/* update the command length field     */

  EIP_INT32 cmdLen = pacCommBufRunner - acIdDataItemLenBuf - 2;
  htols(cmdLen , &acIdDataItemLenBuf);     /* update the data length field     */

  return pacCommBufRunner - pa_pacCommBuf;

}


//-----------------------------------------------------------------------------
void
determineDelayTime(EIP_BYTE *pa_acBufferStart,
    struct SDelayedEncapsulationMessage *pa_pstDelayedMessageBuffer)
{
  EIP_UINT16 unMaxDelayTime;
  pa_acBufferStart += 12; /* start of the sender context */
  unMaxDelayTime = ltohs(&pa_acBufferStart);
  if (0 == unMaxDelayTime)
    {
      unMaxDelayTime = 2000;
    }
  else if (500 > unMaxDelayTime)
    {
      unMaxDelayTime = 500;
    }

  pa_pstDelayedMessageBuffer->m_unTimeOut = (unMaxDelayTime * rand()) / RAND_MAX
      + 1;
}

/*   void RegisterSession(struct S_Encapsulation_Data *pa_S_ReceiveData)
 *   Check supported protocol, generate session handle, send replay back to originator.
 *      pa_nSockfd       socket this request is associated to. Needed for double register check
 *      pa_S_ReceiveData pointer to received data with request/response.
 */
void
handleReceivedRegisterSessionCmd(EIP_INT32 pa_nSockfd,
    struct S_Encapsulation_Data * pa_stReceiveData)
{
  EIP_INT32 i;
  EIP_INT32 nSessionIndex = 0;
  EIP_UINT8 *pacBuf;
  EIP_UINT16 nProtocolVersion = ltohs(
      &pa_stReceiveData->m_acCurrentCommBufferPos);
  EIP_UINT16 nOptionFlag = ltohs(&pa_stReceiveData->m_acCurrentCommBufferPos);

  /* check if requested protocol version is supported and the register session option flag is zero*/
  if ((0 < nProtocolVersion) && (nProtocolVersion <= SUPPORTED_PROTOCOL_VERSION)
      && (0 == nOptionFlag))
    { /*Option field should be zero*/
      /* check if the socket has already a session open */
      for (i = 0; i < OPENER_NUMBER_OF_SUPPORTED_SESSIONS; ++i)
        {
          if (anRegisteredSessions[i] == pa_nSockfd)
            {
              /* the socket has already registered a session this is not allowed*/
              pa_stReceiveData->nSession_handle = i + 1; /*return the already assigned session back, the cip spec is not clear about this needs to be tested*/
              pa_stReceiveData->nStatus =
                  OPENER_ENCAP_STATUS_UNSUPPORTED_PROTOCOL;
              nSessionIndex = INVALID_SESSION;
              pacBuf =
                  &pa_stReceiveData->m_acCommBufferStart[ENCAPSULATION_HEADER_SESSION_HANDLE_POS];
              htoll(pa_stReceiveData->nSession_handle, &pacBuf); /*encapsulate_data will not update the session handle so we have to do it here by hand*/
              break;
            }
        }

      if (INVALID_SESSION != nSessionIndex)
        {
          nSessionIndex = getFreeSessionIndex();
          if (INVALID_SESSION == nSessionIndex) /* no more sessions available */
            {
              pa_stReceiveData->nStatus = OPENER_ENCAP_STATUS_INSUFFICIENT_MEM;
            }
          else
            { /* successful session registered */
              anRegisteredSessions[nSessionIndex] = pa_nSockfd; /* store associated socket */
              pa_stReceiveData->nSession_handle = nSessionIndex + 1;
              pa_stReceiveData->nStatus = OPENER_ENCAP_STATUS_SUCCESS;
              pacBuf =
                  &pa_stReceiveData->m_acCommBufferStart[ENCAPSULATION_HEADER_SESSION_HANDLE_POS];
              htoll(pa_stReceiveData->nSession_handle, &pacBuf); /*encapsulate_data will not update the session handle so we have to do it here by hand*/
            }
        }
    }
  else
    { /* protocol not supported */
      pa_stReceiveData->nStatus = OPENER_ENCAP_STATUS_UNSUPPORTED_PROTOCOL;
    }

  pa_stReceiveData->nData_length = 4;
}

/*   INT8 UnregisterSession(struct S_Encapsulation_Data *pa_S_ReceiveData)
 *   close all corresponding TCP connections and delete session handle.
 *      pa_S_ReceiveData pointer to unregister session request with corresponding socket handle.
 */
EIP_STATUS
handleReceivedUnregisterSessionCmd(
    struct S_Encapsulation_Data * pa_stReceiveData)
{
#ifdef __EIP_SERVER__
	EIP_INT32 i;

  if ((0 < pa_stReceiveData->nSession_handle)
      && (pa_stReceiveData->nSession_handle
          <= OPENER_NUMBER_OF_SUPPORTED_SESSIONS))
    {
      i = pa_stReceiveData->nSession_handle - 1;
      if (EIP_INVALID_SOCKET != anRegisteredSessions[i])
        {
          IApp_CloseSocket(anRegisteredSessions[i]);
          anRegisteredSessions[i] = EIP_INVALID_SOCKET;
          return EIP_OK;
        }
    }

  /* no such session registered */
  pa_stReceiveData->nData_length = 0;
  pa_stReceiveData->nStatus = OPENER_ENCAP_STATUS_INVALID_SESSION_HANDLE;
#endif // #ifdef __EIP_SERVER__

  return EIP_OK_SEND;
}

/*   INT8 SendUnitData(struct S_Encapsulation_Data *pa_S_ReceiveData)
 *   Call Connection Manager.
 *      pa_S_ReceiveData pointer to structure with data and header information.
 */
EIP_STATUS
handleReceivedSendUnitDataCmd(struct S_Encapsulation_Data * pa_stReceiveData)
{
#ifdef __EIP_SERVER__
  EIP_INT16 nSendSize;
  EIP_STATUS eRetVal = EIP_OK_SEND;

  if (pa_stReceiveData->nData_length >= 6)
    {
      /* Command specific data UDINT .. Interface Handle, UINT .. Timeout, CPF packets */
      /* don't use the data yet */
      ltohl(&pa_stReceiveData->m_acCurrentCommBufferPos); /* skip over null interface handle*/
      ltohs(&pa_stReceiveData->m_acCurrentCommBufferPos); /* skip over unused timeout value*/
      pa_stReceiveData->nData_length -= 6; /* the rest is in CPF format*/

      if (EIP_ERROR != checkRegisteredSessions(pa_stReceiveData)) /* see if the EIP session is registered*/
        {
          nSendSize =
              notifyConnectedCPF(pa_stReceiveData,
                  &pa_stReceiveData->m_acCommBufferStart[ENCAPSULATION_HEADER_LENGTH]);

          if (0 < nSendSize)
            { /* need to send reply */
              pa_stReceiveData->nData_length = nSendSize;
            }
          else
            {
              eRetVal = EIP_ERROR;
            }
        }
      else
        { /* received a package with non registered session handle */
          pa_stReceiveData->nData_length = 0;
          pa_stReceiveData->nStatus = OPENER_ENCAP_STATUS_INVALID_SESSION_HANDLE;
        }
    }
  return eRetVal;
#else 
  return EIP_ERROR;
#endif #ifdef __EIP_SERVER__
}

/*   INT8 SendRRData(struct S_Encapsulation_Data *pa_stReceiveData)
 *   Call UCMM or Message Router if UCMM not implemented.
 *      pa_stReceiveData pointer to structure with data and header information.
 *  return status 	0 .. success.
 * 					-1 .. error
 */
EIP_STATUS
handleReceivedSendRRDataCmd(struct S_Encapsulation_Data * pa_stReceiveData)
{
#ifdef __EIP_SERVER__  
  EIP_INT16 nSendSize;
  EIP_STATUS eRetVal = EIP_OK_SEND;

  if (pa_stReceiveData->nData_length >= 6)
    {
      /* Commandspecific data UDINT .. Interface Handle, UINT .. Timeout, CPF packets */
      /* don't use the data yet */
      ltohl(&pa_stReceiveData->m_acCurrentCommBufferPos); /* skip over null interface handle*/
      ltohs(&pa_stReceiveData->m_acCurrentCommBufferPos); /* skip over unused timeout value*/
      pa_stReceiveData->nData_length -= 6; /* the rest is in CPF format*/

      if (EIP_ERROR != checkRegisteredSessions(pa_stReceiveData)) /* see if the EIP session is registered*/
        {
          nSendSize =
              notifyCPF(pa_stReceiveData,
                  &pa_stReceiveData->m_acCommBufferStart[ENCAPSULATION_HEADER_LENGTH]);

          if (nSendSize >= 0)
            { /* need to send reply */
              pa_stReceiveData->nData_length = nSendSize;
            }
          else
            {
              eRetVal = EIP_ERROR;
            }
        }
      else
        { /* received a package with non registered session handle */
          pa_stReceiveData->nData_length = 0;
          pa_stReceiveData->nStatus = OPENER_ENCAP_STATUS_INVALID_SESSION_HANDLE;
        }
    }
  return eRetVal;
#else
  return EIP_ERROR;
#endif // #ifdef __EIP_SERVER__
}

/*   INT8 getFreeSessionIndex()
 *   search for available sessions an return index.
 *  return index of free session in anRegisteredSessions.
 * 			-1 .. no free session available
 */
EIP_INT32
getFreeSessionIndex(void)
{
  EIP_INT32 i;
  for (i = 0; i < OPENER_NUMBER_OF_SUPPORTED_SESSIONS; i++)
    {
      if (EIP_INVALID_SOCKET == anRegisteredSessions[i])
        {
          return i;
        }
    }
  return -1;
}

/*   INT16 createEncapsulationStructure(INT16 pa_sockfd, INT8 *pa_buf, UINT16 pa_length, struct S_Encapsulation_Data *pa_S_Data)
 *   copy data from pa_buf in little endian to host in structure.
 *      pa_length	length of the data in pa_buf.
 *      pa_S_Data	structure to which data shall be copied
 *  return difference between bytes in pa_buf an data_length
 *  		0 .. full package received
 * 			>0 .. more than one packet received
 * 			<0 .. only fragment of data portion received
 */EIP_INT16
createEncapsulationStructure(EIP_UINT8 * pa_buf, /* receive buffer*/
EIP_INT32 pa_length, /* size of stuff in  buffer (might be more than one message)*/
struct S_Encapsulation_Data * pa_stData) /* the struct to be created*/

{
  pa_stData->m_acCommBufferStart = pa_buf;
  pa_stData->nCommand_code = ltohs(&pa_buf);
  pa_stData->nData_length = ltohs(&pa_buf);
  pa_stData->nSession_handle = ltohl(&pa_buf);
  pa_stData->nStatus = ltohl(&pa_buf);
  /*memcpy(pa_stData->anSender_context, pa_buf, SENDER_CONTEXT_SIZE);*/
  pa_buf += SENDER_CONTEXT_SIZE;
  pa_stData->nOptions = ltohl(&pa_buf);
  pa_stData->m_acCurrentCommBufferPos = pa_buf;
  return (pa_length - ENCAPSULATION_HEADER_LENGTH - pa_stData->nData_length);
}

/*   INT8 checkRegisteredSessions(struct S_Encapsulation_Data *pa_S_ReceiveData)
 *   check if received package belongs to registered session.
 *      pa_stReceiveData received data.
 *  return 0 .. Session registered
 *  		-1 .. invalid session -> return unsupported command received
 */
EIP_STATUS
checkRegisteredSessions(struct S_Encapsulation_Data * pa_stReceiveData)
{

  if ((0 < pa_stReceiveData->nSession_handle)
      && (pa_stReceiveData->nSession_handle
          <= OPENER_NUMBER_OF_SUPPORTED_SESSIONS))
    {
      if (EIP_INVALID_SOCKET
          != anRegisteredSessions[pa_stReceiveData->nSession_handle - 1])
        {
          return EIP_OK;
        }
    }

  return EIP_ERROR;
}

void
closeSession(EIP_INT32 pa_nSocket)
{
#ifdef __EIP_SERVER__
  EIP_INT32 i;
  for (i = 0; i < OPENER_NUMBER_OF_SUPPORTED_SESSIONS; ++i)
    {
      if (anRegisteredSessions[i] == pa_nSocket)
        {
          IApp_CloseSocket(pa_nSocket);
          anRegisteredSessions[i] = EIP_INVALID_SOCKET;
          break;
        }
    }
#endif // #ifdef __EIP_SERVER__
}

void
encapShutDown(void)
{
#ifdef __EIP_SERVER__
  EIP_INT32 i;
  for (i = 0; i < OPENER_NUMBER_OF_SUPPORTED_SESSIONS; ++i)
    {
      if (EIP_INVALID_SOCKET != anRegisteredSessions[i])
        {
          IApp_CloseSocket(anRegisteredSessions[i]);
          anRegisteredSessions[i] = EIP_INVALID_SOCKET;
        }
    }
#endif // #ifdef __EIP_SERVER__
}

void
manageEncapsulationMessages()
{
#ifdef __EIP_SERVER__
  EIP_UINT32 i;
  for (i = 0; i < ENCAP_NUMBER_OF_SUPPORTED_DELAYED_ENCAP_MESSAGES; i++)
    {
      if (-1 != g_stDelayedEncapsulationMessages[i].m_nSocket)
        {
          g_stDelayedEncapsulationMessages[i].m_unTimeOut -= OPENER_TIMER_TICK;
          if (0 >= g_stDelayedEncapsulationMessages[i].m_unTimeOut)
            {
              IApp_SendUDPData(
                  &(g_stDelayedEncapsulationMessages[i].m_stSendToAddr),
                  g_stDelayedEncapsulationMessages[i].m_nSocket,
                  &(g_stDelayedEncapsulationMessages[i].m_anMsg[0]),
                  g_stDelayedEncapsulationMessages[i].m_unMessageSize);
              g_stDelayedEncapsulationMessages[i].m_nSocket = -1;
            }
        }
    }
#endif // #ifdef __EIP_SERVER__
}
